//
//  GDSTaskController.m
//  Fugu
//
//  Created by Tim McIntosh on 2005.10.05.
//  Copyright 2005 Green Dome Software. All rights reserved.
//

#import "GDSTaskController.h"


NSString* const GDSTaskActionResultLastReadKey    = @"Last Read";
NSString* const GDSTaskActionResultRawOutputKey   = @"Raw Output";
NSString* const GDSTaskActionResultErrorStringKey = @"Error String";


@implementation GDSTaskActionMonitor

+ (GDSTaskActionMonitor*)monitor
{
    return [[[GDSTaskActionMonitor alloc] init] autorelease];
}

- (id)init
{
    self = [super init];

    if (self != nil)
    {
        NSMutableDictionary* results = [NSMutableDictionary dictionary];

        [results setValue: [NSMutableString string]
                   forKey: GDSTaskActionResultRawOutputKey];

        [self setResults: results];

        [self setOutcome: GDSTaskActionInProgress];
    }

    return self;
}

- (void) dealloc
{
    [mResults release];
    [super dealloc];
}

- (void)setOutcome:(GDSTaskActionOutcome)outcome
{
    mOutcome = outcome;
}

- (GDSTaskActionOutcome)outcome
{
    return mOutcome;
}

- (void)setResults:(NSMutableDictionary*)results
{
    if (results != mResults)
    {
        [mResults release];
        mResults = [results retain];
    }
}

- (NSMutableDictionary*)results
{
    return mResults;
}

@end


@interface GDSTaskController(PrivateMethods)
- (void)setCurrentActionMonitor:(GDSTaskActionMonitor*)monitor;
@end


@implementation GDSTaskController

- (GDSTaskController*)initWithLaunchPath:(NSString*)path arguments:(NSArray*)arguments
{
    self = [super init];

    if (self != nil)
    {
        mTask = [[NSTask alloc] init];

        [mTask setLaunchPath: path];
        [mTask setArguments: arguments];

        [ [ NSNotificationCenter defaultCenter ] addObserver: self
                                                    selector: @selector(taskDidTerminateNotification:)
                                                        name: NSTaskDidTerminateNotification
                                                      object: mTask ];
    }

    return self;
}

- (void) dealloc
{
    [self setDelegate: nil];

    // Normally, task should already be terminated.
    NS_DURING
        kill([mTask processIdentifier], SIGKILL);
//XXX causes hang        [mTask waitUntilExit];
    NS_HANDLER
        /* Exception raised if not already launched - ignore it. */;
    NS_ENDHANDLER

    [[NSNotificationCenter defaultCenter] removeObserver: self];

    // In case Controller is deallocated prematurely:
    [[mCurrentActionMonitor results] setValue: @"GDSTaskController deallocated prematurely"
                                       forKey: GDSTaskActionResultErrorStringKey];
    [mCurrentActionMonitor setOutcome: GDSTaskActionFailure];

    [mCurrentActionMonitor release];
    [mTerminal release];
    [mTask release];

    [super dealloc];
}

- (NSTask*)task
{
    return mTask;
}

- (void)setCurrentActionMonitor:(GDSTaskActionMonitor*)monitor
{
    if (monitor != mCurrentActionMonitor)
    {
        [mCurrentActionMonitor release];
        mCurrentActionMonitor = [monitor retain];
    }
}

- (id)delegate
{
    return mDelegate;
}

- (void)setDelegate:(id)delegate
{
    mDelegate = delegate;
}

- (GDSTaskActionMonitor*)launch
{
    NSAssert3(mTask != nil && mTerminal == nil && mCurrentActionMonitor == nil,
              @"mTask = %p, mTerminal = %p, mCurrentActionMonitor = %p", mTask, mTerminal, mCurrentActionMonitor);

    mTerminal = [[GDSPseudoTerminal pseudoTerminalByLaunchingTask: mTask] retain];

    // Don't want echoing, carriage returns, etc.
    [mTerminal makeRaw];

    NSLog(@"PTY is %@", mTerminal);

    [self setCurrentActionMonitor: [GDSTaskActionMonitor monitor]];

    // Read data in background; will call readCompletionNotification:

    NSAssert(mTerminal != nil, @"Terminal is nil!");
    NSAssert([mTerminal masterFileHandle] != nil, @"Terminal master file handle is nil!");

    [[NSNotificationCenter defaultCenter] addObserver: self
                                             selector: @selector(readCompletionNotification:)
                                                 name: NSFileHandleReadCompletionNotification
                                               object: [mTerminal masterFileHandle]];

    [[mTerminal masterFileHandle] readInBackgroundAndNotify];

//    NSLog(@"Task %@ launched.", mTask);

    return mCurrentActionMonitor;
}

- (GDSTaskActionMonitor*)sendCommand:(NSString*)command
{
    NSAssert(mTerminal != nil, @"sendCommand called prior to launch!");

    NSAssert([mCurrentActionMonitor outcome] != GDSTaskActionInProgress, @"sendCommand called before previous command complete!");

    [self setCurrentActionMonitor: [GDSTaskActionMonitor monitor]];

    [self sendResponse: command];

    return mCurrentActionMonitor;
}

- (void)sendResponse:(NSString*)response
{
    NSData* data = [[response stringByAppendingString: @"\n"] dataUsingEncoding: NSUTF8StringEncoding];

    NSAssert([mCurrentActionMonitor outcome] == GDSTaskActionInProgress, @"sendResponse called with no action in progress!");

    [[mTerminal masterFileHandle] writeData: data];
    [[mTerminal masterFileHandle] synchronizeFile];
    
//    NSLog(@"SENT: %@ (%@)", data, response);
    
    [[mTerminal masterFileHandle] readInBackgroundAndNotify];
}

- (void)updateActionMonitor:(GDSTaskActionMonitor*)monitor
{
    if ([[self delegate] respondsToSelector:@selector(taskController:updateActionMonitor:)])
    {
        [[self delegate] taskController: self updateActionMonitor: mCurrentActionMonitor];
    }
}

- ( void )readCompletionNotification:( NSNotification * )aNotification
{
    NSData *readData = [ [ aNotification userInfo ] objectForKey: NSFileHandleNotificationDataItem ];

    if ([ readData length ] == 0)  // Note: file handle close => read 0 bytes
    {
        // see note in taskDidTerminateNotification
        [mCurrentActionMonitor setOutcome: GDSTaskActionSuccess];
        [self setCurrentActionMonitor: nil];
    }
    else
    {
        NSString*  readString = [[NSString alloc] initWithData: readData
                                                      encoding: NSUTF8StringEncoding];

        [[mCurrentActionMonitor results] setValue: readString forKey: GDSTaskActionResultLastReadKey ];
        [[[mCurrentActionMonitor results] valueForKey: GDSTaskActionResultRawOutputKey] appendString: readString];
        
        [readString release];
        
        [self updateActionMonitor: mCurrentActionMonitor];
        
        if ([mCurrentActionMonitor outcome] == GDSTaskActionInProgress)
        {
//            NSLog(@"Set up for next read...");
            [[mTerminal masterFileHandle] readInBackgroundAndNotify];
        }
        else
        {
            [self setCurrentActionMonitor: nil];
        }
    }
}


- ( void )taskDidTerminateNotification: ( NSNotification * )aNotification
{
//    NSLog(@"Task %@ terminated.", mTask);

    // Check termination status of task
    {
        int status = [mTask terminationStatus];

        // Note:  Premature task termination has been giving WIFSTOPPED()
        //        For example, when invoking sftp with a bad hostname.
        if (WIFEXITED(status) || WIFSTOPPED(status))
        {
            status = WEXITSTATUS(status);

            // NOTE:  Don't set outcome to complete on successful termination.
            // see http://www.cocoadev.com/index.pl?AsyncPOpen
            // "NSTask can send an NSTaskDidTerminate notify *before*
            // the final has-data-available notification."
            if (status != EXIT_SUCCESS)
            {
                [[mCurrentActionMonitor results] setValue: [NSString stringWithFormat: @"Command exited with error code %d", status]
                                                   forKey: GDSTaskActionResultErrorStringKey];

            }
         //   [mCurrentActionMonitor setOutcome: GDSTaskActionExited];
            //XXX readNotification will clean it up.                [self setCurrentActionMonitor: nil];
        }
        else
        {
            NSAssert1(WIFSIGNALED(status), @"Expected task termination due to signal (status = 0x%x)", status);

            [[mCurrentActionMonitor results] setValue: [NSString stringWithFormat: @"Command exited due to signal %d", WTERMSIG(status)]
                                               forKey: GDSTaskActionResultErrorStringKey];

//            [mCurrentActionMonitor setOutcome: GDSTaskActionExited];
//XXX readNotification will clean it up.            [self setCurrentActionMonitor: nil];
        }
    }
}


@end
