/*
 * Copyright (c) 2003 Regents of The University of Michigan.
 * All Rights Reserved.  See COPYRIGHT.
 */
 
 /*
  * Icon caching category for NSImage.
  * Should really be NSImage(Icon)
  */

#import "NSImage(IconForType).h"

#include <sys/types.h>
#include <sys/stat.h>

static NSMutableDictionary      *cachedIconDictionary = nil;

@implementation NSImage(IconForType)

+ ( NSImage * )iconForType: ( NSString * )extension
{
    static NSMutableDictionary	*images = nil;
    NSImage			*icon = nil;
    NSString			*ext = extension;
    
    if ( extension == nil ) {
        return( nil );
    }
    
    if ( [ ext isEqualToString: @"" ] ) {
        ext = @"'doc '";
    }
    
    if ( images == nil ) {
        images = [[ NSMutableDictionary alloc ] init ];
    }
    
    if (( icon = [ images objectForKey: ext ] ) == nil ) {
        if (( icon = [[ NSWorkspace sharedWorkspace ] iconForFileType: ext ] ) != nil ) {
            [ icon optimizeFor16x16Icon ];
            [ images setObject: icon forKey: ext ];
        }
    }
    
    return( icon );
}

+ ( NSImage * )iconForFile: ( NSString * )file
{
    NSMutableDictionary         *icons = nil;
    NSImage			*icon = nil;
    NSString                    *parent = nil;
    
    if ( file == nil || [ file length ] == 0 ) {
        return( nil );
    }
    
    parent = [ file stringByDeletingLastPathComponent ];
    
    if (( icons = [ cachedIconDictionary objectForKey: parent ] ) == nil ) {
        return( [ NSImage iconForType: file ] );
    }
    if (( icon = [ icons objectForKey: file ] ) == nil ) {
        return( [ NSImage iconForType: file ] );
    }
    
    return( icon );
}

- ( void )optimizeFor16x16Icon
{
#if 0  // XXX - this is taking longer than not doing it (on 10.4, anyway). -TKM
    /* only cache 16x16 icons */
    NSEnumerator* e = [[ self representations ] objectEnumerator ];
    NSImageRep*   rep;

    while ( ( rep = [ e nextObject ] ) != nil )
    {
        if ( [ rep pixelsHigh ] != 16  && [[ self representations ] count ] > 1 )
        {
            [ self removeRepresentation: rep ];
        }
    }
#endif
    [ self setScalesWhenResized: YES ];
    [ self setSize: NSMakeSize( 16.0, 16.0 ) ];
}

+ ( void )cacheIconsForPath: ( NSString * )path contents: ( NSArray * )items
{
    NSMutableDictionary* icons = nil;

    if ( path != nil && items != nil || [ path length ] != 0 && [ items count ] != 0 )
    {
        if (( icons = [ cachedIconDictionary objectForKey: path ] ) != nil )
        {
            struct stat   st;

            /* we've already cached the icons at path. see if they're out of date */
            if ( stat( [ path UTF8String ], &st ) < 0 ) {
                /* we'll encounter this error and handle it better if SFTPNode */
                return;
            }

            NSDate*        stamp = [ icons objectForKey: @"CacheTimestamp" ];
            NSTimeInterval t = [ stamp timeIntervalSince1970 ];
            NSTimeInterval s = [[ NSDate date ] timeIntervalSince1970 ];

            if (( s - t ) < 10800 && t > st.st_mtime )
            {
                return;
            }
            
            /* if it's too old, delete it */
            icons = nil;
            [ cachedIconDictionary removeObjectForKey: path ];
        }

        NSWorkspace*  workspace = [ NSWorkspace sharedWorkspace ];

        NSEnumerator* e = [ items objectEnumerator ];
        NSString*     item;

        icons = [ NSMutableDictionary dictionary ];

        while ( ( item = [ e nextObject ] ) != nil )
        {
            NSAutoreleasePool* pool = [[ NSAutoreleasePool alloc ] init ];

            NSString* file = [ path stringByAppendingPathComponent: item ];
            NSImage*  icon = [ workspace iconForFile: file ];

            if ( icon != nil )
            {
                [ icon optimizeFor16x16Icon ];
                [ icons setObject: icon forKey: file ];
            }

            [ pool release ];
        }
        
        if ( cachedIconDictionary == nil )
        {
            cachedIconDictionary = [[ NSMutableDictionary alloc ] init ];
        }

        if ( [ cachedIconDictionary count ] >= 5 )
        {
            NSArray* sortedKeys = [ cachedIconDictionary keysSortedByValueUsingSelector: @selector( cacheDateCompare: ) ];

            [ cachedIconDictionary removeObjectForKey: [ sortedKeys lastObject ]];
        }

        [ icons setObject: [ NSDate date ] forKey: @"CacheTimestamp" ];
        [ cachedIconDictionary setObject: icons forKey: path ];
    }
}

@end
