/*
 * Copyright (c) 2003 Regents of The University of Michigan.
 * All Rights Reserved.  See COPYRIGHT.
 */

#import "SFTPTServer.h"
#import "SFTPController.h"
#import	"SFTPNode.h"
#import "NSString-UnknownEncoding.h"
#import "NSString(SSHAdditions).h"
#import "GDSPseudoTerminal.h"

#include <errno.h>
#include <string.h>
#include <sys/param.h>
#include <sys/wait.h>

#include "argcargv.h"
#include "typeforchar.h"

extern int	errno;
extern char	**environ;

/* used to set which field contains the filename */
static int      fncolumn = -1;

@implementation SFTPTServer

int		connecting = 0;
int		connected = 0;

+ ( void )connectWithPorts: ( NSArray * )ports
{
    NSAutoreleasePool		*pool = [[ NSAutoreleasePool alloc ] init ];
    NSConnection		*cnctnToController;
    SFTPTServer			*serverObject;
    
    cnctnToController = [ NSConnection connectionWithReceivePort:
                            [ ports objectAtIndex: 0 ]
                            sendPort: [ ports objectAtIndex: 1 ]];
                            
    serverObject = [[ self alloc ] init ];
    [ (( SFTPController * )[ cnctnToController rootProxy ] ) setServer: serverObject ];
    [ serverObject release ];
    
    [[ NSRunLoop currentRunLoop ] run ];  
    [ pool release ];
}

- ( id )init
{
    if ( self = [ super init ] )
    {
        _currentTransferName = nil;
        _sftpRemoteObjectList = nil;
        pseudoTerminal = nil;
    }

    return self;
}

- (void)dealloc
{
    [ [ NSNotificationCenter defaultCenter ] removeObserver: self ];

    [ pseudoTerminal release ];

    [ super dealloc ];
}

/* accessor methods */
- ( void )setCurrentTransferName: ( NSString * )name
{
    if ( _currentTransferName != nil ) {
	[ _currentTransferName release ];
	_currentTransferName = nil;
    }
    
    if ( name != nil ) {
	_currentTransferName = [[ NSString alloc ] initWithString: name ];
    } else {
	_currentTransferName = name;
    }
}

- ( NSString * )currentTransferName
{
    return( _currentTransferName );
}

- ( id )remoteObjectList
{
    return( _sftpRemoteObjectList );
}

- ( void )setRemoteObjectList: ( id )objectList
{
    if ( _sftpRemoteObjectList ) {
        [ _sftpRemoteObjectList release ];
        _sftpRemoteObjectList = nil;
    }
    if ( ! objectList ) {
        return;
    }
    
    _sftpRemoteObjectList = [ objectList retain ];
}
/* end accessor methods */

/* sftp/ftp output handler methods */
- ( BOOL )checkForPasswordPromptInBuffer: ( const char * )buf
{
#ifdef notdef
    NSArray             *prompts = nil;
#endif /* notdef */
    BOOL                hasPrompt = NO;
    int                 i, pnum = 0;
    char                *prompts[] = { "password", "passphrase",
                                    "Password:", "PASSCODE:",
                                    "Password for ", "Passcode for ",
				    "CryptoCard Challenge" };
                                    
    if ( buf == NULL ) {
        return( NO );
    }
    
    pnum = ( sizeof( prompts ) / sizeof( prompts[ 0 ] ));
    for ( i = 0; i < pnum; i++ ) {
        if ( strstr( buf, prompts[ i ] ) != NULL ) {
            hasPrompt = YES;
            break;
        }
    }
    
#ifdef notdef
    /* someday we'll allow custom prompt checks */
#endif /* notdef */
    
    return( hasPrompt );
}

- ( BOOL )bufferContainsError: ( const char * )buf
{
    BOOL                hasError = NO;
    int                 i, numerrs = 0;
    char                *errors[] = { "Permission denied",
                                    "Couldn't ", "Secure connection ",
                                    "No address associated with",
                                    "Connection refused",
                                    "Request for subsystem",
                                    "Cannot download",
                                    "ssh_exchange_identification",
                                    "Operation timed out"
                                    "no address associated with",
                                    "REMOTE HOST IDENTIFICATION HAS CHANGED" };
                                    
    if ( buf == NULL ) {
        return( NO );
    }
                                    
    numerrs = ( sizeof( errors ) / sizeof( errors[ 0 ] ));
    for ( i = 0; i < numerrs; i++ ) {
        if ( strstr( buf, errors[ i ] ) != NULL ) {
            hasError = YES;
            break;
        }
    }
    
    return( hasError );
}

- ( BOOL )hasDirectoryListingFormInBuffer: ( const char * )buf
{
    BOOL                hasDirListForm = NO;
    int                 i, numforms = 0;
    char                *lsforms[] = { "ls -l", "ls", "ls " };
    
    if ( buf == NULL ) {
        return( NO );
    }
    
    numforms = ( sizeof( lsforms ) / sizeof( lsforms[ 0 ] ));
    for ( i = 0; i < numforms; i++ ) {
        if ( strncmp( buf, lsforms[ i ], strlen( lsforms[ i ] )) == 0 ) {
            hasDirListForm = YES;
            break;
        }
    }
    
    return( hasDirListForm );
}

- ( BOOL )unknownHostKeyPromptInBuffer: ( const char * )buf
{
    BOOL                isPrompt = NO;
    int                 i, numprompts = 0;
    char                *prompts[] = { "The authenticity of ",
                                        "Host key not found " };
                                        
    numprompts = ( sizeof( prompts ) / sizeof( prompts[ 0 ] ));
    for ( i = 0; i < numprompts; i++ ) {
        if ( strncmp( buf, prompts[ i ], strlen( prompts[ i ] )) == 0 ) {
            isPrompt = YES;
            break;
        }
    }
    
    return( isPrompt );
}

- ( void )parseTransferProgressString: ( const char * )string isUploading: ( BOOL )uploading
{
    int			tac, i, pc_index = -1;
    char		*tmp, **tav, *p;
    char		*t_rate, *t_amount, *t_eta;
    
    if (( tmp = strdup( string )) == NULL ) {
	perror( "strdup" );
	exit( 2 );
    }
    
    if (( tac = argcargv( tmp, &tav )) < 5 ) {
	/* not a transfer progress line we're interested in */
	free( tmp );
	return;
    }
    
    for ( i = ( tac - 1 ); i >= 0; i-- ) {
	if (( p = strrchr( tav[ i ], '%' )) != NULL ) {
	    /* found the %-done field */
	    pc_index = i;
	    p = '\0';
	    break;
	}
    }
    
    t_amount = tav[ pc_index + 1 ];
    t_rate = tav[ pc_index + 2 ];
    
    if ( pc_index == ( tac - 5 )) {
	t_eta = tav[ pc_index + 3 ];
    } else {
	t_eta = "--:--";
    }
    
    if ( uploading ) {
	[ controller updateUploadProgressBarWithValue: strtod( tav[ pc_index ], NULL )
		    amountTransfered: [ NSString stringWithUTF8String: t_amount ]
		    transferRate: [ NSString stringWithUTF8String: t_rate ]
		    ETA: [ NSString stringWithFormat: @"%s ETA", t_eta ]];
    } else {
	[ controller updateDownloadProgressBarWithValue: strtod( tav[ pc_index ], NULL )
		    amountTransfered: [ NSString stringWithUTF8String: t_amount ]
		    transferRate: [ NSString stringWithUTF8String: t_rate ]
		    ETA: [ NSString stringWithFormat: @"%s ETA", t_eta ]];
    }
    
    free( tmp );
}
/* end sftp/ftp output handler methods */

- ( int )atSftpPrompt
{
    return( atprompt );
}

- ( NSString * )retrieveUnknownHostKey
{
    NSString            *key = @"";
    char                buf[ MAXPATHLEN * 2 ];
    char     ch;
    unsigned int i = 0;

    // XXX kludge
    do
    {
        NSData*   readData = [ [ pseudoTerminal masterFileHandle ] readDataOfLength: 1 ];
        
        if ( [ readData length ] != 1 )
        {
            NSLog(@"Key read error");
            return @"";
        }
    }
    while ( i < (MAXPATHLEN * 2 - 1) && (buf[i++] = ch) != '\n' );

    buf[i] = '\0';

    if (( key = [ NSString stringWithUTF8String: buf ] ) == nil ) {
        key = @"";
    }

    return( key );
}

- ( NSMutableDictionary * )remoteObjectFromSFTPLine: ( const char * )object
{
    int			j, tac, len;
    int			datecolumn = -1, ownercolumn = 2;
    char                line[ MAXPATHLEN * 2 ] = { 0 };
    char                *filename = NULL;
    char		**targv;
    char                *p;
    NSMutableDictionary	*infoDictionary = nil;
    NSString		*dateString = nil, *groupName = nil, *name = nil;
    NSData              *nameAsRawBytes = nil;

    if ( strncmp( object, "sftp> ", strlen( "sftp> " )) == 0 ) {
        return( nil );
    }
    
    if ( strlen( object ) >= sizeof( line )) {
        NSLog( @"%s: too long\n", object );
        return( nil );
    }
    strcpy( line, object );
    
    /* break up the string into components */
    if (( tac = argcargv( line, &targv )) <= 0 ) {
        return( nil );
    }
    
    /* 
     * much of the abstraction in here to handle an arbitrary number of fields
     * was suggested by Hugues Martel. He used Obj-C calls. C calls are used here,
     * since otherwise there's a lot of unnecessary conversion.
     * Many thanks to Hugues for this contribution.
     */
        
    /* SSH.com's sftp gives true ls -lF output. 		*/
    /* Do we need to add other chars here (>, /, @, =)? 	*/
    if ( tac > 0 ) {
        if (( p = strchr( targv[ ( tac - 1 ) ], '*' )) != NULL 
                    && *targv[ 0 ] == '-' ) {
            *p = '\0';
        }
    }
    
    /* SSH.com's sftp client writes dir name + : before listing */
    if ( tac == 1 && strcmp( targv[ 0 ], ".:" ) == 0 ) {
        fncolumn = 8;
        goto DOT_OR_DOTDOT;
    } else if ( tac == 1 ) {
	/* prevent crashes */
	goto DOT_OR_DOTDOT;
    }
    
    /*
     * find the filename column. The first line should contain
     * a filename '.' or './'.
     */
    for ( j = 0; j < tac; j++ ) {
        if ( strcmp( targv[ j ], "." ) == 0 || strcmp( targv[ j ], "./" ) == 0 ) {
            fncolumn = j;
            break;
        }
    }
    /* likewise, determine how many fields are used for the date.	*/
    /* based on code submitted by Hugues Martel.			*/
    if ( fncolumn == -1 ) {
        /* probably an invalid line, but might also be a VShell-like    */
        /* server at the root directory.				*/
        if ( isdigit( *targv[ 0 ] ) && strchr( targv[ 4 ], ':' ) != NULL ) {
            fncolumn = 5;
        } else if ( *targv[ 0 ] == 'd' || *targv[ 0 ] == '-'
                    && tac >= 9 ) {
            /* might also be OpenSSH on Cygwin, which doesn't display	*/
            /* a '.' or './' at the root directory. if so, handle it.	*/
            fncolumn = 8;
        }
    }
    if ( fncolumn == -1 || fncolumn > tac ) {
        return( nil );			/* invalid output */
    }

    for ( j = ( fncolumn - 1 ); j >= 0; j-- ) {
        if ( isalpha( *targv[ j ] )) { 	/* we've found the column containing the month */
            datecolumn = j;
            if ( datecolumn != 5 ) {
                int             ind = 0;
                
                NSLog( @"datecolumn: %d\tdate: %s", datecolumn, targv[ j ] );
                for ( ind = 0; ind < tac; ind++ ) {
                    NSLog( @"targv[ %d ]: %s", ind, targv[ ind ] );
                }
                NSLog( @"line: %s", object );
            }
            break;
        }
    }
    if ( datecolumn < 0 ) {	/* potentially dealing with old OpenSSH version */
        if ( *targv[ 0 ] == '0' && strlen( targv[ 0 ] ) > 1
                                && strchr( targv[ 4 ], ':' ) == NULL
                                && fncolumn == 5 ) {
            datecolumn = ( fncolumn - 1 );
            ownercolumn = 1;
        }
    }
    if ( datecolumn >= tac || datecolumn < 0 ) {
        return( nil );
    }
            
    dateString = [ NSString stringWithUTF8String: targv[ datecolumn ]];
    for ( j = ( datecolumn + 1 ); j < fncolumn; j++ ) {
        dateString = [ NSString stringWithFormat: @"%@ %s", dateString, targv[ j ]];
    }
    infoDictionary = [[ NSMutableDictionary alloc ] init ];
    [ infoDictionary setObject: dateString forKey: @"date" ];
    
    if ( datecolumn >= 1 ) {    /* size always comes before date */
        [ infoDictionary setObject: [ NSString stringWithUTF8String:
                                        targv[ ( datecolumn - 1 ) ]]
                            forKey: @"size" ];
    }
        
    if ( fncolumn > 0 && tac >= ( fncolumn + 1 )) {
        if ( tac > ( fncolumn + 1 )) {
            if ( strstr( targv[ 0 ], "sftp>" ) != NULL ) {
                goto DOT_OR_DOTDOT;
            }
            
            for ( j = fncolumn; j < tac; j++ ) {
                len += ( strlen( targv[ j ] ) + 1 );    /* +1 for spaces */
            }

            if (( filename = ( char * )malloc( len )) == NULL ) {
                NSLog( @"malloc: %s", strerror( errno ));
                exit( 2 );
            }
            strlcpy( filename, targv[ fncolumn ], len );
            
            for ( j = fncolumn + 1; j < tac; j++ ) {
                if ( strcmp( targv[ j ], "->" ) == 0 ) {
                    break;
                }
                strlcat( filename, " ", len );
                strlcat( filename, targv[ j ], len );
            }
            
            nameAsRawBytes = [ NSData dataWithBytes: filename length: strlen( filename ) ];
            name = [ NSString stringWithBytesOfUnknownEncoding: filename
                                                    length: strlen( filename ) ];
            free( filename );
        } else {
            if ( strcmp( ".", targv[ fncolumn ] ) == 0
                    || strcmp( "./", targv[ fncolumn ] ) == 0 /* VShell output */
                    || strcmp( "../", targv[ fncolumn ] ) == 0
                    || strcmp( "..", targv[ fncolumn ] ) == 0 ) goto DOT_OR_DOTDOT;
            
            nameAsRawBytes = [ NSData dataWithBytes: targv[ fncolumn ]
                                length: strlen( targv[ fncolumn ] ) ];
            name = [ NSString stringWithBytesOfUnknownEncoding: targv[ fncolumn ]
                                            length: strlen( targv[ fncolumn ] ) ];
        }

        if (( datecolumn - 1 ) == 0 ) {		/* dealing with a VShell server, probably. */
            [ infoDictionary setObject: @"N/A" forKey: @"owner" ];
            [ infoDictionary setObject: @"N/A" forKey: @"group" ];
            /* since VShell doesn't include a mode, we have to invent one */
            /* based on code submitted by Hugues Martel. */
            if ( [ name characterAtIndex: ( [ name length ] - 1 ) ] == '/' ) { /* directory */
                [ infoDictionary setObject: @"d---------" forKey: @"perm" ];
                [ infoDictionary setObject: @"directory" forKey: @"type" ];
            } else {
                [ infoDictionary setObject: @"----------" forKey: @"perm" ];
                [ infoDictionary setObject: @"file" forKey: @"type" ];
            }
        } else if (( datecolumn - 1 ) > 1 ) {	/* probably some unix variant */
            [ infoDictionary setObject: [ NSString stringWithUTF8String: targv[ ownercolumn ]]
                                    forKey: @"owner" ];
            groupName = [ NSString stringWithUTF8String: targv[ ( ownercolumn + 1 ) ]];
            /* possible to have group names containing spaces */
            for ( j = ( ownercolumn + 2 ); j < ( datecolumn - 1 ); j++ ) {
                groupName = [ NSString stringWithFormat: @"%@ %s", groupName, targv[ j ]];
            }
            [ infoDictionary setObject: groupName forKey: @"group" ];

            /* handle old OpenSSH server by translating output */
            if (( datecolumn + 1 ) == fncolumn && *targv[ 0 ] == '0' ) {
                [ infoDictionary setObject:
                                    [[ NSString stringWithUTF8String: targv[ 0 ]]
                                        stringRepresentationOfOctalMode ]
                                    forKey: @"perm" ];
            } else {
                [ infoDictionary setObject: [ NSString stringWithUTF8String: targv[ 0 ]]
                                    forKey: @"perm" ];
            }
            [ infoDictionary setObject:
                [ NSString stringWithUTF8String:
                typeforchar( [[ infoDictionary objectForKey: @"perm" ] characterAtIndex: 0 ] ) ]
                            forKey: @"type" ];
        }
        [ infoDictionary setObject: name forKey: @"name" ];
        [ infoDictionary setObject: nameAsRawBytes forKey: @"NameAsRawBytes" ];
    }
    
    return( [ infoDictionary autorelease ] );
    
DOT_OR_DOTDOT:
    if ( infoDictionary ) {
        [ infoDictionary release ];
    }
    return( nil );
}

- ( void )connectToServerWithParams: ( NSArray * )params
                     fromController: ( SFTPController * )aController
{
    NSAssert( !connecting && !connected, @"Attempt to connect when already connected!" );

    controller   = aController;
    sftpTask     = [ [ NSTask alloc ] init ];

    sethomedir = showrb = NO;
    threestrikes = 0;
    atprompt = pwsent = validpw = NO;
    was_uploading = was_downloading = NO;
    was_changing = was_removing = was_renaming = was_listing = NO;

    [ controller clearLog ]; 

//XXX{ move to controller!
    NSString		*sftppath = nil;

    if (( sftppath = [[ NSUserDefaults standardUserDefaults ]
                            objectForKey: @"SFTPBinaryPath" ] ) == nil ) {
        sftppath = @"/usr/bin/sftp";
    }
    if ( ! [[ NSFileManager defaultManager ] fileExistsAtPath: sftppath ] ) {
        NSLog( @"Couldn't find sftp binary!" );
        return;
    }
//XXX}

    [ sftpTask setLaunchPath: sftppath ];
    [ sftpTask setArguments: params ];

    connecting = YES;

    [ controller addToLog: [ NSString stringWithFormat: @"sftp launch path is %@.\n", sftppath ]];

    [ controller updateHostList ];	/* adds new host to pop-up list */
    [ controller setConnectedWindowTitle ];

    pseudoTerminal = [ [ GDSPseudoTerminal pseudoTerminalByLaunchingTask: sftpTask ] retain];

    NSLog(@"PTY is %@", pseudoTerminal);

    [ [ NSNotificationCenter defaultCenter ] addObserver: self
                                                selector: @selector(taskDidTerminateNotification:)
                                                    name: NSTaskDidTerminateNotification
                                                  object: sftpTask ];

    // Read data in background; will call readCompletionNotification:

    [ [ NSNotificationCenter defaultCenter ] addObserver: self
                                                selector: @selector(readCompletionNotification:)
                                                    name: NSFileHandleReadCompletionNotification
                                                  object: [ pseudoTerminal masterFileHandle ] ];

    [ [ pseudoTerminal masterFileHandle ] readInBackgroundAndNotify ];
}

- ( void )kill
{
    [ sftpTask terminate ];
}

- ( void )sendCommand: ( NSString * )cmd
{
    NSData* data = [ [ cmd stringByAppendingString: @"\n" ] dataUsingEncoding: NSUTF8StringEncoding ];

    [ [ pseudoTerminal masterFileHandle ] writeData: data ];
    [ [ pseudoTerminal masterFileHandle ] synchronizeFile ];
    
    NSLog(@"SENT: %@ (%@)", data, cmd);
}

- ( void )readCompletionNotification: ( NSNotification * )aNotification
{
    static const char nul = '\0';
    NSData *readData = [ [ aNotification userInfo ] objectForKey: @"NSFileHandleNotificationDataItem" ];

    if ( [ readData length ] == 0 )
    {
        // XXX - handle read errors
        NSLog(@"READ ERROR.");
    }
    else
    {
        readData = [ [ readData mutableCopy ] autorelease ];
        [ (NSMutableData*)readData appendBytes: &nul length: 1 ];

        const char		*buf = [ readData bytes ];

#if 1 //def DEBUG
        NSLog( @"RECV: %s", buf );
#endif /* DEBUG */
        
        if ( [ self checkForPasswordPromptInBuffer: buf ] && !validpw ) {
            if ( threestrikes > 0 ) {
                [ controller passError ];
            };
            if ( connecting ) [ controller requestPasswordWithPrompt: ( char * )buf ];
            pwsent = YES;
        } else if ( strstr(( char * )buf, "rename \"" ) != NULL ) {
            was_renaming = YES;
        } else if ( strstr(( char * )buf, "sftp> " ) != NULL ) {
            atprompt = YES;
            if ( !connected ) {
                pwsent++;	/* for key auth */
                validpw++;
                [ controller showRemoteFiles ];
                showrb++;
            } else if ( !sethomedir ) {
                [ controller getListing ];
                sethomedir++;
            } else if ( was_changing || was_renaming ) {
                [ controller getListing ];
                was_changing = NO;
                was_renaming = NO;
            } else {
                NSString* cmd;
                
                /* check to see if there's anything waiting to be uploaded */
                if ( [[ controller uploadQ ] count ] ) {
                    NSDictionary	*dict = [[ controller uploadQ ] objectAtIndex: 0 ];
                    
                    if ( [[ dict objectForKey: @"isdir" ] intValue ] ) {
                        cmd = [ NSString stringWithFormat: @"mkdir \"%@\"", [ dict objectForKey: @"pathfrombase" ] ];
                    } else {
                        char		*p = " ";
                        
                        if ( [[ NSUserDefaults standardUserDefaults ]
                                            boolForKey: @"RetainFileTimestamp" ] ) {
                            p = " -P ";
                        }
                        
                        cmd = [ NSString stringWithFormat: @"put%s\"%@\" \"%@\"",
                            p,
                            [ dict objectForKey: @"fullpath" ],
                            [ dict objectForKey: @"pathfrombase" ] ];
                    }
                    
                    [ self sendCommand: cmd ];         // XXX - handle exception

                    was_uploading = YES;
                    
                    [ self setCurrentTransferName: [[[[ controller uploadQ ] objectAtIndex: 0 ]
                                    objectForKey: @"fullpath" ] lastPathComponent ]];
                    [ controller showUploadProgress ];
                    [ controller updateUploadProgress: 0 ];
                } else if ( was_uploading ) {
                    was_uploading = NO;
                    [ self setCurrentTransferName: nil ];
                    [ controller updateUploadProgress: 0 ];
                }
                
                /* check download queue */
                if ( [[ controller downloadQ ] count ] ) {
                    NSDictionary	*dict = [[ controller downloadQ ] objectAtIndex: 0 ];
                    NSData		*nameAsRawData = [ dict objectForKey: @"rpath" ];
                    NSString        *transferName = nil;
                    char		*p = " ";
                    
                    if ( [[ NSUserDefaults standardUserDefaults ]
                                        boolForKey: @"RetainFileTimestamp" ] ) {
                        p = " -P ";
                    }
                    
                    cmd = [ NSString stringWithFormat: @"get%s\"%.*s\" \"%@\"",
                        p,
                        [ nameAsRawData length ], [ nameAsRawData bytes ],
                        [ dict objectForKey: @"lpath" ] ];
                    
                    [ self sendCommand: cmd ];         // XXX - handle exception

                    was_downloading = YES;
                    
                    transferName = [ NSString stringWithBytesOfUnknownEncoding: ( char * )[ nameAsRawData bytes ]
                                                                        length: [ nameAsRawData length ]];
                    [ self setCurrentTransferName: [ transferName lastPathComponent ]];
                    [ controller showDownloadProgressWithMessage:
                        ( char * )[[ transferName lastPathComponent ] UTF8String ]];
                    [ controller removeFirstItemFromDownloadQ ];
                } else if ( was_downloading ) {
                    was_downloading = NO;
                    [ controller finishedDownload ];
                    [ self setCurrentTransferName: nil ];
                }
                
                /* check remove queue */
                if ( [[ controller removeQ ] count ] ) {
                    was_removing = YES;
                    [ controller deleteFirstItemFromRemoveQueue ];
                } else if ( was_removing ) {
                    was_removing = NO;
                    [ controller getListing ];
                } else if ( was_listing ) {
                    was_listing = NO;
                    [ controller finishedCommand ];
                }
            }
        } else {
            atprompt = NO;
            
            if ( strncmp(( char * )buf, "Permission denied, ",
                         strlen( "Permission denied, " )) == 0 ) {
                pwsent = NO;
                threestrikes++;
            } else if ( [ self bufferContainsError: buf ] ) {
                [ controller connectionError: [ NSString stringWithUTF8String: buf ]];
            } else if ( [ self currentTransferName ] != nil ) {
                if ( strstr(( char * )buf, [[ self currentTransferName ] UTF8String ] ) != NULL
                     && strrchr(( char * )buf, '%' ) != NULL ) {
                    if ( was_downloading ) {
                        [ self parseTransferProgressString: ( char * )buf
                                               isUploading: NO ];
                    } else if ( was_uploading ) {
                        [ self parseTransferProgressString: ( char * )buf
                                               isUploading: YES ];
                    }
                }
            } else if ( strstr(( char * )buf, "passphrase for key" ) != NULL ) {
                pwsent = NO;
                threestrikes = 0;	/* if pubkey auth fails, password prompt will appear */
                [ controller requestPasswordWithPrompt: ( char * )buf ];
            } else if ( strstr(( char * )buf, "Changing owner on" ) != NULL
                        || strstr(( char * )buf, "Changing group on" )
                        || strstr(( char * )buf, "Changing mode on" )) {
                [ controller setBusyStatusWithMessage: [ NSString stringWithUTF8String:
                    ( void * )buf ]];
                was_changing = YES;
                if ( strstr(( char * )buf, "Couldn't " ) != NULL ) {
                    [ controller sessionError: [ NSString stringWithUTF8String: ( void * )buf ]];
                }
            } else if ( [ self unknownHostKeyPromptInBuffer: buf ] ) {
                NSMutableDictionary    *hostInfo = nil;
                
                hostInfo = [ NSMutableDictionary dictionaryWithObjectsAndKeys:
                    [ NSString stringWithUTF8String: buf ], @"msg",
                    [ self retrieveUnknownHostKey ], @"key", nil ];
                
                [ controller getContinueQueryForUnknownHost: ( NSDictionary * )hostInfo ];
            } else if ( strstr(( char * )buf, "Removing " ) != NULL ) {
                [ controller setBusyStatusWithMessage:
                    [ NSString stringWithUTF8String: ( char * )buf ]];
            }
        }
        
        /* moved to separate if block: sometimes ls and sftp> occur in same buffer */
        if ( [ self hasDirectoryListingFormInBuffer: buf ] && connected ) {
            [ controller addToLog: [ NSString stringWithUTF8String: ( char * )buf ]];
            was_listing = YES;
            [ self collectListing ];
            [ controller loadRemoteBrowserWithItems: [ self remoteObjectList ]];
        }
        
        if ( strstr(( char * )buf, "Remote working" ) != NULL ) {
            char		*p, *q, *tmp;
            
            tmp = strdup(( char * )buf );
            
            if (( q = strrchr( tmp, '\r' )) != NULL ) *q = '\0';
            
            p = strchr( tmp, '/' );
            
            [ controller setRemotePathPopUp:
                [ NSString stringWithBytesOfUnknownEncoding: p 
                                                     length: strlen( p ) ]];
            free( tmp );
        }
        
        if ( threestrikes >= 3 ) {
            [ controller cancelConnection: nil ];
        }
        
        if ( buf[ 0 ] != '\0' ) {
            [ controller addToLog: [ NSString stringWithUTF8String: ( void * )buf ]];
        }

        NSLog(@"Set up for next read...");
        [ [ pseudoTerminal masterFileHandle ] readInBackgroundAndNotify ];
    }
}

- ( void )taskDidTerminateNotification: ( NSNotification * )aNotification
{
    int		status;

    [ [ NSNotificationCenter defaultCenter ] removeObserver: self
                                                       name: NSFileHandleReadCompletionNotification
                                                     object: pseudoTerminal ];

    [ pseudoTerminal release ];
    pseudoTerminal = nil;

    [ [ NSNotificationCenter defaultCenter ] removeObserver: self
                                                    name: NSTaskDidTerminateNotification
                                                  object: sftpTask ];

    [ self setCurrentTransferName: nil ];
    connecting = connected = NO;

    [ controller cleanUp ];

    [ controller addToLog: @"\nsftp task ended.\n" ];

    status = [ sftpTask terminationStatus ];

    if ( WIFEXITED( status )) {
        [ controller addToLog: @"Normal exit\n" ];
    } else if ( WIFSIGNALED( status )) {
        [ controller addToLog: @"WIFSIGNALED: " ];
        [ controller addToLog: [ NSString stringWithFormat: @"signal = %d\n", status ]];
    } else if ( WIFSTOPPED( status )) {
        [ controller addToLog: @"WIFSTOPPED\n" ];
    }

    [ sftpTask release ];
    sftpTask   = nil;

    controller = nil;
}

// Run command & return response
- ( NSString * )executeCommand: ( NSString * )cmd
{
    [ self sendCommand: cmd ];

    //
    // Collect response from server
    //
    NSMutableString     *listing = nil;
    
    do {
        NSData* readData = [ [ pseudoTerminal masterFileHandle ] availableData ];
        
        // XXX handle exceptions
        
        if ( [ readData length ] == 0 )
        {
            // XXX error
            break;
        }
        else
        {
            NSString* readString = [ [ NSString alloc ] initWithData: readData
                                                            encoding: NSUTF8StringEncoding ];
            
            NSLog(@"LISTING: %@", readString);
            
            [ controller addToLog: readString ];
            
            if ( listing == nil )
            {
                listing = [ [ readString mutableCopy ] autorelease ];
            }
            else
            {
                [ listing appendString: readString ];
            }
            
            [ readString release ];
        }
        
    } while ( [ listing rangeOfString: @"sftp>" ].location == NSNotFound );
    
    
}

- ( void )collectListing
{
    //
    // Process response line-by-line
    //
    NSMutableArray* items = nil;
    NSEnumerator*   lineEnumerator = [ [ listing componentsSeparatedByString: @"\n" ] objectEnumerator ];
    NSString*       line;

    while ( (line = [ lineEnumerator nextObject ]) != nil ) {
        NSMutableDictionary* object;

        if ( [ self bufferContainsError: [ line UTF8String ] ] ) {
            [ controller sessionError: line ];
            continue;
        }
        
#ifdef SSH_COM_SUPPORT
        if ( [ line rangeOfSting: @"<Press any key" ].location != NSNotFound ) {
            /* SSH.com's sftp makes you hit a key to get to the prompt. Whee. */
            [ self sendCommand: @"" ];
            continue;
        }
#endif /* SSH_COM_SUPPORT */
        
        if (( object = [ self remoteObjectFromSFTPLine: [ line UTF8String ] ] ) != nil ) {
            if ( items == nil ) {
                items = [ NSMutableArray array ];
            }
            [ items addObject: object ];
        }
    }
    
    [ controller finishedCommand ];
    [ self setRemoteObjectList: items ];
}

@end
