//
//  SFTPClientTaskController.m
//  Fugu
//
//  Created by Tim McIntosh on 2005.10.06.
//  Copyright 2005 Green Dome Software. All rights reserved.
//

#import "SFTPClientTaskController.h"
#import "GDSTaskController.h"

#import "NSString+SFTPAdditions.h"
#import "NSArray(Extensions).h"

#import "SFTPLsCommand.h"
#import "SFTPPwdCommand.h"

@implementation NSDictionary(FileListMethods)

- (NSString*)name
{
    return [self objectForKey: @"name"];
}

- (NSString*)date
{
    return [self objectForKey: @"date"];
}

- (NSString*)size
{
    return [self objectForKey: @"size"];
}

- (NSString*)type
{
    return [self objectForKey: @"type"];
}

- (NSString*)perm
{
    return [self objectForKey: @"perm"];
}

- (NSString*)owner
{
    return [self objectForKey: @"owner"];
}

- (NSString*)group
{
    return [self objectForKey: @"group"];
}

@end

@interface SFTPClientTaskController(PrivateMethods)
+ (NSString*)sshLaunchPath;

- (void)setUser:(NSString*)user;
- (void)setServer:(NSString*)server;
- (void)setPort:(NSNumber*)port;

- (void)_sendCommand:(NSString*)command;
- (void)startNextCommand;

- (void)setCurrentCommand:(SFTPCommand*)command;

- (void)stopObserving;
@end


static NSString* const kSftpDefaultPath = @"/usr/bin/sftp";

static NSString* const kPrompt          = @"prompt";
static NSString* const kResponse        = @"response";


@implementation SFTPClientTaskController

+ (void)initialize
{
    [[NSUserDefaults standardUserDefaults] registerDefaults:
        [NSDictionary dictionaryWithObject: kSftpDefaultPath
                                    forKey: @"SFTPBinaryPath"]];

#ifndef GNUSTEP
    [self setKeys: [NSArray arrayWithObjects: @"currentCommand", nil ]
           triggerChangeNotificationsForDependentKey: @"commandQueue"];

    [self setKeys: [NSArray arrayWithObjects: @"state", @"commandQueue.count", @"currentCommand", nil ]
           triggerChangeNotificationsForDependentKey: @"isBusy"];

    [self setKeys: [NSArray arrayWithObject: @"state" ]
           triggerChangeNotificationsForDependentKey: @"isConnected"];

    [self setKeys: [NSArray arrayWithObject: @"state" ]
           triggerChangeNotificationsForDependentKey: @"sessionName"];
#endif
}

+ (NSString*)sftpLaunchPath
{
    NSString* path = [[NSUserDefaults standardUserDefaults] objectForKey: @"SFTPBinaryPath"];

    if (![[NSFileManager defaultManager] fileExistsAtPath: path])
    {
        NSLog(@"Warning: can't find SFTP program at %@, using default %@", path, kSftpDefaultPath);
        path = kSftpDefaultPath;
    }

    return path;
}

+ (NSString*)sshLaunchPath
{
    return [[[self sftpLaunchPath] stringByDeletingLastPathComponent] stringByAppendingPathComponent: @"ssh"];
}

+ (double)sshVersion
{
    double version = -1.0;  // unsupported version

//    NSLog(@"Getting SSH version.");
    
    GDSTaskController* tc = [[GDSTaskController alloc] initWithLaunchPath: [SFTPClientTaskController sshLaunchPath]
                                                                arguments: [NSArray arrayWithObject: @"-V"]];

    GDSTaskActionMonitor* monitor = [[tc launch] retain];
    
    do
    {
        [[NSRunLoop currentRunLoop] runMode: NSDefaultRunLoopMode
                                 beforeDate: [NSDate distantFuture]];
    }
    while ([monitor outcome] == GDSTaskActionInProgress);
    
    if ([monitor outcome] != GDSTaskActionSuccess)
    {
        NSLog(@"Warning: couldn't determine SSH version, ssh -V failed: %@",
              [[monitor results] valueForKey: GDSTaskActionResultErrorStringKey]);
    }
    else
    {
        NSString* versionString = [[monitor results] valueForKey: GDSTaskActionResultRawOutputKey];
        
        if (![versionString hasPrefix: @"OpenSSH_"])
        {
            NSLog(@"Warning: unsupported SSH version: %@", versionString);
        }
        else
        {
            version = [[versionString substringWithRange: NSMakeRange(8, 3)] doubleValue];
        }
    }
    
    [monitor release];

    [tc release];

    return version;
}

+ (BOOL)hasSupportedSftpVersion
{
    return [self sshVersion] > 0.0;
}

- (SFTPClientTaskController*)initWithServer:(NSString*)server
                                       port:(NSNumber*)port
                                       user:(NSString*)user
                                    options:(NSArray*)options
{
    if (self = [super init])
    {
        NSMutableArray* arguments = [NSMutableArray array];

        if (port != nil)
        {
            /* only specify port if given to us, so we don't override .ssh/config */
            [arguments addObject: [NSString stringWithFormat: @"-oPort=%hu", [port unsignedShortValue]]];
        }

        if (options != nil)
        {
            [arguments addObjectsFromArray: options];
        }

        [arguments addObject: [NSString stringWithFormat: @"%@@%@", user, server]];

//        NSLog(@"Arguments: %@", arguments);

        mTaskController = [[GDSTaskController alloc] initWithLaunchPath: [[self class] sftpLaunchPath]
                                                              arguments: arguments];

        [mTaskController setDelegate: self];

        mCommandQueue = [[NSMutableArray alloc] init];

        // Determine version once per instance
        mSshVersion = [[self class] sshVersion];

        [self setUser: user];
        [self setServer: server];
        [self setPort: port];
        
        mPath = @"/";  // Note: can't use setter here, as it queues an ls command
    }

    return self;
}

- (void) dealloc
{
    if ([self state] != SFTPClientTaskStateNotConnected)
    {
        [self disconnect];
    }

    [self setDelegate: nil];

    [self setError: nil];
    [self setPath: nil];
    [self setFiles: nil];

    [self stopObserving];

    [mCommandQueue release];
    mCommandQueue = nil;

    [mTaskController release];
    mTaskController = nil;

    [self setUser: nil];
    [self setServer: nil];
    [self setPort: nil];

    [super dealloc];
}

- (double)sshVersion
{
    return mSshVersion;
}

- (void)setUser:(NSString*)user
{
    // XXX doesn't handle case of setting same value
    [mUser release];
    mUser = [user copy];
}

- (void)setServer:(NSString*)server
{
    // XXX doesn't handle case of setting same value
    [mServer release];
    mServer = [server copy];
}

- (void)setPort:(NSNumber*)port
{
    // XXX doesn't handle case of setting same value
    [mPort release];
    mPort = [port copy];
}


- (NSString*)user
{
    return mUser;
}

- (NSString*)server
{
    return mServer;
}

- (NSNumber*)port
{
    return mPort;
}

- (NSString*)sessionName
{
    return ([self isConnected]
            ? [NSString stringWithFormat: @"%@@%@", [self user], [self server]]
            : nil);
}

- (id)delegate
{
    return mDelegate;
}

- (void)setDelegate:(id)delegate
{
    mDelegate = delegate;
}

- (void)setState:(SFTPClientTaskState)state
{
    mState = state;
}

- (SFTPClientTaskState)state
{
    return mState;
}

- (BOOL)isBusy
{
    BOOL isBusy;
    
    switch ([self state])
    {
        case SFTPClientTaskStateConnecting:
        case SFTPClientTaskStateBusy:
            isBusy = YES;
            break;
            
        case SFTPClientTaskStateReady:
            isBusy = [self currentCommand] != nil || [[self commandQueue] isNotEmpty];
            break;
            
        default:
            isBusy = NO;

    }

    return isBusy;
}

- (void)setError:(NSString*)error
{
    if (error != mError)
    {
        [mError release];
        mError = [error copy];

//        NSLog(@"Error set to \"%@\"", error ? error : @"nil");
    }
}

- (NSString*)error
{
    return mError;
}

// Called to resume operation after pausing due to error
- (void)resume
{
    NSAssert([self state] == SFTPClientTaskStateReady, @"Incorrect state for resume.");

    [self setError: nil];
    [self startNextCommand];
}

- (SFTPCommand*)currentCommand
{
    return mCurrentCommand;
}

- (void)setCurrentCommand:(SFTPCommand*)command
{
    if (command != mCurrentCommand)
    {
        [mCurrentCommand release];
        mCurrentCommand = [command retain];
    }
}

- (NSString*)path
{
    return mPath;
}

- (NSString*)absolutePathFromString:(NSString*)path
{
    return [path hasPrefix: @"/"] ? path : [[self path] stringByAppendingPathComponent: path];
}

- (void)setPath:(NSString*)path
{
    path = [self absolutePathFromString: path];

    if ([mPath isNotEqualTo: path])
    {
        [mPath release];

        mPath = [path copy];

        [self refresh];
    }
}

- (void)changeToParentPath
{
    [self setPath: [[self path] stringByDeletingLastPathComponent]];
}

- (void)setFiles:(NSArray*)files
{
    if (files != mFiles)
    {
        [mFiles release];
        mFiles = [files copy];
    }
}

- (NSArray*)files
{
    return mFiles;
}

- (void)stopObserving
{
    [mActionMonitor removeObserver: self forKeyPath: @"outcome"];
    mActionMonitor = nil;
}

- (void)observeMonitor:(GDSTaskActionMonitor*)monitor
{
    [self stopObserving];

    mActionMonitor = monitor;

#ifndef GNUSTEP
    [monitor addObserver: self
              forKeyPath: @"outcome"
                 options: NSKeyValueObservingOptionNew
                 context: nil];
#endif
}

- (void)connect
{
    [self setState: SFTPClientTaskStateConnecting];

    [self observeMonitor: [mTaskController launch]];
}

- (void)disconnect
{
    // FIXME - should issue quit or send signal if busy.  See SFTPController.m

    NSAssert([self state] != SFTPClientTaskStateNotConnected, @"disconnect called when not connected.");

    [ self flushQueue ];
    [self setCurrentCommand: nil];

    // XXX - clean this up.  Need to give user notification when waiting.
    NSTask* task = [mTaskController task];

    switch ([self state])
    {
        case SFTPClientTaskStateReady:
            [self _sendCommand: @"quit"];
            break;

        default:
            kill([task processIdentifier], SIGKILL);  // XXX - need category for this.
//            [task terminate];
    }

    // Wait up to 5 seconds for task exit -TBD

    NSDate* timeout = [NSDate dateWithTimeIntervalSinceNow: 5.0];

    while ([[mTaskController task] isRunning])
    {
        [[NSRunLoop currentRunLoop] runMode: NSDefaultRunLoopMode
                                 beforeDate: timeout];
        
        if ([[NSDate date] isGreaterThanOrEqualTo: timeout])
        {
            NSLog(@"Giving up on task...");
            break;
        }
    }

    [self setState: SFTPClientTaskStateNotConnected];

    [[self delegate] sftpClientDisconnected: self];
}

- (BOOL)isConnected
{
    // XXX key/value dependency
    return [self state] > SFTPClientTaskStatePassword;
}

- (void)_sendCommand:(NSString*)command
{
    [self setError: nil];
    [self observeMonitor: [mTaskController sendCommand: command]];
}

- (void)sendPassword:(NSString*)password
{
    NSAssert([self state] == SFTPClientTaskStateConnecting, @"Incorrect state");
    [self _sendCommand: password];
}

- (void)sendUnknownHostKeyResponse:(BOOL)continueConnecting
{
    NSAssert([self state] == SFTPClientTaskStateConnecting, @"Incorrect state");
    [self _sendCommand: continueConnecting ? @"yes" : @"no"];
}

- (void)startNextCommand
{
    // Commands may have inter-dependencies, so when an error occurs,
    // pause until instructed to resume by the delegate.
    if ([self state] == SFTPClientTaskStateReady && [self error] == nil)
    {
        SFTPCommand* command = nil;

#ifndef GNUSTEP
        [self willChangeValueForKey: @"commandQueue"];
#endif

        if ([mCommandQueue isNotEmpty])
        {
            command = [[mCommandQueue objectAtIndex: 0] retain];
            [mCommandQueue removeObjectAtIndex: 0];
        }

        [self setCurrentCommand: command];  // Note: must retain command before removing from queue
        [command release];

        if (command != nil)
        {
//            NSLog(@"Starting new command: %@", [command command]);

            [self setState: SFTPClientTaskStateBusy];
            [self _sendCommand: [[self currentCommand] command]];
        }

#ifndef GNUSTEP
        [self didChangeValueForKey: @"commandQueue"];
#endif
    }
}

- (void)queueCommand:(SFTPCommand*)command
{
#ifndef GNUSTEP
    [self willChangeValueForKey: @"commandQueue"];
#endif
    [command setController: self];
    [mCommandQueue addObject: command];
#ifndef GNUSTEP
    [self didChangeValueForKey: @"commandQueue"];
#endif

    [self startNextCommand];
}

- (void)cancelCommand:(SFTPCommand*)command
{
//    NSLog(@"Cancelling %@", command);

#ifndef GNUSTEP
    [self willChangeValueForKey: @"commandQueue"];
#endif
    [mCommandQueue removeObject: command];
#ifndef GNUSTEP
    [self didChangeValueForKey: @"commandQueue"];
#endif
}

- (void)flushQueue
{
#ifndef GNUSTEP
    [self willChangeValueForKey: @"commandQueue"];
#endif
    [mCommandQueue removeAllObjects];
#ifndef GNUSTEP
    [self didChangeValueForKey: @"commandQueue"];
#endif
}

- (NSArray*)commandQueue
{
    NSArray*     queue = [NSArray array];
    SFTPCommand* first = [self currentCommand];

    if (first != nil)
    {
        queue = [queue arrayByAddingObject: first];
    }

    if (mCommandQueue != nil)
    {
        queue = [queue arrayByAddingObjectsFromArray: mCommandQueue];
    }

    return queue;
}

- (BOOL)needsRefresh
{
    return mNeedsRefresh;
}

- (void)setNeedsRefresh:(BOOL)needsRefresh
{
    mNeedsRefresh = needsRefresh;
}

- (void)refresh
{
    if ([self path] != nil)
    {
        [self setNeedsRefresh: NO];
        [self queueCommand: [[[SFTPLsCommand alloc] initWithPath: [self path]] autorelease]];
    }
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context
{
    if ([keyPath isEqualToString: @"outcome"])
    {
        GDSTaskActionOutcome outcome  = [object outcome];

        [self stopObserving];

//        NSLog(@"Got outcome: %d", outcome);

        NSString* prompt   = [[object results] valueForKey: kPrompt];
        NSArray*  response = [[object results] valueForKey: kResponse];
        NSString* raw      = [[object results] valueForKey: GDSTaskActionResultRawOutputKey];
        
//        NSLog(@"Response: %@  Prompt: \"%@\"", response, prompt);

        // Stop observing monitor.  Note: may cause monitor to be released!
//        [object removeObserver: self forKeyPath: keyPath];

        // XXX - currently not being used - see GDSTaskController.
        if (outcome == GDSTaskActionExited)
        {
//            NSLog(@"Disconnected");
    
            [self setError: [response lastObject]];

            switch ([self state])
            {
                case SFTPClientTaskStateConnecting:
                    [[self delegate] sftpClientConnectionFailed: self];
                    break;

                default:
                    [[self delegate] sftpClientDisconnected: self];
            }
        }
        else  if (outcome == GDSTaskActionFailure)
        {
            [self setError: [[object results] valueForKey: GDSTaskActionResultErrorStringKey]];
            
#if 0
            switch ([self state])
            {
                // TBD - what to do here.
            }
#endif
        }
        else  // Success
        {
            switch ([self state])
            {
                case SFTPClientTaskStateConnecting:
                    if ([raw containsUnknownHostKeyPrompt])
                    {
                        /*
                         Response:
                           "Connecting to www.radiofreeomaha.net...", 
                           ...
                           "The authenticity of host 'www.radiofreeomaha.net (24.252.22.227)' can't be established.", 
                           "DSA key fingerprint is 2d:5a:95:a6:6c:fa:a0:fc:7b:a4:4a:a7:37:d7:a2:d2."
                         Prompt: "Are you sure you want to continue connecting (yes/no)? "
                         */                         
                        NSString* keyInfo = [[response subarrayWithRange: NSMakeRange(1, [response count] - 1)] componentsJoinedByString: @"\n"];  // XXX - does this work with all supported clients?
                        [[self delegate] sftpClient: self atUnknownHostKeyPrompt: keyInfo];
                    }
                    else if ([prompt containsSftpPasswordPrompt])
                    {
//XXX - only on retry..                        [self setError: @"Invalid Password"];
                        [[self delegate] sftpClient: self atPasswordPrompt: prompt];
                    }
                    else if ([prompt containsSftpPrompt])
                    {
                        [self queueCommand: [[[SFTPPwdCommand alloc] init] autorelease]];
                        // Note:  order of setState: with respect to the above
                        //        is important so that queued commands are not
                        //        started in the current run loop cycle.
                        [self setState: SFTPClientTaskStateReady];
                        [[self delegate] sftpClientConnected: self ];
                    }
                    else
                    {
                        if ([[response lastObject] containsConnectionClosed])
                        {
                            /*
                             Response:
                               "Connecting to aneuho..."
                               "ssh: aneuho: No address associated with nodename"
                               "Connection closed"
                             */
                            [self setError: [response objectAtIndex: 1]];
                        }
                        else if (response != nil)
                        {
                            [self setError: [response lastObject]];
                        }
                        else
                        {
                            [self setError: [[object results] valueForKey: GDSTaskActionResultRawOutputKey]];
                        }
                        [[self delegate] sftpClientConnectionFailed: self];
                        [self disconnect];
                    }
                    break;

                case SFTPClientTaskStateBusy:
                    [[self currentCommand] parseResponse: response];
                    // Note:  order of setState: with respect to the above
                    //        is important so that queued commands are not
                    //        started in the current run loop cycle.
                    [self setState: SFTPClientTaskStateReady];

                    // Notify delegate of pause due to error.
                    if ([self error] != nil && [[self delegate] respondsToSelector: @selector(sftpClient:pausedDueToError:)])
                    {
                        [[self delegate] sftpClient: self pausedDueToError: [self error]];
                    }
                    break;
            }
        }

        if ([[self commandQueue] count] == 1 && [self needsRefresh])
        {
            [self refresh];
        }
        else
        {
            [self startNextCommand];
        }
    }
}

- (void)taskController:(GDSTaskController*)controller updateActionMonitor:(GDSTaskActionMonitor*)monitor
{
    NSMutableString* response = [[monitor results] valueForKey: GDSTaskActionResultRawOutputKey];

    // Pass along raw read data to delegate, if desired
    if ( [[self delegate] respondsToSelector: @selector(sftpClient:readRawData:)] )
    {
        [[self delegate] sftpClient: self readRawData: [[monitor results] valueForKey: GDSTaskActionResultLastReadKey]];
    }

    // Convert \r\n sequence to \n
    [response replaceOccurrencesOfString: @"\r\n"
                              withString: @"\n"
                                 options: 0
                                   range: NSMakeRange(0, [response length])];

//    NSLog(@"Updating action monitor for response:\n\"%@\"", response);

    // can happen @ shutdown:  NSAssert([self state] != SFTPClientTaskStateNotConnected, @"Inconsistent state");
    NSAssert([self state] != SFTPClientTaskStateReady, @"Inconsistent state");

    BOOL  atPrompt = NO;

    switch ([self state])
    {
        case SFTPClientTaskStateConnecting:
            atPrompt =  [response containsUnknownHostKeyPrompt] ||
                        [response containsSftpPasswordPrompt] ||
                        [response containsSftpPrompt] ||
                        [response containsConnectionClosed];
            break;

        case SFTPClientTaskStateBusy:
#ifdef SSH_COM_SUPPORT
            /* SSH.com's sftp makes you hit a key to get to the prompt. Whee. */
            if ( [response containsString: @"<Press any key"] )
            {
                XXX - this may not be correct... -TKM
                NSArray* lines = [response componentsSeparatedByString: @"\n"];
                // Save partial response
                mPartialResponse = [[lines subarrayWithRange: NSMakeRange(0, [lines count] - 1)] retain];

                [controller sendResponse: @""];
                return;
            }
            else
#endif
            {
                [[self currentCommand] updateProgress: response];
                atPrompt = [response containsSftpPrompt];
            }
            break;
    }

    if (atPrompt)
    {
//        NSLog(@"At prompt.");
        NSArray* lines = [response componentsSeparatedByString:@"\n"];

#ifdef SSH_COM_SUPPORT
        if (mPartialResponse != nil)
        {
            lines = [mPartialResponse arrayByAddingObjectsFromArray: lines];
            [mPartialResponse release];
            mPartialResponse = nil;
        }
#endif

        // Special case for SSH v1 using challenge/response with Password: prompt
        if ([response hasSuffix: @"Password:\nResponse: "] ||
            [response hasSuffix: @"Password: \nResponse: "])
        {
            // Strip off Response: prompt in this case.
            lines = [lines subarrayWithRange: NSMakeRange(0, [lines count] - 1)];
        }

        // Remove prompt from response
        [[monitor results] setValue: [lines lastObject]
                             forKey: kPrompt];
        [[monitor results] setValue: [lines subarrayWithRange: NSMakeRange(0, [lines count] - 1)]
                             forKey: kResponse];

        [monitor setOutcome: GDSTaskActionSuccess];
    }        
}

@end
