//
//  SFTPClientTaskControllerTests.m
//  Fugu
//
//  Created by Tim McIntosh on 2005.10.07.
//  Copyright 2005 __MyCompanyName__. All rights reserved.
//

#import "SFTPClientTaskControllerTests.h"
#import "SFTPClientTaskController.h"
#import "SFTPFileTransferCommand.h"
#import "SFTPCdCommand.h"
#import "SFTPRenameCommand.h"
#import "SFTPRmCommand.h"

@implementation SFTPClientTaskControllerTests

- (void)setError:(NSString*)error
{
    if (error != mError)
    {
        [mError release];
        mError = [error copy];
    }
}

- (NSString*)error
{
    return mError;
}

- (void)sftpClient:(SFTPClientTaskController*)client atUnknownHostKeyPrompt:(NSString*)prompt
{
    [client sendUnknownHostKeyResponse:YES];
}

- (void)sftpClient:(SFTPClientTaskController*)client atPasswordPrompt:(NSString*)prompt
{
    [client sendPassword: @"weare138"];
}

- (void)sftpClientConnectionFailed:(SFTPClientTaskController*)client
{
    NSLog(@"Connection failed.");
    UKFail();
}

- (void)sftpClient:(SFTPClientTaskController*)client pausedDueToError:(NSString*)error
{
    NSLog(@"Command error: %@, resuming.", error);

    [self setError: error];

    [client resume];
}

- (void)trackProgress:(SFTPFileTransferCommand*)cmd
{
    [cmd addObserver: self
          forKeyPath: @"progress"
             options: 0
             context: NULL];

    [cmd addObserver: self
          forKeyPath: @"isComplete"
             options: 0
             context: NULL];
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context
{
    if ([keyPath isEqualToString: @"progress"])
    {
        NSLog(@"Transfer progress: %d", (unsigned int)[object progress]);
    }
    else if ([keyPath isEqualToString: @"isComplete"])
    {
        NSLog(@"Command complete: %@", [object command]);
        
        [object removeObserver: self forKeyPath: @"progress"];
        [object removeObserver: self forKeyPath: @"isComplete"];
    }
    else
    {
        NSAssert(0, @"Unknown key path!");
    }
}

// These should exist:
#define LOCAL_TEST_FILE @"/Network/Users/tmcintos/starwarsWhosOnFirst.wmv"
#define REMOTE_TEST_DIR @"/usr/home/mcintosh"

- (void)test
{
    NSLog(@"Testing SFTP...");

    SFTPClientTaskController* sftp = [[SFTPClientTaskController alloc] initWithServer: @"www.radiofreeomaha.net"
                                                                                 port: [NSNumber numberWithUnsignedShort: 2222]
                                                                                 user: @"mcintosh"
                                                                              options: [NSArray array]];

    [sftp setDelegate: self];

    // Test connect
    NSLog(@"Connecting...");
    [sftp connect];
    [[NSRunLoop currentRunLoop] runUntilDate: [NSDate dateWithTimeIntervalSinceNow: 10.0]];
    // Expect successful connection, in ready state
    UKIntsEqual(SFTPClientTaskStateReady, [sftp state]);

    SFTPCommand*  cmd;

    // test 'cd' failure
    [sftp queueCommand: cmd = [[[SFTPCdCommand alloc] initWithRemotePath: @"/nonexistent"] autorelease]];
    UKObjectsSame(cmd, [sftp currentCommand]);
    [[NSRunLoop currentRunLoop] runUntilDate: [NSDate dateWithTimeIntervalSinceNow: 1.0]];
    // Expect in ready state, error logged
    UKIntsEqual(SFTPClientTaskStateReady, [sftp state]);
    UKNotNil([self error]);
    [self setError: nil];

    // test 'cd' success
    [sftp queueCommand: cmd = [[[SFTPCdCommand alloc] initWithRemotePath: REMOTE_TEST_DIR] autorelease]];
    UKObjectsSame(cmd, [sftp currentCommand]);
    [[NSRunLoop currentRunLoop] runUntilDate: [NSDate dateWithTimeIntervalSinceNow: 1.0]];
    // Expect in ready state, no error logged
    UKIntsEqual(SFTPClientTaskStateReady, [sftp state]);
    UKNil([self error]);
    [self setError: nil];

    // test 'put' failure case 1 - nonexistent local file
    [sftp queueCommand: cmd = [[[SFTPFileTransferCommand alloc] initWithLocalPath: @"/nonexistent"
                                                                       remotePath: REMOTE_TEST_DIR @"/test_file"
                                                                transferDirection: SFTPFileTransferDirectionUpload] autorelease]];
    UKObjectsSame(cmd, [sftp currentCommand]);
    [[NSRunLoop currentRunLoop] runUntilDate: [NSDate dateWithTimeIntervalSinceNow: 1.0]];
    // Expect in ready state, error logged
    UKIntsEqual(SFTPClientTaskStateReady, [sftp state]);
    UKNotNil([self error]);
    [self setError: nil];


    // test 'put' failure case 2 - nonexistent remote directory
    [sftp queueCommand: cmd = [[[SFTPFileTransferCommand alloc] initWithLocalPath: LOCAL_TEST_FILE
                                                                       remotePath: @"/nonexistent/nonexistent"
                                                                transferDirection: SFTPFileTransferDirectionUpload] autorelease]];
    UKObjectsSame(cmd, [sftp currentCommand]);
    [[NSRunLoop currentRunLoop] runUntilDate: [NSDate dateWithTimeIntervalSinceNow: 1.0]];
    // Expect in ready state, error logged
    UKIntsEqual(SFTPClientTaskStateReady, [sftp state]);
    UKNotNil([self error]);
    [self setError: nil];

    // test 'put' success
    [sftp queueCommand: cmd = [[[SFTPFileTransferCommand alloc] initWithLocalPath: LOCAL_TEST_FILE
                                                                       remotePath: REMOTE_TEST_DIR @"/test_file"
                                                                transferDirection: SFTPFileTransferDirectionUpload] autorelease]];
    UKObjectsSame(cmd, [sftp currentCommand]);
    [self trackProgress: (SFTPFileTransferCommand*) cmd];
    [[NSRunLoop currentRunLoop] runUntilDate: [NSDate dateWithTimeIntervalSinceNow: 70.0]];
    // Expect in ready state, no error logged
    UKIntsEqual(SFTPClientTaskStateReady, [sftp state]);
    UKNil([self error]);
    [self setError: nil];

    // test 'get' success
    [sftp queueCommand: cmd = [[[SFTPFileTransferCommand alloc] initWithLocalPath: @"test_file"
                                                                       remotePath: REMOTE_TEST_DIR @"/test_file"
                                                                transferDirection: SFTPFileTransferDirectionDownload] autorelease]];
    UKObjectsSame(cmd, [sftp currentCommand]);
    [self trackProgress: (SFTPFileTransferCommand*) cmd];
    [[NSRunLoop currentRunLoop] runUntilDate: [NSDate dateWithTimeIntervalSinceNow: 70.0]];
    // Expect in ready state, no error logged
    UKIntsEqual(SFTPClientTaskStateReady, [sftp state]);
    UKNil([self error]);
    [self setError: nil];

    // test 'rename' success
    [sftp queueCommand: cmd = [[[SFTPRenameCommand alloc] initWithOrigPath: REMOTE_TEST_DIR @"/test_file"
                                                                   newPath: REMOTE_TEST_DIR @"/test_file2" ] autorelease]];
    UKObjectsSame(cmd, [sftp currentCommand]);
    [[NSRunLoop currentRunLoop] runUntilDate: [NSDate dateWithTimeIntervalSinceNow: 1.0]];
    // Expect in ready state, no error logged
    UKIntsEqual(SFTPClientTaskStateReady, [sftp state]);
    UKNil([self error]);
    [self setError: nil];

    // test 'rename' success for a file
    [sftp queueCommand: cmd = [[[SFTPRmCommand alloc] initWithRemotePath: REMOTE_TEST_DIR @"/test_file2"
                                                             isDirectory: NO  ] autorelease]];
    UKObjectsSame(cmd, [sftp currentCommand]);
    [[NSRunLoop currentRunLoop] runUntilDate: [NSDate dateWithTimeIntervalSinceNow: 1.0]];
    // Expect in ready state, no error logged
    UKIntsEqual(SFTPClientTaskStateReady, [sftp state]);
    UKNil([self error]);
    [self setError: nil];

    // FIXME - need mkdir command
}

@end
