/*
 * Copyright (c) 2003 Regents of The University of Michigan.
 * All Rights Reserved.  See COPYRIGHT.
 */

#import "SFTPController.h"

#import "SFTPPrefs.h"
#import "SFTPNode.h"
#import "SFTPErrorHandler.h"
#import "SFTPItemCell.h"

#import "SFTPAppController.h"
#import "SCPController.h"

#import "NDLaunchServicesObject.h"

#import "SFTPClientTaskController.h"
#import "SFTPFileTransferCommand.h"
#import "SFTPRenameCommand.h"
#import "SFTPRmCommand.h"
#import "SFTPMkdirCommand.h"
#import "SFTPCustomCommand.h"

#import "NSArray(Extensions).h"
#import "NSAttributedString-Ellipsis.h"
#import "NSCalendarDate(ConvertToSeconds).h"
#import "NSFileManager(mktemp).h"
#import "NSImage(IconForType).h"
#import "NSImage+SystemIcons.h"
#import "NSMenu(Extensions).h"
#import "NSMutableArray(Extensions).h"
#import "NSSet(ImageExtensions).h"
#import "NSString(SSHAdditions).h"
#import "NSString(FSRefAdditions).h"
#import "NSString-UnknownEncoding.h"
#import "NSString(URLAdditions).h"
#import "NSString+SFTPAdditions.h"
#import "NSWorkspace(LaunchServices).h"
#import "NSWorkspace(SystemVersionNumber).h"

#include <ApplicationServices/ApplicationServices.h>
#include <Carbon/Carbon.h>
#include <CoreFoundation/CoreFoundation.h>

#if 1  // XXX - thin these out - tkm
#include <sys/types.h>
#include <sys/file.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <pwd.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <util.h>
#endif

#include "keychain.h"

#import "MainMenuTags.h"

#define LOCAL_MENU_SHOW_INSPECTOR   800
#define LOCAL_MENU_OPEN             801
#define LOCAL_MENU_OPEN_WITH        820
#define LOCAL_MENU_PREVIEW          802
#define LOCAL_MENU_RENAME           803
#define LOCAL_MENU_UPLOAD           804
#define LOCAL_MENU_UPLOAD_WITH_SCP  805
#define LOCAL_MENU_MOVE_TO_TRASH    806

#define REMOTE_MENU_SHOW_INSPECTOR  900
#define REMOTE_MENU_OPEN            901
#define REMOTE_MENU_OPEN_WITH       920
#define REMOTE_MENU_PREVIEW         902
#define REMOTE_MENU_RENAME          903
#define REMOTE_MENU_DOWNLOAD        904
#define REMOTE_MENU_DNLOAD_WITH_SCP 905
#define REMOTE_MENU_DELETE          906

static NSString * const FuguRemoteFilesPboardType = @"FuguRemoteFilesPboardType";

// KVO context keys
static NSString * const ClientConnectionStateObservation = @"ClientConnectionStateObservation";


int namecmp( id ob1, id ob2, void *context )
{
    unsigned		*sorttype;
    
    sorttype = ( unsigned * )context;
    
    return( [[ ob1 objectForKey: @"name" ]
                compare: [ ob2 objectForKey: @"name" ]
                options: *sorttype ] );
}

int datecmp( id ob1, id ob2, void *context )
{
    NSComparisonResult	result = NSOrderedSame;
    off_t		d1, d2;
    
    d1 = [ NSCalendarDate secondsFromFormattedDateString:
                            [ ob1 objectForKey: @"date" ]];
    d2 = [ NSCalendarDate secondsFromFormattedDateString:
                            [ ob2 objectForKey: @"date" ]];
    
    if ( d1 < d2 ) {
        result = NSOrderedAscending;
    } else if ( d2 < d1 ) {
        result = NSOrderedDescending;
    } else {
        result = namecmp( ob1, ob2, context );
    }
    
    return( result );
}

int sizecmp( id ob1, id ob2, void *context )
{
    NSComparisonResult	result = NSOrderedSame;
    NSString		*size1 = [ ob1 objectForKey: @"size" ];
    NSString		*size2 = [ ob2 objectForKey: @"size" ];
    long long int	s1, s2;
    
    s1 = strtoll(( char * )[ size1 UTF8String ], NULL, 10 );
    s2 = strtoll(( char * )[ size2 UTF8String ], NULL, 10 );
    
    if ( s1 < s2 ) {
        result = NSOrderedAscending;
    } else if ( s2 < s1 ) {
        result = NSOrderedDescending;
    } else {
        result = namecmp( ob1, ob2, context );
    }
    
    return( result );
}

int ownercmp( id ob1, id ob2, void *context )
{
    return( [[ ob1 objectForKey: @"owner" ]
                compare: [ ob2 objectForKey: @"owner" ]
                options: NSLiteralSearch ] );
}

int groupcmp( id ob1, id ob2, void *context )
{
    return( [[ ob1 objectForKey: @"group" ]
                compare: [ ob2 objectForKey: @"group" ]
                options: NSLiteralSearch ] );
}

int permcmp( id ob1, id ob2, void *context )
{
    return( [[ ob1 objectForKey: @"perm" ]
                compare: [ ob2 objectForKey: @"perm" ]
                options: NSLiteralSearch ] );
}

int ( *sortFunctionForIdentifier( id identifier ))( id, id, void * )
{
    if ( [ identifier isEqualToString: @"namecolumn" ] ) {
        return( &namecmp );
    } else if ( [ identifier isEqualToString: @"datecolumn" ] ) {
        return( &datecmp );
    } else if ( [ identifier isEqualToString: @"sizecolumn" ] ) {
        return( &sizecmp );
    } else if ( [ identifier isEqualToString: @"ownercolumn" ] ) {
        return( &ownercmp );
    } else if ( [ identifier isEqualToString: @"groupcolumn" ] ) {
        return( &groupcmp );
    } else if ( [ identifier isEqualToString: @"permcolumn" ] ) {
        return( &permcmp );
    }
    
    return( 0 );
}

@interface SFTPController(PrivateMethods)
- ( SFTPClientTaskController * )client;
- ( void )setClient: ( SFTPClientTaskController * )client;
@end

@implementation SFTPController

- ( id )init
{
    if ( self = [ super initWithWindowNibName: @"SFTPWindow" ] )
    {
        mToolbarItems = [[ NSDictionary alloc ] initWithContentsOfFile: [[ NSBundle mainBundle ] pathForResource: @"SFTPToolbarItems" ofType: @"plist" ]];
//        NSLog(@"Toolbaritems: %@", mToolbarItems);
        
        /* watch for important notifications */
        [[ NSNotificationCenter defaultCenter ] addObserver: self
                                                   selector: @selector( reloadDefaults )
                                                       name: SFTPPrefsChangedNotification
                                                     object: nil ];

        kqueue = [[UKKQueue alloc] init];
        [kqueue setDelegate: self];

        [ self window ];   // Load NIB
    }

    return self;
}

- ( void )dealloc
{
    [[ NSNotificationCenter defaultCenter ] removeObserver: self ];

    [ dirImage release ];
    [ fileImage release ];
    
    [ remoteDirContents release ];
    [ dotlessRDir release ];
    [ dotlessLDir release ];
    [ remoteDirPath release ];
    [ rSwitchArray release ];
    
    [ openWithMenu release ];

    [ kqueue release ];

    [ mToolbarItems release ];

    [ super dealloc ];
}

- ( SFTPClientTaskController *)client
{
    return mClient;
}

- ( void )setClient: ( SFTPClientTaskController *)c
{
    if ( c != mClient )
    {
        [ mClient removeObserver: self
                      forKeyPath: @"isConnected" ];

        [ mClient removeObserver: self
                      forKeyPath: @"currentCommand" ];
        
        [ mClient removeObserver: self
                      forKeyPath: @"currentCommand.progress" ];
        
        [ mClient removeObserver: self
                      forKeyPath: @"currentCommand.isComplete" ];
        
        [ mClient removeObserver: self
                      forKeyPath: @"files" ];
        
        [ mClient removeObserver: self
                      forKeyPath: @"path" ];
    
        [ mClient autorelease ];
        mClient = [ c retain ];

        [ mClient setDelegate: self ];

        [ mClient addObserver: self
                   forKeyPath: @"isConnected"
                      options: NSKeyValueObservingOptionNew
                      context: ClientConnectionStateObservation ];

        // XXX - change other observations over to use context key

        [ mClient addObserver: self
                   forKeyPath: @"currentCommand"
                      options: NSKeyValueObservingOptionNew
                      context: NULL ];
        
        [ mClient addObserver: self
                   forKeyPath: @"currentCommand.progress"
                      options: NSKeyValueObservingOptionNew
                      context: NULL ];
        
        [ mClient addObserver: self
                   forKeyPath: @"currentCommand.isComplete"
                      options: NSKeyValueObservingOptionNew
                      context: NULL ];
        
        [ mClient addObserver: self
                   forKeyPath: @"files"
                      options: NSKeyValueObservingOptionNew
                      context: NULL ];
        
        [ mClient addObserver: self
                   forKeyPath: @"path"
                      options: NSKeyValueObservingOptionNew
                      context: NULL ];
    }
}

- ( IBAction )openWithMenuItem: ( id )sender
{
    // Just a dummy method so that validateMenuItem is called.
}

- ( BOOL )validateMenuItem: ( NSMenuItem * )anItem
{
//    NSLog(@"validate: %@", anItem);
    
    switch ( [ anItem tag ] )
    {
        case VIEW_MENU_COLUMNS_FILENAME:
            [ anItem setState: [ localBrowser columnWithIdentifier: @"filename" ] != -1 ? NSOnState : NSOffState ];
            break;

        case VIEW_MENU_COLUMNS_DATE:
            [ anItem setState: [ localBrowser columnWithIdentifier: @"datecolumn" ] != -1 ? NSOnState : NSOffState ];
            break;

        case VIEW_MENU_COLUMNS_SIZE:
            [ anItem setState: [ localBrowser columnWithIdentifier: @"sizecolumn" ] != -1 ? NSOnState : NSOffState ];
            break;

        case VIEW_MENU_COLUMNS_OWNER:
            [ anItem setState: [ localBrowser columnWithIdentifier: @"ownercolumn" ] != -1 ? NSOnState : NSOffState ];
            break;

        case VIEW_MENU_COLUMNS_GROUP:
            [ anItem setState: [ localBrowser columnWithIdentifier: @"groupcolumn" ] != -1 ? NSOnState : NSOffState ];
            break;

        case  VIEW_MENU_COLUMNS_PERMS:
            [ anItem setState: [ localBrowser columnWithIdentifier: @"permcolumn" ] != -1 ? NSOnState : NSOffState ];
            break;

        case FILE_MENU_DISCONNECT:
        case VIEW_MENU_FOCUS_ON_REMOTE:
        case LOCAL_MENU_UPLOAD:
        case WINDOW_MENU_TOGGLE_CONSOLE:
            return [ [ self client ] isConnected ];

        case VIEW_MENU_TOGGLE_HIDDEN:
            if ( dotflag )
            {
                [ anItem setTitle: NSLocalizedString( @"Hide Special Files", @"Hide Special Files" ) ];
            }
            else
            {
                [ anItem setTitle: NSLocalizedString( @"Show Hidden Files", @"Show Hidden Files" ) ];
            }
            break;

        case FILE_MENU_ITEM_OPEN_WITH:
            [ [ anItem menu ] setSubmenu: [ self openWithMenu ] forItem: anItem ];
            /*FALLTHROUGH*/

        case FILE_MENU_OPEN:
            if ( [[ self window ] firstResponder ] == localBrowser )
            {
                return YES;
            }
            else if ( [[ self window ] firstResponder ] != remoteBrowser )
            {
                return NO;
            }
            /*FALLTHROUGH*/

        case REMOTE_MENU_OPEN:
	    if ( [[[ self remoteSelectedFile ] type ] isNotEqualTo: kSftpFileFileType ] )
            {
                return( NO );
            }
    }

    return( YES );
}

- ( NSDictionary* )localFileAtIndex: ( int )anIndex
{
    NSDictionary* selectedFile = nil;

    if ( anIndex > -1 )
    {
        selectedFile = [( dotflag ? localDirContents : dotlessLDir ) objectAtIndex: anIndex ];
    }

    return selectedFile;
}

- ( NSDictionary* )remoteFileAtIndex: ( int )anIndex
{
    NSDictionary* selectedFile = nil;

    if ( anIndex > -1 )
    {
        selectedFile = [( dotflag ? remoteDirContents : dotlessRDir ) objectAtIndex: anIndex ];
    }

    return selectedFile;
}

- ( NSDictionary* )localSelectedFile
{
    return [ self localFileAtIndex: [ localBrowser selectedRow ] ];
}

- ( NSArray* )localSelectedFiles
{
    NSMutableArray* files;

    NSEnumerator*   e = [ localBrowser selectedRowEnumerator ];
    NSNumber*       row;

    while ( ( row = [ e nextObject ] ) != nil )
    {
        [ files addObject: [ self localFileAtIndex: [ row intValue ] ] ];
    }

    return [ NSArray arrayWithArray: files ];
}

- ( NSArray* )remoteSelectedFiles
{
    NSMutableArray* files;
    
    NSEnumerator*   e = [ remoteBrowser selectedRowEnumerator ];
    NSNumber*       row;
    
    while ( ( row = [ e nextObject ] ) != nil )
    {
        [ files addObject: [ self remoteFileAtIndex: [ row intValue ] ] ];
    }
    
    return [ NSArray arrayWithArray: files ];
}

- ( NSDictionary* )remoteSelectedFile
{
    return [ self remoteFileAtIndex: [ remoteBrowser selectedRow ] ];
}

- ( IBAction )focusOnLocalPane: ( id )sender
{
    [[ self window ] makeFirstResponder: localBrowser ];
}

- ( IBAction )focusOnRemotePane: ( id )sender
{
    [[ self window ] makeFirstResponder: remoteBrowser ];
}

- ( void )toolbarSetup
{
    NSToolbar *sftptbar = [[[ NSToolbar alloc ] initWithIdentifier:  @"RXTranscript tbar" ] autorelease ];
    
    [ sftptbar setAllowsUserCustomization: YES ];
    [ sftptbar setAutosavesConfiguration: YES ];
    [ sftptbar setDisplayMode: NSToolbarDisplayModeIconAndLabel ];
    
    [ sftptbar setDelegate: self ];
    [ [ self window ] setToolbar: sftptbar ];
}

#pragma mark -- required toolbar delegate methods --

- ( NSToolbarItem * )toolbar: ( NSToolbar * )toolbar itemForItemIdentifier: ( NSString * )itemIdent willBeInsertedIntoToolbar: ( BOOL )flag
{
    NSDictionary*  info =  [ mToolbarItems objectForKey: itemIdent ];
    NSToolbarItem* item = nil;

    if ( info == nil )
    {
        NSLog(@"Unknown toolbar item ID: %@", itemIdent);
    }
    else
    {
        NSString* tooltipIdent = [ NSString stringWithFormat: @"%@ Tooltip", itemIdent ];

        item = [[[ NSToolbarItem alloc ] initWithItemIdentifier: itemIdent ] autorelease ];
        
        [ item setLabel: [ [ NSBundle mainBundle] localizedStringForKey: itemIdent
                                                                  value: itemIdent
                                                                  table:  @"SFTPToolbar"] ]; 
        [ item setPaletteLabel: [ item label ]];
        [ item setToolTip: [ [ NSBundle mainBundle] localizedStringForKey: tooltipIdent
                                                                    value: [ @"No translation: " stringByAppendingString: tooltipIdent ]
                                                                    table: @"SFTPToolbar"] ];
        [ item setTarget: self ];
        [ item setAction: NSSelectorFromString( [ info objectForKey: @"selector" ] ) ];
        
        SEL s = NSSelectorFromString( [ info objectForKey: @"imageSelector" ] );
        NSString* arg = [ info objectForKey: @"imageArgument" ];
        
        if ( s == @selector(imageForSystemIcon:) )
        {
            const char* cstr = [ arg cString ];
            ResType iconRes = (cstr[0] << 24) | (cstr[1] << 16) | (cstr[2] << 8) | cstr[3];
            
            [ item setImage: [ NSImage imageForSystemIcon: iconRes ] ];
        }
        else
        {
            [ item setImage: [(NSObject*)[NSImage class] performSelector: s withObject: arg ] ];
        }
    }
    
    return item;
}

- ( BOOL )validateToolbarItem: ( NSToolbarItem * )tItem
{
    NSDictionary* info = [ mToolbarItems objectForKey: [ tItem itemIdentifier ] ];

#if 0 // XXX fix - tkm
    if ( [[ tItem itemIdentifier ] isEqualToString: SFTPToolbarDeleteIdentifier ] )
    {
        if ( [[ [ self window ] firstResponder ] isEqual: remoteBrowser ] )
        {
            [ tItem setLabel:
                NSLocalizedStringFromTable( @"Delete", @"SFTPToolbar",
                                            @"Delete" ) ];
            [ tItem setToolTip:
                NSLocalizedStringFromTable( @"Delete Tooltip", @"SFTPToolbar",
                                            @"Delete selected item from server." ) ];
            [ tItem setImage: [ NSImage imageForSystemIcon: kToolbarDeleteIcon]];
            
            return [ [ self client ] isConnected ];
        }
        else
        {
            [ tItem setLabel:
                NSLocalizedStringFromTable( @"Trash", @"SFTPToolbar",
                                            @"Trash" ) ];
            [ tItem setToolTip:
                NSLocalizedStringFromTable(@"Trash Tooltip", @"SFTPToolbar",
                                           @"Move selected local items or directories to the trash." ) ];
            [ tItem setImage: [ NSImage imageForSystemIcon: kTrashIcon ]];
            
            return  [[ [ self window ] firstResponder ] isEqual: localBrowser ];
        }
    }
    else if ( [[ tItem itemIdentifier ] isEqualToString: SFTPToolbarRemoteItemPreviewIdentifier ] )
    {
        if ( ! [[ [ self window ] firstResponder ] isKindOfClass: [ SFTPTableView class ]] )
        {
            return( NO );
        }
    }
    else if ( [[ tItem itemIdentifier ] isEqualToString: SFTPToolbarEditDocumentIdentifier ] )
    {
        if ( [[ [ self window ] firstResponder ] isEqual: localBrowser ] )
        {
            return [[ self localSelectedFile ] type ] != kSftpDirectoryFileType;
        }
        else if ( [[ [ self window ] firstResponder ] isEqual: remoteBrowser ] )
        {
            return [[ self remoteSelectedFile ] type ] != kSftpDirectoryFileType;
        }
        else
        {
            return( NO );
        }
    }
    else
#endif
    if ( ![[ info objectForKey: @"whenDisconnected" ] boolValue ] )
    {
        return [[ self client ] isConnected ];
    }

    return YES ;
}

- ( NSArray * )toolbarDefaultItemIdentifiers: ( NSToolbar * )toolbar
{
    // XXX clean up -tkm
    return [ NSArray arrayWithObjects:
        @"Local Home",
        @"Local History",
        NSToolbarSeparatorItemIdentifier,
        NSToolbarFlexibleSpaceItemIdentifier,
        @"Go To",
        @"Reload",
        @"Info",
        @"Edit",
        @"New Folder",
        @"Delete",
        @"Disconnect",
        NSToolbarFlexibleSpaceItemIdentifier,
        NSToolbarSeparatorItemIdentifier,
        @"Remote Home",
        @"Remote History",
        nil ];
}

- ( NSArray * )toolbarAllowedItemIdentifiers: ( NSToolbar * )toolbar
{
    return [[ mToolbarItems allKeys ] arrayByAddingObjectsFromArray: [ NSArray arrayWithObjects: 
        NSToolbarSeparatorItemIdentifier,
        NSToolbarFlexibleSpaceItemIdentifier,
        nil ]];
}

#pragma mark -- end required toolbar delegate methods --

- ( void )awakeFromNib
{
    NSUserDefaults	*defaults;
    NSString		*hdir, *identifier;
    NSArray		*columnArray;
    SFTPItemCell	*cell;
    NSFont              *logFont = nil;
    int			i, count;

    openWithMenu = [ [ NSMenu alloc ] initWithTitle: NSLocalizedString(@"Open With", @"Open With Menu") ];
    [ openWithMenu setAutoenablesItems: NO ];
    
    /* get images for display */
    dirImage = [[[ NSWorkspace sharedWorkspace ] iconForFileType: @"'fldr'" ] retain ];
    [ dirImage setScalesWhenResized: YES ];
    [ dirImage setSize: NSMakeSize( 16.0, 16.0 ) ];
    fileImage = [[[ NSWorkspace sharedWorkspace ] iconForFileType: @"'doc '" ] retain ];
    [ fileImage setScalesWhenResized: YES ];
    [ fileImage setSize: NSMakeSize( 16.0, 16.0 ) ];
    linkImage = [[ NSImage imageNamed: @"symlink" ] retain ];
    
    [ localBox setContentView: localView ];
    [ localView setNeedsDisplay: YES ];
    [ remoteBox setContentView: loginView ];
    [ loginView setNeedsDisplay: YES ];
    [ logField setEditable: NO ];

    [ [ self window ] setFrameAutosaveName: [ [ self  window ] title ] ];
    
    [ imagePreviewPanel setFrameAutosaveName: @"ImagePreviewPanel" ];
    
    [ infoPanel setFrameAutosaveName: @"InfoPanel" ];
    
    [ logField setString: @"" ];
    if (( logFont = [ NSFont fontWithName: @"Courier" size: 9 ] ) == nil ) {
        logFont = [ NSFont systemFontOfSize: 9 ];
    }
    if ( logFont != nil ) {
        [ logField setFont: logFont ];
    }
    
    [ manualComField registerForDraggedTypes:
                [ NSArray arrayWithObject: NSFilenamesPboardType ]];
    
    [ localBrowser setAction: nil ];//@selector( localBrowserSingleClick: ) ];
    [ localBrowser setDoubleAction: @selector( localBrowserDoubleClick: ) ];
    [ localBrowser setAutosaveTableColumns: YES ];
    [ localBrowser setNextKeyView: remoteHost ];
    [ loginDirField setNextKeyView: localBrowser ];
    [ localBrowser setAllowsColumnSelection: NO ];
    
    [ remoteBrowser setAction: nil ];//@selector( remoteBrowserSingleClick: ) ];
    [ remoteBrowser setDoubleAction: @selector( remoteBrowserDoubleClick: ) ];
    [ remoteBrowser setAutosaveTableColumns: YES ];
    [ remoteBrowser setNextKeyView: localBrowser ];
    [ remoteBrowser setAllowsColumnSelection: NO ];
    
    [ remoteHost setStringValue: @"" ];
    [ remoteHost setAction: nil ];	/* don't intercept 'return' presses */
    
    [ popUpFavs removeAllItems ];

    remoteHistoryMenu = [[ NSMenu alloc ] init ];
    localHistoryMenu = [[ NSMenu alloc ] init ];
    
    /*	set up visible table columns */
    while ( [[ localBrowser tableColumns ] count ] > 1 ) {
        [ localBrowser removeTableColumn:
                [[ localBrowser tableColumns ] lastObject ]];
    }
    while ( [[ remoteBrowser tableColumns ] count ] > 1 ) {
        [ remoteBrowser removeTableColumn:
                [[ remoteBrowser tableColumns ] lastObject ]];
    }
    
    columnArray = [[ NSUserDefaults standardUserDefaults ]
                        objectForKey: @"VisibleColumns" ];
    
    if ( columnArray == nil ) {
        columnArray = [ NSArray arrayWithObjects:
                            [ NSDictionary dictionaryWithObjectsAndKeys:
                                    @"sizecolumn", @"identifier",
                    [ NSNumber numberWithFloat: SIZE_COLUMN_WIDTH ], @"width", nil ],
                            [ NSDictionary dictionaryWithObjectsAndKeys:
                                    @"datecolumn", @"identifier",
                    [ NSNumber numberWithFloat: DATE_COLUMN_WIDTH ], @"width", nil ], nil ];
        [[ NSUserDefaults standardUserDefaults ] setObject: columnArray
                                                forKey: @"VisibleColumns" ];
    }

    count = [ columnArray count ];
    
    for ( i = 0; i < count; i++ ) {
        id			o;
        NSString		*identifier = @"", *title = @"column";
        float			width = 0.0;
        
        o = [ columnArray objectAtIndex: i ];
        
        if ( [ o isKindOfClass: [ NSString class ]] ) {
            identifier = o;
        } else if ( [ o isKindOfClass: [ NSDictionary class ]] ) {
            identifier = [ o objectForKey: @"identifier" ];
        }

        title = ColumnTitleFromIdentifier( identifier );
        width = [[[ columnArray objectAtIndex: i ]
                        objectForKey: @"width" ] floatValue ];

        [ localBrowser addTableColumnWithIdentifier: identifier
                columnTitle: title width: width ];
        [ remoteBrowser addTableColumnWithIdentifier: identifier
                columnTitle: title width: width ];
    }
    [ localBrowser sizeLastColumnToFit ];
    [ remoteBrowser sizeLastColumnToFit ];
    
    identifier = [[ NSUserDefaults standardUserDefaults ]
                    objectForKey: @"RemoteBrowserSortingIdentifier" ];
    if ( identifier == nil ) {
        identifier = @"namecolumn";
        [[ NSUserDefaults standardUserDefaults ]
                    setObject: identifier
                    forKey: @"RemoteBrowserSortingIdentifier" ];
        [[ NSUserDefaults standardUserDefaults ]
                    setObject: [ NSNumber numberWithInt: 0 ]
                    forKey: @"RemoteBrowserSortDirection" ];
    }
    if ( identifier != nil ) {
        NSTableColumn		*tc = nil;
        NSArray			*a = [ remoteBrowser tableColumns ];
        int			sortdirection = 0;
        
        for ( i = 0; i < [ a count ]; i++ ) {
            if ( [[[ a objectAtIndex: i ] identifier ]
                            isEqualToString: identifier ] ) {
                tc = [ a objectAtIndex: i ];
                break;
            }
        }
        
        if ( tc != nil ) {
            NSImage		*image = nil;
            
            [ remoteBrowser setHighlightedTableColumn: tc ];
            sortdirection = [[[ NSUserDefaults standardUserDefaults ]
                                objectForKey: @"RemoteBrowserSortDirection" ]
                                intValue ];
            
            if ( sortdirection == 0 ) {
                image = [ NSImage imageNamed: @"NSAscendingSortIndicator" ];
            } else {
                image = [ NSImage imageNamed: @"NSDescendingSortIndicator" ];
            }
            [ remoteBrowser setIndicatorImage: image inTableColumn: tc ];
        }
    }
                    
    
    /* set type of cell for tables */
    cell = [[ SFTPItemCell alloc ] init ];
    [ cell setAction: nil ];
    [[[ localBrowser tableColumns ]
            objectAtIndex: 0 ]
            setDataCell: cell ];
    [[[ remoteBrowser tableColumns ]
            objectAtIndex: 0 ]
            setDataCell: cell ];
    [ cell release ];
                        
    [ localBrowser registerForDraggedTypes:
            [ NSArray arrayWithObjects: FuguRemoteFilesPboardType, nil ]];
    [ localBrowser setDataSource: self ];
    localDirPath = [[ NSString alloc ] init ];
    localDirContents = [[ NSMutableArray alloc ] init ];
    dotlessLDir = [[ NSMutableArray alloc ] init ];
    remoteDirContents = [[ NSMutableArray alloc ] init ];
    
    rSwitchArray = [[ NSArray alloc ] initWithObjects:
                    roReadSwitch, roWriteSwitch, roExecSwitch,
                    rgReadSwitch, rgWriteSwitch, rgExecSwitch,
                    raReadSwitch, raWriteSwitch, raExecSwitch, nil ];
                    
    [ remoteBrowser registerForDraggedTypes:
                [ NSArray arrayWithObject: NSFilenamesPboardType ]];
    [ remoteBrowser setDataSource: self ];
    [ remoteBrowser reloadData ];
    
    /* set self to info tab's delegate */
    [ infoTabView setDelegate: self ];
        
    /* get defaults from prefs */
    defaults = [ NSUserDefaults standardUserDefaults ];
    hdir = [ defaults objectForKey: @"defaultdir" ];
    if ( hdir == nil || ! [ hdir length ] ) hdir = NSHomeDirectory();
    [ self localBrowserReloadForPath: hdir ];
        
    [ remoteHost addItemsWithObjectValues: [ defaults objectForKey: @"RecentServers" ]];
    
    if ( [ defaults objectForKey: @"PostEditBehaviour" ] == nil ) {
        [ defaults setObject: [ NSNumber numberWithInt: 0 ]
                    forKey: @"PostEditBehaviour" ];
    }
    
    /* get favorites from prefs */
    [ self reloadDefaults ];
    
    [ self toolbarSetup ];

    // Rendezvous menu - need to reset image each time menu items are changed:

    [[ rendezvousPopUp itemAtIndex: 0] setImage: [ NSImage imageNamed: @"zeroconf" ] ];

    [[ NSNotificationCenter defaultCenter ] addObserver: self
                                               selector: @selector(rendezvousMenuUpdate:)
                                                   name: NSMenuDidAddItemNotification
                                                 object: [ rendezvousPopUp menu ] ];

    /* search for rendezvous-enabled ssh servers */
    [ [ NSApp delegate ] scanForSSHServers ];

    [ [ self window ] makeKeyAndOrderFront: nil ];
    [[ remoteHost window ] makeFirstResponder: remoteHost ];
}

- ( IBAction )toggleDots: ( id )sender
{
    dotflag = !dotflag;

    [ localBrowser reloadData ];

    if ( [ [ self client ] isConnected ] ) {
        [ remoteBrowser reloadData ];
    }
}

- ( IBAction )toggleGroupColumn: ( id )sender
{
    int			state = [ sender state ];
    
    if ( state == NSOnState ) {
        [ sender setState: NSOffState ];
        [ localBrowser removeTableColumn:
                [ localBrowser tableColumnWithIdentifier: @"groupcolumn" ]];
        [ remoteBrowser removeTableColumn:
                [ remoteBrowser tableColumnWithIdentifier: @"groupcolumn" ]];
    } else {
        NSString	*identifier = @"groupcolumn";
        NSString	*title = ColumnTitleFromIdentifier( identifier );
        
        [ localBrowser addTableColumnWithIdentifier: identifier
            columnTitle: title width: GROUP_COLUMN_WIDTH ];
        [ remoteBrowser addTableColumnWithIdentifier: identifier
            columnTitle: title width: GROUP_COLUMN_WIDTH ];
        [ localBrowser moveColumn: ( [ localBrowser numberOfColumns ] - 1 )
                        toColumn: 1 ];
        [ remoteBrowser moveColumn: ( [ remoteBrowser numberOfColumns ] - 1 )
                        toColumn: 1 ];
                                        
        [ sender setState: NSOnState ];
    }
}

- ( IBAction )toggleModDateColumn: ( id )sender;
{
    int			state = [ sender state ];
    
    if ( state == NSOnState ) {
        [ sender setState: NSOffState ];
        [ localBrowser removeTableColumn:
                [ localBrowser tableColumnWithIdentifier: @"datecolumn" ]];
        [ remoteBrowser removeTableColumn:
                [ remoteBrowser tableColumnWithIdentifier: @"datecolumn" ]];
    } else {
        NSString	*identifier = @"datecolumn";
        NSString	*title = ColumnTitleFromIdentifier( identifier );
        
        [ localBrowser addTableColumnWithIdentifier: identifier
            columnTitle: title width: DATE_COLUMN_WIDTH ];
        [ remoteBrowser addTableColumnWithIdentifier: identifier
            columnTitle: title width: DATE_COLUMN_WIDTH ];
        [ localBrowser moveColumn: ( [ localBrowser numberOfColumns ] - 1 )
                        toColumn: 1 ];
        [ remoteBrowser moveColumn: ( [ remoteBrowser numberOfColumns ] - 1 )
                        toColumn: 1 ];
                                        
        [ sender setState: NSOnState ];
    }
}

- ( IBAction )toggleModeColumn: ( id )sender
{
    int			state = [ sender state ];
    
    if ( state == NSOnState ) {
        [ sender setState: NSOffState ];
        [ localBrowser removeTableColumn:
                [ localBrowser tableColumnWithIdentifier: @"permcolumn" ]];
        [ remoteBrowser removeTableColumn:
                [ remoteBrowser tableColumnWithIdentifier: @"permcolumn" ]];
    } else {
        NSString	*identifier = @"permcolumn";
        NSString	*title = ColumnTitleFromIdentifier( identifier );
        
        [ localBrowser addTableColumnWithIdentifier: identifier
            columnTitle: title width: MODE_COLUMN_WIDTH ];
        [ remoteBrowser addTableColumnWithIdentifier: identifier
            columnTitle: title width: MODE_COLUMN_WIDTH ];
        [ localBrowser moveColumn: ( [ localBrowser numberOfColumns ] - 1 )
                        toColumn: 1 ];
        [ remoteBrowser moveColumn: ( [ remoteBrowser numberOfColumns ] - 1 )
                        toColumn: 1 ];
                                        
        [ sender setState: NSOnState ];
    }
}

- ( IBAction )toggleOwnerColumn: ( id )sender
{
    int			state = [ sender state ];
    
    if ( state == NSOnState ) {
        [ sender setState: NSOffState ];
        [ localBrowser removeTableColumn:
                [ localBrowser tableColumnWithIdentifier: @"ownercolumn" ]];
        [ remoteBrowser removeTableColumn:
                [ remoteBrowser tableColumnWithIdentifier: @"ownercolumn" ]];
    } else {
        NSString	*identifier = @"ownercolumn";
        NSString	*title = ColumnTitleFromIdentifier( identifier );
        
        [ localBrowser addTableColumnWithIdentifier: identifier
            columnTitle: title width: OWNER_COLUMN_WIDTH ];
        [ remoteBrowser addTableColumnWithIdentifier: identifier
            columnTitle: title width: OWNER_COLUMN_WIDTH ];
        [ localBrowser moveColumn: ( [ localBrowser numberOfColumns ] - 1 )
                        toColumn: 1 ];
        [ remoteBrowser moveColumn: ( [ remoteBrowser numberOfColumns ] - 1 )
                        toColumn: 1 ];
                                        
        [ sender setState: NSOnState ];
    }
}

- ( IBAction )toggleSizeColumn: ( id )sender
{
    int			state = [ sender state ];
    
    if ( state == NSOnState ) {
        [ sender setState: NSOffState ];
        [ localBrowser removeTableColumn:
                [ localBrowser tableColumnWithIdentifier: @"sizecolumn" ]];
        [ remoteBrowser removeTableColumn:
                [ remoteBrowser tableColumnWithIdentifier: @"sizecolumn" ]];
    } else {
        NSString	*identifier = @"sizecolumn";
        NSString	*title = ColumnTitleFromIdentifier( identifier );
        
        [ localBrowser addTableColumnWithIdentifier: identifier
            columnTitle: title width: SIZE_COLUMN_WIDTH ];
        [ remoteBrowser addTableColumnWithIdentifier: identifier
            columnTitle: title width: SIZE_COLUMN_WIDTH ];
        [ localBrowser moveColumn: ( [ localBrowser numberOfColumns ] - 1 )
                        toColumn: 1 ];
        [ remoteBrowser moveColumn: ( [ remoteBrowser numberOfColumns ] - 1 )
                        toColumn: 1 ];
                                        
        [ sender setState: NSOnState ];
    }
}

- ( IBAction )showEditedDocumentsViewer: ( id )sender
{
    [ editedDocumentsPanel makeKeyAndOrderFront: sender ];
}

- ( IBAction )showActivityViewer: ( id )sender
{
    [ activityPanel makeKeyAndOrderFront: sender ];
}

- ( IBAction )toggleConsole: ( id )sender
{
    [ logDrawer toggle: sender ];
}

- ( void )reloadDefaults
{
    int			i;
    id			dobj, fobj;
    struct passwd	*pw;
    NSArray		*favs;
    NSUserDefaults	*defaults = [ NSUserDefaults standardUserDefaults ];
    
    /* reload favorites from prefs */
    [ popUpFavs removeAllItems ];
    favs = [ defaults objectForKey: @"Favorites" ];
    [ popUpFavs addItemWithTitle: @"" ];
    NSImage* img = [ NSImage imageForSystemIcon: kToolbarFavoritesIcon ];
    [img setSize:NSMakeSize(16, 16)];
    [[ popUpFavs itemAtIndex: 0 ] setImage: img]; // imageNamed: @"favorites.png" ]];
    for ( i = 0; i < [ favs count ]; i++ ) {
        fobj = [ favs objectAtIndex: i ];
        if ( [ fobj isKindOfClass: [ NSString class ]] ) {
            if ( [ fobj isEqualToString: @"" ] ) continue;
            [ popUpFavs addItemWithTitle: fobj ];
        } else {
            NSString	*title = nil;
            if ( [ fobj objectForKey: @"nick" ] != nil &&
                    ! [[ fobj objectForKey: @"nick" ] isEqualToString: @"" ] ) {
                title = [ fobj objectForKey: @"nick" ];
            } else if ( [ fobj objectForKey: @"host" ] != nil &&
                    ! [[ fobj objectForKey: @"host" ] isEqualToString: @"" ] ) {
                title = [ fobj objectForKey: @"host" ];
            } else {
                continue;
            }
            [ popUpFavs addItemWithTitle: title ];
        }
        [[ popUpFavs lastItem ] setImage: [ NSImage imageNamed: @"favorites.png" ]];
    }
    
    [ popUpFavs setEnabled: [ popUpFavs numberOfItems ] > 1 ];

    /* don't change things if user's entered anything */
    if ( [[ remoteHost stringValue ] length ] ) return;
    
    if (( dobj = [ defaults objectForKey: @"defaultuser" ] ) != nil ) {
        [ userName setStringValue: dobj ];
    } else if (( pw = getpwuid( getuid())) != NULL ) {
        [ userName setStringValue: [ NSString stringWithUTF8String: pw->pw_name ]];
    }
    if (( dobj = [ defaults objectForKey: @"defaulthost" ] ) != nil ) {
        [ remoteHost setStringValue: dobj ];
    }
    if (( dobj = [ defaults objectForKey: @"defaultport" ] ) != nil ) {
        [ portField setStringValue: dobj ];
    }
    if (( dobj = [ defaults objectForKey: @"defaultrdir" ] ) != nil ) {
        [ loginDirField setStringValue: dobj ];
    }
}

- ( IBAction )refreshBrowsers: ( id )sender
{
    [ self localBrowserReloadForPath: localDirPath ];
    if ( ![ [ self client ] isConnected ] ) return;
    [ [ self client ] refresh ];
}

- ( void )loadRemoteBrowserWithItems: ( NSArray * )items
{
    int			sorting;
    int			selectedrow = [ remoteBrowser selectedRow ];
    BOOL		asciisort = [[ NSUserDefaults standardUserDefaults ]
                                            boolForKey: @"ASCIIOrderSorting" ];

    [ remoteDirContents removeAllObjects ];
    [ dotlessRDir removeAllObjects ];
    
    [ remoteDirContents setArray: items ];
    
    if ( asciisort ) {
        sorting = NSLiteralSearch;
    } else {
        sorting = NSCaseInsensitiveSearch;
    }
    
    [ remoteDirContents sortUsingFunction:
            sortFunctionForIdentifier( [[ remoteBrowser highlightedTableColumn ]
                                            identifier ] )
                        context: ( void * )&sorting ];
                        
    if ( [[[ NSUserDefaults standardUserDefaults ]
            objectForKey: @"RemoteBrowserSortDirection" ] intValue ] ) {
        [ remoteDirContents reverse ];
    }
    [ dotlessRDir addObjectsFromArray: [ remoteDirContents visibleItems ]];

    [ remoteBrowser reloadData ];
    
    [ remoteBrowser scrollRowToVisible: 0 ];

    if ( [ infoPanel isVisible ] ) {
	if ( selectedrow >= 0 ) {
	    [ remoteBrowser selectRow: selectedrow byExtendingSelection: NO ];
	    [ self getInfo: nil ];
	}
    }
}

- ( void )setRemotePathPopUp: ( NSString * )pwd
{
    int			i;
    NSMutableArray	*rPathComponents = [[ NSMutableArray alloc ] init ];
    NSArray		*tmp = [[ pwd componentsSeparatedByString: @"/" ] copy ];
    NSImage		*slashImage;
    NSMenuItem		*item;

    [ remoteDirPath release ];
    remoteDirPath = [ pwd copy ];
    
    if ( remoteHome == nil ) {
        char		*p;

//NSLog( @"setting home directory" );
        p = strchr(( char * )[ pwd UTF8String ], '/' );
        for ( i = 0; i < strlen( p ); i++ ) {
            if ( p[ i ] == '\r' ) p[ i ] = '\0';
        }
        if ( p != NULL ) remoteHome = [[ NSString stringWithUTF8String: p ] retain ];
    }
    
    for ( i = ([ tmp count ] - 1 ); i > 0; i-- ) {
        if ( [[ tmp objectAtIndex: i ] isEqualToString: @"" ] ) continue;
        [ rPathComponents addObject: [ tmp objectAtIndex: i ]];
    }
    [ tmp release ];
    
    [ rPathPopUp removeAllItems ];
    [ rPathComponents addObject: @"/" ];
    for ( i = 0; i < [ rPathComponents count ]; i++ ) {
        item = [[ NSMenuItem alloc ] initWithTitle: [ rPathComponents objectAtIndex: i ]
                                        action: NULL
                                        keyEquivalent: @"" ];
        [ item setImage: dirImage ];
        [[ rPathPopUp menu ] addItem: item ];
        [ item release ];
    }
    slashImage = [[ NSWorkspace sharedWorkspace ] iconForFile: @"/" ];
    [ slashImage setSize: NSMakeSize( 16, 16 ) ];
    [[ rPathPopUp lastItem ] setImage: slashImage ];
    [ rPathPopUp selectItemWithTitle: [ rPathComponents objectAtIndex: 0 ]];
    [ rPathComponents release ];
    
    /* setup remote history menu; don't add path if we're reloading */
    if ( [ remoteHistoryMenu numberOfItems ] == 0
            || ! [ remoteDirPath isEqualToString: [[ remoteHistoryMenu itemAtIndex: 0 ] title ]] ) {
        NSImage			*img = dirImage;
        
        if ( [ remoteDirPath isEqualToString: @"/" ] ) {
            img = nil;
            img = slashImage;
        }
        [ remoteHistoryMenu insertItemWithTitle: remoteDirPath
                            action: @selector( cdFromRemoteHistoryMenu: )
                            keyEquivalent: @""
                            atIndex: 0 ];
        [[ remoteHistoryMenu itemAtIndex: 0 ] setImage: img ];
        while ( [ remoteHistoryMenu numberOfItems ] > 20 ) {
            [ remoteHistoryMenu removeItemAtIndex: ( [ remoteHistoryMenu numberOfItems ] - 1 ) ];
        }
    }
    
    /* close the preview panel, if open */
    if ( [ imagePreviewPanel isVisible ] ) {
        [ imagePreviewPanel close ];
    }
}

#pragma mark -- SFTPClientTaskController delegate methods --

- (void)sftpClientConnected:(SFTPClientTaskController*)c
{
    [ remoteBox setContentView: nil ];
    [ remoteBox setContentView: remoteView ];
    
    [ commandButton setEnabled: YES ];
    
    [[ remoteBrowser window ] makeFirstResponder: remoteBrowser ];
    [ localBrowser setNextKeyView: remoteBrowser ];

    // XXX - consider passing in initial directory to client at init time.
    // XXX - handle double-quotes in paths by switching to single quotes, as Fugu originally did here.
    if ( [[ loginDirField stringValue ] isNotEqualTo: @"" ] )
    {
        [ [ self client ] setPath: [ loginDirField stringValue ] ];
    }
}

- (void)sftpClient:(SFTPClientTaskController*)c readRawData:(NSString*)partialRead
{
    [ self addToLog: partialRead ];
}

- (void)sftpClientConnectionFailed:(SFTPClientTaskController*)c
{
    NSString* errMsg = [ [ self client ] error ];

    // Note:  errMsg may be nil if this gets called due to cancellation
    //        during initial connection.
    if ( errMsg != nil )
    {
        NSRunAlertPanel(NSLocalizedString(@"Connection to Server Failed", @"Connection Failed"),
                        [NSString stringWithFormat: NSLocalizedString(@"The following error was received when attempting to connect:\n\n%@\n\n"
                                                                      @"Please check your connection settings and try again.",
                                                                      @"Connection Failed Message"), errMsg],
                        @"OK", @"", @"" );
    }
}

- (void)sftpClient:(SFTPClientTaskController*)c pausedDueToError:(NSString*)error
{
    NSString* defaultButton = NSLocalizedString(@"OK", @"OK");
    NSString* altButton = nil;
    NSString* msg =
        [ NSString stringWithFormat:
            NSLocalizedString(@"The following error was received when attempting to %@: %@",
                              @"SFTP Command Error Message"),
            [[ c currentCommand ] displayName ],
            error ];

    if ( [[ c commandQueue ] count ] > 0 )
    {
        msg = [ msg stringByAppendingString:
            NSLocalizedString(@"\n\nThere are more commands waiting to be processed.\n"
                              @"Would you like to resume or abort the queued commands?",
                              @"SFTP Queued Commands Message") ];
        altButton     = NSLocalizedString(@"Abort", @"Abort");
        defaultButton = NSLocalizedString( @"Resume", @"Resume" );
    }

    int rc = NSRunAlertPanel( NSLocalizedString(@"Command Failed", @"SFTP Command Error Title"), 
                              msg,
                              defaultButton,
                              altButton,
                              nil );
    
    if ( rc != NSAlertDefaultReturn )
    {
        [ c flushQueue ];
    }
    [ c resume ];
}

#pragma mark -- End client task controller delegate methods --

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context
{
    id newValue = [ change valueForKey: NSKeyValueChangeNewKey ];
    id command  = [ [ self client ] currentCommand ];

//    NSLog(@"Observed value '%@' changed: %@", keyPath, change);

    if ( context == ClientConnectionStateObservation)
    {
        // XXX - may not be needed? was used to update window title
    }
    else if ([ keyPath isEqualToString: @"currentCommand" ])
    {
        if ( [ command isKindOfClass: [ SFTPFileTransferCommand class ] ] )
        {
            if ( [ command isUpload ] )
            {
                [ self showUploadProgress ];
                [ self updateUploadProgress ];
            }
            else
            {
                [ self showDownloadProgressWithMessage: [ command file ] ];
            }
        }   
    }
    else if ([ keyPath isEqualToString: @"currentCommand.isComplete" ])
    {
        if ( ! [ newValue isKindOfClass: [ NSNull class ] ] )
        {
            if ( [ newValue boolValue ] == YES )
            {
//                NSLog(@"command %@ completed", command);

                if ( [ command isKindOfClass: [ SFTPFileTransferCommand class ] ] )
                {
                    if ( [ command isUpload ] )
                    {
                        [ self updateUploadProgress ];
                    }
                    else
                    {
                        [ self finishedDownload: [ command path ] ];
                    }
                }
                else
                {
                    [ self finishedCommand ];  // XXX - clean up - tkm
                }
            }
        }
    }
    else if ( [ keyPath isEqualToString: @"currentCommand.progress" ] && [ [ [ self client ] currentCommand ] isKindOfClass: [ SFTPFileTransferCommand class ] ] )
    {
        if ( [ command isUpload ] )
        {
            [ self updateUploadProgressBarWithValue: [ command progress ]
                                   amountTransfered: [ command bytesTransferred ]
                                       transferRate: [ command transferRate ]
                                                ETA: [ command eta ] ];
        }
        else
        {
            [ self updateDownloadProgressBarWithValue: [ command progress ]
                                     amountTransfered: [ command bytesTransferred ]
                                         transferRate: [ command transferRate ]
                                                  ETA: [ command eta ] ];
        }
    }
    else if ( [ keyPath isEqualToString: @"path" ] )
    {
//        NSLog(@"Path change %@", change);

        [ self setRemotePathPopUp: newValue ];
    }
    else if ( [ keyPath isEqualToString: @"files" ] )
    {
//        NSLog(@"Files change.");

        [ self loadRemoteBrowserWithItems: newValue ];
    }
}

- ( void )finishedCommand
{
    /* enable buttons on completion */
    [ commandButton setEnabled: YES ];
    [ rPathPopUp setEnabled: YES ];
    [ lPathPopUp setEnabled: YES ];
}

- ( IBAction )sendManualCommand: ( id )sender
{
    NSString* command = [[ manualComField stringValue ] stringByTrimmingCharactersInSet: [ NSCharacterSet whitespaceCharacterSet ] ];
    NSString* lcommand = [ command lowercaseString ];

    if ( [ command isNotEqualTo: @"" ] &&
         ![ lcommand hasPrefix: @"quit" ] &&
         ![ lcommand hasPrefix: @"exit" ] )
    {
        [ [ self client ] queueCommand: [ [ [ SFTPCustomCommand alloc ] initWithCommand: command ] autorelease ] ];
    }

    [ manualComField selectText: nil ];
}

- (BOOL)windowShouldClose:(id)sender
{
    if ( sender == [ self window ] )
    {
        [ self disconnect: nil ];
        
        return [[ self client ] isConnected ] == NO;
    }

    return YES;
}

- ( IBAction )disconnect: ( id )sender
{
    if ( [[ self editedDocuments ] count ] )
    {
	int rc = NSRunAlertPanel([NSString stringWithFormat:
            NSLocalizedString( @"Disconnect %@?",
                               @"Open Document Disconnect Warning Title" ),
            [[ self client ] sessionName]],
                                 NSLocalizedString( @"Some remote documents may still be open.  If you disconnect, "
                                                    @"you will lose any unsaved changes.",
                                                    @"Open Document Disconnect Warning Text" ),
                                 NSLocalizedString( @"Disconnect", @"Disconnect" ),
                                 NSLocalizedString( @"Review Tracked Documents", @"Review Tracked Documents" ),
                                 NSLocalizedString( @"Cancel", @"Cancel" ),
                                 [[ self editedDocuments ] count ] );
        
	switch ( rc )
        {
            case NSAlertDefaultReturn:
                [ self willChangeValueForKey: @"editedDocuments" ];  // XXX - encapsulate
                [ editedDocuments removeAllObjects ];
                [ self didChangeValueForKey: @"editedDocuments" ];
                break;

            case NSAlertAlternateReturn:
                [ editedDocumentsPanel makeKeyAndOrderFront: nil ];
                /* FALLTHROUGH */

            default:
                return;
	}
    }

    if ( [ [ self client ] isConnected ] )
    {
        [ [ self client ] disconnect ];  // XXX old code just sent quit - TBD - TKM
    }

    if ( [ self cachedPreviews ] != nil )
    {
        [ cachedPreviews release ];
        cachedPreviews = nil;
    }

    [ logDrawer close ];
    [ localBrowser setNextKeyView: remoteHost ];
}

- (void)sftpClientDisconnected:(SFTPClientTaskController*)c
{
    if ( [ rPathPopUp numberOfItems ] )
    {
        [[ rPathPopUp itemAtIndex: 0 ] setImage: nil ];
    }
    [ rPathPopUp removeAllItems ];
    
    {
        int i;
        for ( i = [ remoteHistoryMenu numberOfItems ]; i > 0; i-- )
        {
            [ remoteHistoryMenu removeItemAtIndex: 0 ];
        }
    }
        
    /* disable buttons */
    [ commandButton setEnabled: NO ];
    
    [ remoteDirContents removeAllObjects ];
    [ dotlessRDir removeAllObjects ];
    [ remoteBrowser reloadData ];
    
    [ connectButton setEnabled: YES ];
    [[ localBrowser window ] makeFirstResponder: localBrowser ];
    [ remoteBox setContentView: nil ];
    [ remoteBox setContentView: loginView ];
    [[ remoteHost window ] makeFirstResponder: remoteHost ];
    [ passErrorField setStringValue: @"" ];
    
    remoteHome = nil;

    [ self setGotPasswordFromKeychain: NO ];
    [ addToKeychainSwitch setState: NSOffState ];
    if ( [ infoPanel isVisible ] ) [ infoPanel close ];

    [ self setClient: nil ];
}

- ( void )localBrowserReloadForPath: ( NSString * )path
{
    int			i, sorting, sortdirection = 0;
    NSArray		*bits;
    NSString		*fullpath, *componentpath, *identifier;
    NSMutableArray	*tmp, *lpath;
    NSImage		*img;
    NSMenuItem		*item;
    SFTPNode		*node;
    
    fullpath = [ path stringByExpandingTildeInPath ];

    if ( access(( char * )[ fullpath UTF8String ], F_OK | R_OK | X_OK ) < 0 ) {
        NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                NSLocalizedString( @"Couldn't view %@: %s",
                        @"Couldn't view %@: %s" ),
                NSLocalizedString( @"OK", @"OK" ), @"", @"",
                fullpath, strerror( errno ));
        return;
    }
    
    if ( localDirPath != nil ) {
        [ localDirPath release ];
        localDirPath = nil;
    }
    localDirPath = [ fullpath copy ];
    lpath = [[ NSMutableArray alloc ] init ];
    [ fullpath retain ];
    
    /* setup path popup button */
    bits = [ fullpath componentsSeparatedByString: @"/" ];
    
    for ( i = ( [ bits count ] - 1 ); i > 0; i-- ) {
        if ( ! [ ( NSString * )[ bits objectAtIndex: i ] length ] ) continue;
        [ lpath addObject: [ bits objectAtIndex: i ]];
    }
    
    [ lpath addObject: @"/" ];
    [ lPathPopUp removeAllItems ];
    
    componentpath = fullpath;
    
    for ( i = 0; i < [ lpath count ]; i++ ) {
        img = [[ NSWorkspace sharedWorkspace ] iconForFile: componentpath ];
                                    
        [ img setScalesWhenResized: YES ];
        [ img setSize: NSMakeSize( 16.0, 16.0 ) ];
        
        item = [[ NSMenuItem alloc ] initWithTitle: [ lpath objectAtIndex: i ]
                                        action: NULL
                                        keyEquivalent: @"" ];
        [ item setImage: img ];
        [[ lPathPopUp menu ] addItem: item ];
        [ item release ];
        img = nil;
        componentpath = [ componentpath stringByDeletingLastPathComponent ];
    }
    img = [[ NSWorkspace sharedWorkspace ] iconForFile: @"/" ];
    [ img setSize: NSMakeSize( 16, 16 ) ];
    [[ lPathPopUp lastItem ] setImage: img ];
    [ lPathPopUp selectItemWithTitle: [ lpath objectAtIndex: 0 ]];
    [ lpath release ];
    
    /* add path to history menu; don't add if we're reloading */
    if ( [ localHistoryMenu numberOfItems ] == 0
                || ! [ fullpath isEqualToString: [[ localHistoryMenu itemAtIndex: 0 ] title ]] ) {
        [ localHistoryMenu insertItemWithTitle: fullpath
                            action: @selector( cdFromLocalHistoryMenu: )
                            keyEquivalent: @""
                            atIndex: 0 ];
        img = nil;
        img = [[ NSWorkspace sharedWorkspace ] iconForFile: fullpath ];
        [ img setSize: NSMakeSize( 16, 16 ) ];
        [[ localHistoryMenu itemAtIndex: 0 ] setImage: img ];
        while ( [ localHistoryMenu numberOfItems ] > 20 ) {
            [ localHistoryMenu removeItemAtIndex: ( [ localHistoryMenu numberOfItems ] - 1 ) ];
        }
    }
    
    [ localDirContents release ];
    localDirContents = nil;
    
    [ dotlessLDir removeAllObjects ];
    
    if (( identifier = [[ NSUserDefaults standardUserDefaults ]
                objectForKey: @"LocalBrowserSortingIdentifier" ] ) == nil ) {
        identifier = @"namecolumn";
        [[ NSUserDefaults standardUserDefaults ]
                setObject: identifier
                forKey: @"LocalBrowserSortingIdentifier" ];
    }
    sortdirection = [[[ NSUserDefaults standardUserDefaults ]
                        objectForKey: @"LocalBrowserSortDirection" ]
                        intValue ];
    
    if ( ! [[[ localBrowser highlightedTableColumn ]
                identifier ] isEqualToString: identifier ] ) {
        NSArray			*a = [ localBrowser tableColumns ];
        
        for ( i = 0; i < [ a count ]; i++ ) {
            if ( [[[ a objectAtIndex: i ] identifier ]
                        isEqualToString: identifier ] ) {
                break;
            }
        }
        [ localBrowser setHighlightedTableColumn: [ a objectAtIndex: i ]];
        [ localBrowser setIndicatorImage:
                [ NSImage imageNamed: @"NSAscendingSortIndicator" ]
                        inTableColumn: [ a objectAtIndex: i ]];
    }
    
    sorting = ( int )[[ NSUserDefaults standardUserDefaults ]
                            boolForKey: @"ASCIIOrderSorting" ];
    
    if ( sorting ) {
        sorting = NSLiteralSearch;
    } else {
        sorting = NSCaseInsensitiveSearch;
    }
    
    node = [[ SFTPNode alloc ] init ];
    tmp = [ node itemsAtPath: fullpath showHiddenFiles: YES ];
    localDirContents = [ tmp mutableCopy ];
    [ localDirContents
            sortUsingFunction: sortFunctionForIdentifier( identifier )
            context: ( void * )&sorting ];
    if ( sortdirection ) {
        [ localDirContents reverse ];
        [ localBrowser setIndicatorImage:
                [ NSImage imageNamed: @"NSDescendingSortIndicator" ]
                inTableColumn: [ localBrowser highlightedTableColumn ]];
    }

    [ dotlessLDir addObjectsFromArray: [ localDirContents visibleItems ]];
    
    [ node release ];
    [ fullpath release ];

    [ localBrowser reloadData ];
}

- ( IBAction )localBrowserSingleClick: ( id )browser
{
    if ( [ infoPanel isVisible ] ) [ self getInfo: nil ];
}

- ( void )changeDirectory: ( id )sender
{
    if ( [[ [ self window ] firstResponder ] isEqual: localBrowser ] && [ localBrowser selectedRow ] >= 0 ) {
        [ self localBrowserDoubleClick: nil ];
    } else if ( [[ [ self window ] firstResponder ] isEqual: remoteBrowser ] && [ [ self client ] isConnected ] ) {
        [ self remoteBrowserDoubleClick: nil ];
    }
}

- ( void )dotdot: ( id )sender
{
    if ( [[ [ self window ] firstResponder ] isEqual: localBrowser ] ) {
        [ self localCdDotDot: nil ];
    } else if ( [[ [ self window ] firstResponder ] isEqual: remoteBrowser ] && [ [ self client ] isConnected ] ) {
        [ self remoteCdDotDot: nil ];
    }
}

- ( IBAction )localBrowserDoubleClick: ( id )sender
{
    NSDictionary*  file = [ self localFileAtIndex: [ sender clickedRow ] ];

    if ( file == nil )
    {
        NSLog(@"Clicked row is nil");  // XXX
        
        file = [ self localSelectedFile ];
    }

    if ( file == nil )
    {
        NSBeep();
        NSLog(@"%s no file selected!", __FUNCTION__);
    }
    else
    {
        NSString* path = [ file name ];

        if ( [[ file type ] isEqualToString: @"alias" ] )
        {
            /* resolve alias or link, determine if directory */
            path = [ file objectForKey: @"resolvedAlias" ];
        }

        BOOL isDirectory;
        BOOL fileExists = [[ NSFileManager defaultManager ] fileExistsAtPath: path isDirectory: &isDirectory ];

        if (  !fileExists )
        {
            NSBeep();
        }
        else if ( isDirectory )
        {
            [ self localBrowserReloadForPath: path ];
        }
        else
        {
            [self openLocalFileInEditor: nil];
        }
    }
}

- ( IBAction )uploadButtonClick: ( id )sender
{
    [ self uploadFiles: [[ self localSelectedFiles ] valueForKey: @"name" ] toDirectory: remoteDirPath ];
}

- ( IBAction )remoteBrowserSingleClick: ( id )browser
{
    if ( [ infoPanel isVisible ] ) [ self getInfo: nil ];
}

- ( IBAction )downloadButtonClick: ( id )sender
{
    NSEnumerator	*en = [ remoteBrowser selectedRowEnumerator ];
    NSMutableArray	*marray = [[ NSMutableArray alloc ] init ];
    NSAutoreleasePool	*p = [[ NSAutoreleasePool alloc ] init ];
    NSDictionary	*d;
    id			nobj;
    
    while (( nobj = [ en nextObject ] ) != nil ) {
        d = [ ( dotflag ? remoteDirContents : dotlessRDir ) objectAtIndex: [ nobj intValue ]];
        
        if ( [[ d objectForKey: @"type" ] isEqualToString: @"directory" ] ) {
            int			rc;

            rc = NSRunAlertPanel( NSLocalizedString( @"Warning: OpenSSH's sftp client "
                                                    @"cannot yet download directories.",
                                                    @"Warning: OpenSSH's sftp client "
                                                    @"cannot yet download directories." ),
                NSLocalizedString( @"Would you like to download %@ to %@ with SCP instead?",
                                @"Would you like to download %@ to %@ with SCP instead?" ),
                NSLocalizedString( @"Download", @"Download" ),
                NSLocalizedString( @"Cancel", @"Cancel" ), @"",
                [ d objectForKey: @"name" ], localDirPath );
            switch ( rc ) {
            case NSAlertDefaultReturn:
                [ self scpRemoteItem: [ d objectForKey: @"name" ]
                            fromHost: [ remoteHost stringValue ]
                         toLocalPath: localDirPath userName: [ userName stringValue ]];
            case NSAlertAlternateReturn:
            default:
                continue;
            }
        }
        [ marray addObject: d ];
    }
    [ p release ];
    [ self downloadFiles: marray toDirectory: localDirPath ];
    [ marray release ];
}


- ( IBAction )remoteBrowserDoubleClick: ( id )browser
{
    // Header cell double click will lead to clicked row = -1
    int row = [ remoteBrowser clickedRow ] > -1 ? [ remoteBrowser clickedRow ] : [ remoteBrowser selectedRow ];

    if (row > -1)
    {
        NSDictionary* remoteItem = [ dotflag ? remoteDirContents : dotlessRDir objectAtIndex: row ];

        unichar       type = [[ remoteItem objectForKey: @"perm" ] characterAtIndex: 0 ];
    
        if ( type == 'd' || type == 'l' )
        {
            [ [ self client ] setPath: [ remoteItem objectForKey: @"name" ] ];
        }
        else
        {
            /* File - double click edits in-place; option double click downloads */
            if ([[NSApp currentEvent] modifierFlags] & NSAlternateKeyMask)
            {
                [ self downloadFiles: [ NSArray arrayWithObject: remoteItem ]
                         toDirectory: localDirPath ];
            }
            else
            {
                [ self openRemoteFileInEditor: nil];
            }
        }
    }
}


- ( IBAction )showLocalHistoryMenu: ( id )sender
{
    NSEvent		*e = [ NSApp currentEvent ];
    
    if ( ! [ sender isKindOfClass: [ NSToolbarItem class ]] ) {
	return;
    }
			    
    [ NSMenu popUpContextMenu: localHistoryMenu withEvent: e forView: [ sender view ]];
}

- ( IBAction )showRemoteHistoryMenu: ( id )sender
{
    NSEvent*  e = [ NSApp currentEvent ];

    [ NSMenu popUpContextMenu: remoteHistoryMenu withEvent: e forView: nil ];
}

- ( void )uploadFile: ( NSString * )lpath toPath: ( NSString * )rpath
{
    [ [ self client ] queueCommand: [ [ [ SFTPFileTransferCommand alloc ] initWithLocalPath: lpath
                                                                                 remotePath: rpath
                                                                          transferDirection: SFTPFileTransferDirectionUpload ]
        autorelease ] ];
}

- ( void )uploadFiles: ( NSArray * )lfiles toDirectory: ( NSString * )rpath
{
    if ( [ [ self client ] isConnected ] )  // XXX may not be needed, see similar comment in downloads -tkm
    {
        BOOL noSafetyNet = [[ NSUserDefaults standardUserDefaults ] boolForKey: @"TrapezeWithNoNet" ];
        BOOL clobber = NO;

        NSEnumerator* localFilePathsEnumerator = [ lfiles objectEnumerator ];
        NSString*     localFilePath;
        
        while ( ( localFilePath = [ localFilePathsEnumerator nextObject ] ) != nil )
        {
            NSString* localFileName  = [ localFilePath lastPathComponent ];
            BOOL      skip = NO;
        
            if ( !noSafetyNet && !clobber )
            {
                NSEnumerator* remoteFilesEnumerator = [ remoteDirContents objectEnumerator ];
                NSDictionary* remoteFile;
                
                while ( ( remoteFile = [ remoteFilesEnumerator nextObject ] ) != nil )
                {
                    NSString* remoteFileName = [ remoteFile objectForKey: @"name" ];

                    if ( [ rpath isEqualToString: remoteDirPath ] && [ remoteFileName isEqualToString: localFileName ] )
                    {
                        int rc = NSRunAlertPanel( [ NSString stringWithFormat:
                                                        NSLocalizedString( @"%@ exists. Overwrite?", @"%@ exists. Overwrite?" ), localFileName],
                                                  @"", NSLocalizedString( @"Cancel", @"Cancel" ),
                                                  NSLocalizedString( @"Overwrite", @"Overwrite" ),
                                                  NSLocalizedString( @"Overwrite All", @"Overwrite All" ));

                        switch ( rc )
                        {
                            default:
                            case NSAlertDefaultReturn:
                                skip = YES;
                                return;

                            case NSAlertAlternateReturn:
                                break;
                                
                            case NSAlertOtherReturn:
                                clobber = YES;
                                break;
                        }
                    }
                }
            }

            if ( !skip || clobber )
            {
                struct stat st;

                if ( stat(( void * )[localFilePath UTF8String ], &st ) < 0 )
                {
                    NSLog( @"stat %@: %s", localFilePath, strerror( errno ));
                    continue;
                }
                
                NSString* remoteFilePath = [ NSString stringWithFormat: @"%@/%@", rpath, localFileName];

                if ( st.st_mode & S_IFDIR )
                {
                    [ self uploadDirectory: localFilePath toDirectory: remoteFilePath ];
                }
                else
                {
                    [ self uploadFile: localFilePath toPath: remoteFilePath ];
                }
            }
        }
    }
}

#pragma mark -- Download --

- ( void )downloadFiles: ( NSArray * )rpaths toDirectory: ( NSString * )lpath
{
    if ( [ [ self client ] isConnected ] )  // XXX - not sure if really needed if client checks it... -TKM
    {
        NSArray* lpathContents = [[ NSFileManager defaultManager ] directoryContentsAtPath: lpath ];

        BOOL noSafetyNet = [[ NSUserDefaults standardUserDefaults ] boolForKey: @"TrapezeWithNoNet" ];
        BOOL clobber = NO;

        NSEnumerator* remoteFilesEnumerator = [ rpaths objectEnumerator ];
        NSDictionary* remoteFile;
        
        while ( ( remoteFile = [ remoteFilesEnumerator nextObject ] ) != nil )
        {
            NSString* remoteFileName = [ remoteFile objectForKey: @"name" ];
            BOOL     skip = NO;

            if ( !noSafetyNet && !clobber )
            {
                NSEnumerator* localFilesEnumerator = [ lpathContents objectEnumerator ];
                NSString*     localFileName;
                
                while ( ( localFileName = [ localFilesEnumerator nextObject ] ) != nil )
                {
                    if ( [ localFileName isEqualToString: remoteFileName ] && !clobber )
                    {
                        int  rc =
                        NSRunAlertPanel( [ NSString stringWithFormat:
                            NSLocalizedString( @"%@ exists. Overwrite?", @"%@ exists. Overwrite?" ), remoteFileName],
                                         @"", NSLocalizedString( @"Cancel", @"Cancel" ),
                                         NSLocalizedString( @"Overwrite", @"Overwrite" ),
                                         NSLocalizedString( @"Overwrite All", @"Overwrite All" ));

                        switch ( rc )
                        {
                            case NSAlertDefaultReturn:
                                skip = YES;
                                break;
                                
                            case NSAlertAlternateReturn:
                                break;
                                
                            case NSAlertOtherReturn:
                                clobber = YES;
                                break;
                        }
                        
                        break;
                    }
                }
            }

            if ( !skip || clobber )
            {
                [ [ self client ] queueCommand: [ [ [ SFTPFileTransferCommand alloc ] initWithLocalPath: lpath
                                                                                    remotePath: remoteFileName
                                                                             transferDirection: SFTPFileTransferDirectionDownload ]
                    autorelease ] ];
            }
        }
    }
}

- ( void )showDownloadProgressWithMessage: ( NSString *)msg
{
    [ downloadTextField setStringValue:
        [ NSString stringWithFormat:
            NSLocalizedString( @"Downloading %@...", @"Downloading %@..." ), msg ] ];
    [ downloadProgBar setAnimationDelay: ( 20.0 / 60.0 ) ];
    [ downloadProgBar setUsesThreadedAnimation: YES ];
    [ downloadProgBar startAnimation: nil ];

    if ( timer == nil )
    {
        [ downloadTimeField setStringValue: @"00:00:00" ];
        dltime = 0;
        timer = [ NSTimer scheduledTimerWithTimeInterval: 5.0 target: self
                                                selector: @selector( dlUpdate ) userInfo: nil
                                                 repeats: YES ];
    }

    if ( ![ downloadSheet isVisible ] )
    {
        [ NSApp beginSheet: downloadSheet
            modalForWindow: [ self window ]
             modalDelegate: self
            didEndSelector: NULL
               contextInfo: nil ];
    }
}

- ( void )updateDownloadProgressBarWithValue: ( double )value
	    amountTransfered: ( NSString * )amount
	    transferRate: ( NSString * )rate
	    ETA: ( NSString * )eta
{
    double		pc_done = value;

    [ downloadProgBar stopAnimation: nil ];
    [ downloadProgBar setIndeterminate: NO ];
    [ downloadProgBar setMinValue: 0.0 ];
    [ downloadProgBar setMaxValue: 100.0 ];
    [ downloadProgBar setDoubleValue: pc_done ];
    [ downloadProgInfo setStringValue: [ NSString stringWithFormat: @"%@ %@ %@", amount, rate, eta ]];
    [ downloadProgPercentDone setStringValue: [ NSString stringWithFormat: @"%d%%", ( int )value ]];
}

- ( void )dlUpdate
{
    dltime += 5.0;
    [ downloadTimeField setStringValue: [ NSString clockStringFromInteger: dltime ]];
}

- ( void )finishedDownload: ( NSString * )remotePath
{
    // XXX - move to client, as well as similar code in upload - can be generalized (pass in class)
    NSEnumerator*  e = [ [ [ self client ] commandQueue ] objectEnumerator ];
    id             cmd;
    unsigned int  downloadCount = 0;

    while ( ( cmd = [ e nextObject ] ) != nil )
    {
        if ( [ cmd isKindOfClass: [ SFTPFileTransferCommand class ] ] && ![ cmd isComplete ] )
        {
            // XXX currently doesn't tell the difference between uploads & downloads, but GUI makes them mutually exclusive
            downloadCount++;
        }
    }

    if ( downloadCount == 0 ) 
    {
        // XXX factor this out - tkm
        [ downloadSheet orderOut: nil ];
        [ NSApp endSheet: downloadSheet ];
        [ downloadProgBar setIndeterminate: YES ];
        [ downloadProgBar stopAnimation: nil ];
    }

    if ( remotePath != nil )
    {
	NSDictionary*  editedDocument = [[ self editedDocuments ] lastObject ];

        if ( [ remotePath isEqualToString: [ editedDocument objectForKey: @"remotepath" ] ] )
        {
            [ self editFile: [ editedDocument objectForKey: @"localpath" ]
                 remotePath: [ editedDocument objectForKey: @"remotepath" ]
                 withEditor: [ editedDocument objectForKey: @"editor" ] ];
        }
    }

    if ( [ self previewedImage ] != nil )
    {
        [ self displayPreview ];
    }
    
    [ timer invalidate ];
    timer = nil;

    [ [ NSWorkspace sharedWorkspace ] noteFileSystemChanged ];

    [ self localBrowserReloadForPath: localDirPath ];
}

#pragma mark -- end download --

- ( void )delete: ( id )sender
{
    if ( [ sender isEqual: localBrowser ]
                || [[ [ self window ] firstResponder ] isEqual: localBrowser ] ) {
        [ self deleteLocalFile: nil ];
    } else if ( [ sender isEqual: remoteBrowser ]
                || [[ [ self window ] firstResponder ] isEqual: remoteBrowser ] ) {
        [ self deleteRemoteFile: nil ];
    }
}

- ( IBAction )deleteLocalFile: ( id )sender
{
    NSArray* dir   = dotflag ? localDirContents : dotlessLDir;
    NSArray* files = [ dir subarrayWithIndicesFromEnumerator: [ localBrowser selectedRowEnumerator ] ];

    [ self deleteLocalFiles: [ files valueForKeyPath: @"name.lastPathComponent" ] ];
}

- ( void )deleteLocalFiles: ( NSArray* )files
{
    NSString*			fn;
    int				rc, optag;

//    NSLog(@"Deleting: %@", files);

    switch ( [ files count ] ) {
        case 0:
            return;
        case 1:
            fn = [ files objectAtIndex: 0 ];
            break;
        default:
            fn = NSLocalizedString( @"the selected items", @"the selected items" );
            break;
    }

    rc = NSRunAlertPanel( [ NSString stringWithFormat:
                NSLocalizedString( @"Do you want to move %@ to the Trash?",
                        @"Do you want to move %@ to the Trash?" ), fn ],
            NSLocalizedString( @"You cannot undo this action.",
                        @"You cannot undo this action." ),
            NSLocalizedString( @"Delete", @"Delete" ),
            NSLocalizedString( @"Cancel", @"Cancel" ), @"" );

    if ( rc != NSAlertAlternateReturn) {
        if ( [[ NSWorkspace sharedWorkspace ] performFileOperation: NSWorkspaceRecycleOperation
                                                            source: localDirPath destination: @""
                                                             files: files tag: &optag ] == NO ) {
            NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                             NSLocalizedString( @"Couldn't delete %@.", @"Couldn't delete %@." ),
                             NSLocalizedString( @"OK", @"OK" ), @"", @"", fn );
            return;
        }

        [[ NSWorkspace sharedWorkspace ] noteFileSystemChanged: localDirPath ];

        [ self localBrowserReloadForPath: localDirPath ];
    }
}

- ( IBAction )deleteRemoteFile: ( id )sender
{
    NSArray* dir   = dotflag ? remoteDirContents : dotlessRDir;
    NSArray* files = [ dir subarrayWithIndicesFromEnumerator: [ remoteBrowser selectedRowEnumerator ] ];

    if ( ![ [ self client ] isConnected ] ) return;

    [ self deleteRemoteFiles: files ];
}

// Argument is a remote file property list
- ( void )deleteRemoteFiles: (NSArray*)files
{
    NSString*			fn;

//    NSLog(@"Deleting: %@", files);

    switch ( [ files count ] )
    {
    case 0:
        return;

    case 1:
        fn = [[ files objectAtIndex: 0 ] objectForKey: @"name" ];
        break;

    default:
        fn = NSLocalizedString( @"The selected items", @"The selected items" );
        break;
    }

    int rc = NSRunAlertPanel( [ NSString stringWithFormat: NSLocalizedString( @"Delete %@?", @"Delete %@?" ), fn ],
                              NSLocalizedString( @"%@ will be deleted immediately. "
                                                 @"You will not be able to undo this action.",
                                                 @"%@ will be deleted immediately. "
                                                 @"You will not be able to undo this action." ),
                              NSLocalizedString( @"Delete", @"Delete" ),
                              NSLocalizedString( @"Cancel", @"Cancel" ), @"", fn );

    if ( rc != NSAlertAlternateReturn )
    {
        NSEnumerator*   fileEnumerator = [ files objectEnumerator ];
        NSDictionary*   file;
        
        while ( (file = [ fileEnumerator nextObject ]) != nil )
        {
            [ [ self client ] queueCommand:
                [ [ [ SFTPRmCommand alloc ] initWithRemotePath: [ remoteDirPath stringByAppendingPathComponent: [file objectForKey: @"name" ] ]
                                                   isDirectory: [[ file objectForKey: @"perm" ] characterAtIndex: 0 ] == 'd' ]
                    autorelease ] ];
        }
    }
}

- ( IBAction )toggleDirCreationButtons: ( id )sender
{
    [ localDirCreateButton setState: NSOffState ];
    [ remoteDirCreateButton setState: NSOffState ];
    [ sender setState: NSOnState ];
    if ( [[ sender title ] isEqualToString:
            NSLocalizedString( @"Locally", @"Locally" ) ] ) {
        [ newDirCreateButton setAction: @selector( makeLDir ) ];
    } else {
        [ newDirCreateButton setAction: @selector( makeRDir ) ];
    }
}

- ( IBAction )createNewDirectory: ( id )sender
{
    id			activeTableView = nil;
    
    if ( [ [ self window ] isKeyWindow ] ) {
        activeTableView = [ [ self window ] firstResponder ];
        
        if ( [ activeTableView isEqual: localBrowser ] || ! [ [ self client ] isConnected ] ) {
            [ localDirCreateButton setState: NSOnState ];
            [ remoteDirCreateButton setState: NSOffState ];
        } else {
            [ remoteDirCreateButton setState: NSOnState ];
            [ localDirCreateButton setState: NSOffState ];
        }
    }
    
    if ( [ localDirCreateButton state ] ) {
        [ newDirCreateButton setAction: @selector( makeLDir ) ];
    } else if ( [ remoteDirCreateButton state ] ) {
        [ newDirCreateButton setAction: @selector( makeRDir ) ];
    }
    [ NSApp beginSheet: newDirPanel
            modalForWindow: [ self window ]
            modalDelegate: self
            didEndSelector: NULL
            contextInfo: nil ];
}

- ( void )makeRDir
{
    if ( [ [ self client ] isConnected ] )
    {
        [ [ self client ] queueCommand: [ [ [ SFTPMkdirCommand alloc ] initWithRemotePath: [ newDirNameField stringValue ] ] autorelease ] ];

        [ self dismissNewDirPanel: nil ];
    }
}

- ( void )makeLDir
{
    NSString		*lpath = [ NSString stringWithFormat:
                                    @"%@/%@",
                                    localDirPath,
                                    [ newDirNameField stringValue ]];
                                    
    if ( mkdir(( char * )[ lpath UTF8String ], 0755 ) != 0 ) {
        NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
            NSLocalizedString( @"Couldn't create %@: %s", @"Couldn't create %@: %s" ),
            @"OK", @"", @"", lpath, strerror( errno ));
    }
    
    [ self localBrowserReloadForPath: localDirPath ];
    [[ NSWorkspace sharedWorkspace ] noteFileSystemChanged: localDirPath ];
    
    [ self dismissNewDirPanel: nil ];
}

- ( IBAction )dismissNewDirPanel: ( id )sender
{
    [ newDirPanel orderOut: nil ];
    [ NSApp endSheet: newDirPanel ];
}

#if 0 // XXX - fix this -tkm
- ( IBAction )changeROwner: ( id )sender
{
    char	cmd[ MAXPATHLEN ];
    int		i;
    
    if ( ![[ rOwnerField stringValue ] length ] ) {
        goto CHOWN_ERROR;
    }
    for ( i = 0; i < [[ rOwnerField stringValue ] length ]; i++ ) {
        if ( !isdigit( [[ rOwnerField stringValue ] characterAtIndex: i ] )) {
            goto CHOWN_ERROR;
        }
    }
    
    if ( snprintf( cmd, MAXPATHLEN, "chown %s \"%s\"",
                ( char * )[[ rOwnerField stringValue ] UTF8String ],
                ( char * )[[ rWhereField stringValue ] UTF8String ] )
            > ( MAXPATHLEN - 1 )) {
        NSLog( @"Buffer overflow." );
        return;
    }
    
    [ self writeCommand: cmd ];
    return;

CHOWN_ERROR:
    NSBeginAlertSheet( NSLocalizedStringFromTable( @"Invalid owner user identification number.",
                        @"SFTPInfoPanel",
                        @"Invalid owner user identification number." ),
                          @"OK", @"", @"", infoPanel,
                          self, NULL, nil, NULL,
                        NSLocalizedStringFromTable(
                            @"You must provide a numeric uid for this action.", @"SFTPInfoPanel",
                            @"You must provide a numeric uid for this action." ));
}

- ( IBAction )changeRGroup: ( id )sender
{
    char	cmd[ MAXPATHLEN ];
    int		i;
    
    if ( ![[ rGroupField stringValue ] length ] ) {
        goto CHGRP_ERROR;
    }
    for ( i = 0; i < [[ rGroupField stringValue ] length ]; i++ ) {
        if ( !isdigit( [[ rGroupField stringValue ] characterAtIndex: i ] )) {
            goto CHGRP_ERROR;
        }
    }
    
    if ( snprintf( cmd, MAXPATHLEN, "chgrp %s \"%s\"",
                ( char * )[[ rGroupField stringValue ] UTF8String ],
                ( char * )[[ rWhereField stringValue ] UTF8String ] )
            > ( MAXPATHLEN - 1 )) {
        NSLog( @"Buffer overflow." );
        return;
    }
    
    [ self writeCommand: cmd ];
    return;

CHGRP_ERROR:
    NSBeginAlertSheet( NSLocalizedStringFromTable( @"Invalid group identification number.",
                        @"SFTPInfoPanel",
                        @"Invalid group identification number." ),
                          @"OK", @"", @"", infoPanel,
                          self, NULL, nil, NULL,
                        NSLocalizedStringFromTable(
                            @"You must provide a numeric gid for this operation.",
                            @"SFTPInfoPanel",
                            @"You must provide a numeric gid for this operation." ));
}
#endif

- ( IBAction )remotePermissionSwitchClick: ( id )sender
{
    int			own = 0, grp = 0, oth = 0;
    char		bit;
    
    if ( [[ rSwitchArray objectAtIndex: 0 ] state ] == NSOnState ) own += 4;
    if ( [[ rSwitchArray objectAtIndex: 1 ] state ] == NSOnState ) own += 2;
    if ( [[ rSwitchArray objectAtIndex: 2 ] state ] == NSOnState ) own += 1;
    if ( [[ rSwitchArray objectAtIndex: 3 ] state ] == NSOnState ) grp += 4;
    if ( [[ rSwitchArray objectAtIndex: 4 ] state ] == NSOnState ) grp += 2;
    if ( [[ rSwitchArray objectAtIndex: 5 ] state ] == NSOnState ) grp += 1;
    if ( [[ rSwitchArray objectAtIndex: 6 ] state ] == NSOnState ) oth += 4;
    if ( [[ rSwitchArray objectAtIndex: 7 ] state ] == NSOnState ) oth += 2;
    if ( [[ rSwitchArray objectAtIndex: 8 ] state ] == NSOnState ) oth += 1;
    
    if ( ![[ rPermField stringValue ] length ] ) bit = '0';
    else bit = [[ rPermField stringValue ] characterAtIndex: 0 ];
    
    if ( !isdigit( bit )) bit = '0';
    
    [ rPermField setStringValue: [ NSString stringWithFormat: @"%c%d%d%d",
            bit, own, grp, oth ]];
}

#if 0 // XXX - fix this -tkm
- ( IBAction )changeRemoteMode: ( id )sender
{
    int			i;
    char		cmd[ MAXPATHLEN ];
    
    if ( [[ rPermField stringValue ] length ] != 4 ) goto CHMOD_ERROR; 
    for ( i = 0; i < [[ rPermField stringValue ] length ]; i++ ) {
        if ( !isdigit( [[ rPermField stringValue ] characterAtIndex: i ] )) {
            goto CHMOD_ERROR;
        }
    }
    
    if ( snprintf( cmd, MAXPATHLEN, "chmod %s \"%s\"",
                        ( char * )[[ rPermField stringValue ] UTF8String ],
                        ( char * )[[ rWhereField stringValue ] UTF8String ] )
                > ( MAXPATHLEN - 1 )) {
        NSLog( @"buffer overflow" );
        return;
    }
    
    [ self writeCommand: cmd ];
    return;
    
CHMOD_ERROR:
    NSBeginAlertSheet( NSLocalizedStringFromTable( @"Invalid octal permissions set.",
                        @"SFTPInfoPanel", @"Invalid octal permissions set." ),
                          @"OK", @"", @"", infoPanel,
                          self, NULL, nil, NULL,
                        NSLocalizedStringFromTable(
                        @"You must provide a four-digit octal permissions set for this operation.",
                            @"SFTPInfoPanel",
                        @"You must provide a four-digit octal permissions set for this operation." ));
}
#endif

- ( IBAction )changeLOwnerAndGroup: ( id )sender
{
    uid_t	u;
    gid_t	g;
    
    if ( ![[ ownerField stringValue ] length ] ) return;
    if ( ![[ groupField stringValue ] length ] ) return;
    
    g = ( gid_t )[ groupField intValue ];
    u = ( uid_t )[ ownerField intValue ];
    
    /* only root can change owners */
    if ( getuid() != 0 && u != getuid()) u = getuid();
    
    if ( chown(( char * )[[ whereField stringValue ] UTF8String ], u, g ) < 0 ) {
        NSBeginAlertSheet( NSLocalizedStringFromTable( @"Error changing owner.", @"SFTPInfoPanel",
                            @"Error changing owner." ),
                          @"OK", @"", @"", infoPanel,
                          self, NULL, nil, NULL,
                          @"chown %@: %s", [ whereField stringValue ],
                            strerror( errno ));
        return;
    }
}

- ( IBAction )changeLocalMode: ( id )sender
{
    mode_t	mode;
    
    if ( ![[ permField stringValue ] length ] ) return;
    
    mode = ( mode_t )( strtol(( char * )[[ permField stringValue ] UTF8String ], NULL, 8 ));
    if ( errno == ERANGE || errno == EINVAL ) {
        NSLog( @"strtol failed: %s", strerror( errno ));
        return;
    }
    if ( chmod(( char * )[[ whereField stringValue ] UTF8String ], mode ) < 0 ) {
        NSBeginAlertSheet( NSLocalizedStringFromTable( @"Error changing permissions.",
                            @"SFTPInfoPanel", @"Error changing permissions." ),
                          @"OK", @"", @"", infoPanel,
                          self, NULL, nil, NULL,
                          @"chmod %@: %s", [ infoPathField stringValue ],
                            strerror( errno ));
        return;
    }
    [ self localBrowserReloadForPath: localDirPath ];
}

- ( IBAction )localPermissionSwitchClick: ( id )sender
{
    int			own = 0, grp = 0, oth = 0;
    char		bit;
    
    if ( [ loReadSwitch state ] == NSOnState ) own += 4;
    if ( [ loWriteSwitch state ] == NSOnState ) own += 2;
    if ( [ loExecSwitch state ] == NSOnState ) own += 1;

    if ( [ lgReadSwitch state ] == NSOnState ) grp += 4;
    if ( [ lgWriteSwitch state ] == NSOnState ) grp += 2;
    if ( [ lgExecSwitch state ] == NSOnState ) grp += 1;

    if ( [ laReadSwitch state ] == NSOnState ) oth += 4;
    if ( [ laWriteSwitch state ] == NSOnState ) oth += 2;
    if ( [ laExecSwitch state ] == NSOnState ) oth += 1;
    
    if ( ![[ permField stringValue ] length ] ) bit = '0';
    else bit = [[ permField stringValue ] characterAtIndex: 0 ];
    
    if ( !isdigit( bit )) bit = '0';
    
    [ permField setStringValue: [ NSString stringWithFormat: @"%c%d%d%d",
            bit, own, grp, oth ]];
}

- ( IBAction )toggleGoToButtons: ( id )sender
{
    [ localGotoButton setState: NSOffState ];
    [ remoteGotoButton setState: NSOffState ];
    [ sender setState: NSOnState ];
    if ( [[ sender title ] isEqualToString:
            NSLocalizedString( @"Locally",
                    @"Locally" ) ] ) {
        [ gotoButton setAction: @selector( gotoLocalDirectory: ) ];
    } else {
        [ gotoButton setAction: @selector( gotoRemoteDirectory: ) ];
    }
}

- ( IBAction )getGotoDirPanel: ( id )sender
{
    id			activeTableView = nil;
    
    if ( [ [ self window ] isKeyWindow ] ) {
        activeTableView = [ [ self window ] firstResponder ];
        
        if ( [ activeTableView isEqual: localBrowser ] || ! [ [ self client ] isConnected ] ) {
            [ remoteGotoButton setState: NSOffState ];
            [ localGotoButton setState: NSOnState ];
        } else {
            [ remoteGotoButton setState: NSOnState ];
            [ localGotoButton setState: NSOffState ];
        }
    }
    
    if ( [ localGotoButton state ] ) {
        [ gotoButton setAction: @selector( gotoLocalDirectory: ) ];
    } else if ( [ remoteGotoButton state ] ) {
        [ gotoButton setAction: @selector( gotoRemoteDirectory: ) ];
    }

    [ NSApp beginSheet: gotoDirPanel
            modalForWindow: [ self window ]
            modalDelegate: self
            didEndSelector: NULL
            contextInfo: nil ];
}

- ( IBAction )gotoLocalDirectory: ( id )sender
{
    [ self localBrowserReloadForPath:
            [[ gotoDirNameField stringValue ] stringByExpandingTildeInPath ]];
    [ self dismissGotoDirPanel: nil ];
}

- ( IBAction )gotoRemoteDirectory: ( id )sender
{
    if ( [ [ self client ] isConnected ] )
    {
        /*
         * try to handle a ~ in a remote dir path.
         * this will fail if some home dirs are
         * not in the same location as the logged in user's.
         * will also fail if user specified a dir
         * in the connect pane.
         */
        NSString* newpath = [ [ gotoDirNameField stringValue ] stringByExpandingTildeInRemotePathWithHomeSetAs: remoteHome ];

        [ [ self client ] setPath: newpath ];

        [ self dismissGotoDirPanel: nil ];

        [ gotoDirNameField insertItemWithObjectValue: [ gotoDirNameField stringValue ] atIndex: 0 ];
    }
}

- ( IBAction )dismissGotoDirPanel: ( id )sender
{
    [ gotoDirPanel orderOut: nil ];
    [ NSApp endSheet: gotoDirPanel ];
}

- ( IBAction )cdRemoteHome: ( id )sender
{
    if ( [ [ self client ] isConnected ] )
    {
        [ [ self client ] setPath: remoteHome ];
    }
}

- ( IBAction )cdLocalHome: ( id )sender
{
    NSString		*hdir = [[ NSUserDefaults standardUserDefaults ]
                                    objectForKey: @"defaultdir" ];
                                    
    if ( hdir == nil || ! [ hdir length ] ) hdir = NSHomeDirectory();
    [ self localBrowserReloadForPath: hdir ];
}

- ( IBAction )remoteCdDotDot: ( id )sender
{
    if ( [ [ self client ] isConnected ] )
    {
        [ [ self client ] changeToParentPath ];
    }
}

- ( IBAction )localCdDotDot: ( id )sender
{
    NSString			*item;
    
    item = [ localDirPath stringByDeletingLastPathComponent ];
    if ( item == nil || ![ item length ] ) return;
    [ self localBrowserReloadForPath: item ];
}

- ( IBAction )cdFromRemoteHistoryMenu: ( id )sender
{
    [ [ self client ] setPath: [ sender title ] ];
}

- ( IBAction )cdFromLocalHistoryMenu: ( id )sender
{
    [ self localBrowserReloadForPath: [ sender title ]];
}

- ( IBAction )cdFromLPathPopUp: ( id )sender
{
    NSString	*path, *component;
    int		index, i = 0;
    
    path = [ NSString stringWithString: localDirPath ];
    component = [ lPathPopUp titleOfSelectedItem ];
    index = [ lPathPopUp indexOfSelectedItem ];
    
    if ( [ component isEqualToString: @"/" ] ) {
        [ self localBrowserReloadForPath: @"/" ];
        return;
    }
    
    if ( ! [ path containsString: component ] ) {
        NSLog( @"Screwed up local popup list." );
        [ self localBrowserReloadForPath: path ];
        return;
    }
    for ( i = 0; i != index; i++ ) {
        path = [ path stringByDeletingLastPathComponent ];
    }
    if ( ! [[ path lastPathComponent ] isEqualToString: component ] ) {
        NSLog( @"Failed to find %@ in %@", component, localDirPath );
        return;
    }
    
    [ self localBrowserReloadForPath: path ];
}

- ( IBAction )cdFromRPathPopUp: ( id )sender
{
    if ( [ [ self client ] isConnected ] )
    {
        NSString* path = [ NSString stringWithString: remoteDirPath ];
        NSString* component = [ rPathPopUp titleOfSelectedItem ];
        int index = [ rPathPopUp indexOfSelectedItem ];
        
        if ( [ component isEqualToString: @"/" ] )
        {
            [ [ self client ] setPath: @"/" ];   // XXX - why the special case? (investigate)
            return;
        }
        
        if ( ! [ path containsString: component ] )
        {
            NSLog( @"Screwed up remote popup list. Correcting." );
            [ [ self client ] refresh ];
            return;
        }

        int i;
        for ( i = 0; i != index; i++ ) {
            path = [ path stringByDeletingLastPathComponent ];
        }
        if ( ! [[ path lastPathComponent ] isEqualToString: component ] ) {
            NSLog( @"Failed to find %@ in %@", component, remoteDirPath );
            return;
        }
        
        [ [ self client ] setPath: path ];
    }
}

- ( void )performSpringLoadedActionInTable: ( NSTableView * )table
{
    NSPoint location = [ table convertPoint: [[ table window ] convertScreenToBase: [ NSEvent mouseLocation ] ]
                                   fromView: [[ table window ] contentView ]];

    int     row     = [ table rowAtPoint: location ];

    if ( row < 0 )
    {
        NSBeep();
    }
    else
    {
        if ( [ self springLoadedRootPath ] == nil )
        {
            [ self setSpringLoadedRootPathInTable: table ];
        }

        if ( [ table isEqual: localBrowser ] )
        {
            NSDictionary* file = [ self localFileAtIndex: row ];
            NSString*     path = [ file objectForKey: @"name" ];

            if ( [[ file objectForKey: @"type" ] isEqualToString: NSLocalizedString( @"directory", @"directory" ) ] )
            {
                [ self localBrowserReloadForPath: path ];
            }
        }
        else if ( [ table isEqual: remoteBrowser ] )
        {
            NSDictionary* file = [ self remoteFileAtIndex: row ];
            NSString*     path = [ file objectForKey: @"name" ];

            if ( [[ file objectForKey: @"type" ] isEqualToString: NSLocalizedString( @"directory", @"directory" ) ] )
            {
                [[ self client ] setPath: path ];
            }
        }
    }
}

- ( void )setSpringLoadedRootPathInTable: ( NSTableView * )table
{
    if ( _springLoadedRootPath != nil ) {
        [ _springLoadedRootPath release ];
        _springLoadedRootPath = nil;
    }
    
NSLog( @"setting springloaded root" );
    if ( [ table isEqual: remoteBrowser ] ) {
        _springLoadedRootPath = [ remoteDirPath copy ];
    } else if ( [ table isEqual: localBrowser ] ) {
        _springLoadedRootPath = [ localDirPath copy ];
    }
}

- ( NSString * )springLoadedRootPath
{
    return( _springLoadedRootPath );
}

- ( void )springLoadedActionCancelledInTable: ( SFTPTableView * )table
{
    NSString*  path = [ self springLoadedRootPath ];

    if ( path != nil )
    {
        if ( [ table isEqual: localBrowser ] )
        {
            [ self localBrowserReloadForPath: path ];
        }
        else if ( [ table isEqual: remoteBrowser ] )
        {
            [[ self client ] setPath: path ];
        }
        
        if ( _springLoadedRootPath != nil )
        {
            [ _springLoadedRootPath release ];
            _springLoadedRootPath = nil;
        }
    }
}

#pragma mark -- Upload --

// ASSumptions: (i.e. pre-conditions)
//  1. local directory exists
//  2. remote directory does NOT exist.  Only case where it would exist is
//     if replacing an existing directory, which should've been deleted
//     before calling this.  XXX this is currently not the case; uploadFiles:
//     should schedule a deletion.

- ( void )uploadDirectory: ( NSString * )localDirectoryPath toDirectory: ( NSString* )remoteDirectoryPath
{
#if 1
    // Notes:
    //  Finder doesn't follow sym links when copying
    //  The links themselves are copied
    //  NSDirectoryEnumerator doesn't traverse links to directories
    //  It does cross device boundaries
    //  Old implementation didn't pass LOGICAL or PHYSICAL flag to fts_open - unknown behavior (haven't tried it).
    //  Not sure how aliases are treated
    //
    //  Behavior here:
    //  Symlinks to files will upload the file.
    //  Preference to follow directory links? (always follow, always skip, prompt)
    //  Will skip for now.

    // TBD - add dependencies to SFTP commands, so that all dependent commands are cancelled if one fails...

    // call file manager to get directory enumerator for local path
    NSDirectoryEnumerator*  e = [ [ NSFileManager defaultManager ] enumeratorAtPath: localDirectoryPath ];
    NSString*  relativePath;

    // queue command to make base directory
    [ [ self client ] queueCommand: [ [ [ SFTPMkdirCommand alloc ] initWithRemotePath: remoteDirectoryPath ] autorelease ] ];

    // traverse directory tree.
    while ( ( relativePath = [ e nextObject ] ) != nil )
    {
        NSString* localPath  = [ localDirectoryPath stringByAppendingPathComponent: relativePath ];
        NSString* remotePath = [ remoteDirectoryPath stringByAppendingPathComponent: relativePath ];

        NSLog(@"Local: %@  Remote: %@", localPath, remotePath);

        NSString* fileType = [ [ e fileAttributes ] objectForKey: NSFileType ];

        if ( fileType == NSFileTypeDirectory)
        {
            [ [ self client ] queueCommand: [ [ [ SFTPMkdirCommand alloc ] initWithRemotePath: remotePath ] autorelease ] ];
        }
        else if ( fileType == NSFileTypeSymbolicLink || fileType == NSFileTypeRegular )
        {
            BOOL isDirectory;

            // Note: we know the file exists, so return value is not checked
            [ [ NSFileManager defaultManager ] fileExistsAtPath: localPath
                                                    isDirectory: &isDirectory ];
            if ( isDirectory )
            {
                // symbolic link to a directory:
#warning "uploadDirectory:toDirectory skips symlinks to directories"
                continue;
            }

            [ self uploadFile: localPath toPath: remotePath ];
        }
        else
        {
                NSLog(@"%s: skipping %@", _cmd, relativePath);
        }
    }    
    
#else  //XXX
    NSString		*pfb = nil;
    NSAutoreleasePool	*p = [[ NSAutoreleasePool alloc ] init ];
    FTS                 *tree = NULL;
    FTSENT              *item = NULL;
    char		dirpath[ MAXPATHLEN ];
    char                *ftspaths[ 2 ] = { NULL, NULL };
    int                 isdir = 0;
    
    if ( [ directoryPath length ] > sizeof( dirpath )) {
        NSLog( @"%@: too long\n" );
        return;
    }
    strcpy( dirpath, [ directoryPath UTF8String ] );
    
    ftspaths[ 0 ] = dirpath;
    
    if (( tree = fts_open( ftspaths, FTS_COMFOLLOW, 0 )) == NULL ) {
        NSLog( @"fts_open %s: %s", dirpath, strerror( errno ));
        return;
    }

    while (( item = fts_read( tree )) != NULL ) {
        isdir = 0;
        
        switch ( item->fts_info ) {
        case FTS_D:
            isdir = 1;
            break;
            
        case FTS_DP:
            continue;
        
        case FTS_ERR:
            NSLog( @"fts_read: %s", strerror( item->fts_errno ));
            continue;
        
        default:
            break;
        }

        pfb = [ NSString pathFromBaseDir: basedir
                fullPath: [ NSString stringWithUTF8String: item->fts_path ]];

        [ uploadQueue addObject:
                [ NSDictionary dictionaryWithObjectsAndKeys:
                [ NSNumber numberWithInt: isdir ], @"isdir",
                [ NSString stringWithUTF8String: item->fts_path ], @"fullpath",
                pfb, @"pathfrombase", nil ]];
    }
    if ( errno ) {
        NSLog( @"fts_read: %s", strerror( errno ));
    }
    
    if ( fts_close( tree ) != 0 ) {
        NSLog( @"fts_close: %s", strerror( errno ));
    }
    
    [ p release ];
#endif
}

- ( void )showUploadProgress
{
    if ( [ uploadProgPanel isVisible ] || [ [ self window ] isMiniaturized ] ) {
	return;
    }
    
    [ uploadProgBar setMaxValue: ( double )[ [ [ self client ] commandQueue ] count ]];
    [ uploadProgBar setUsesThreadedAnimation: YES ];
    [ uploadProgBar setIndeterminate: NO ];
    [ uploadTimeField setStringValue: @"00:00:00" ];
    
    ultime = 0;
    timer = [ NSTimer scheduledTimerWithTimeInterval: 5.0 target: self
			selector: @selector( ulUpdate ) userInfo: nil
			repeats: YES ];
    
    /* disable manual command button while uploading */
    [ commandButton setEnabled: NO ];

    if ( [ [ self window ] isVisible ] ) {
	[ NSApp beginSheet: uploadProgPanel
		modalForWindow: [ self window ]
		modalDelegate: self
		didEndSelector: NULL
		contextInfo: nil ];
    }
}

- ( void )ulUpdate
{
    ultime += 5.0;
    [ uploadTimeField setStringValue: [ NSString clockStringFromInteger: ultime ]];
}

- ( void )updateUploadProgress
{
    NSEnumerator*  e = [ [ [ self client ] commandQueue ] objectEnumerator ];
    id             cmd;
    unsigned int  uploadCount = 0;

    while ( ( cmd = [ e nextObject ] ) != nil )
    {
        if ( [ cmd isKindOfClass: [ SFTPFileTransferCommand class ] ] && ![ cmd isComplete ] )
        {
            uploadCount++;
        }
    }

    if ( uploadCount > 0 ) 
    {
        cmd = [ [ self client ] currentCommand ];
        
        if ( [ cmd isKindOfClass: [ SFTPFileTransferCommand class ] ] )
        {
            [ uploadProgName setStringValue: [ (SFTPFileTransferCommand*)cmd file ] ];
        }

        [ uploadProgItemsLeft setIntValue: uploadCount ];
        
#if 0  // XXX fix - probably just delete - tkm  
        if ( ssh_version < 3.6 ) {
            [ uploadProgBar incrementBy: 1.0 ];
            [ uploadProgBar displayIfNeeded ];
        }
    }
    else if ( /* ??? */ )
    {
        [ uploadProgName setStringValue: [ (SFTPFileTransferCommand*)[ [ self client ] currentCommand ] file ] ];
        [ uploadProgItemsLeft setIntValue: uploadCount ];

	if ( ssh_version < 3.6 ) {
	    [ uploadProgBar setIndeterminate: YES ];
	    [ uploadProgBar startAnimation: nil ];
	    [ uploadProgBar displayIfNeeded ];
	}
#endif
    }
    else
    {
        [ uploadProgBar stopAnimation: nil ];
        [ uploadProgBar setDoubleValue: 0.0 ];
        [ uploadProgName setStringValue: @"" ];
        [ uploadProgItemsLeft setStringValue: @"" ];
	[ uploadProgInfo setStringValue: @"" ];

        [ uploadProgPanel orderOut: nil ];
        [ NSApp endSheet: uploadProgPanel ];

        [ timer invalidate ];
        timer = nil;

        /* update remote display to show newly uploaded items */
        [ remoteDirContents release ];
        remoteDirContents = nil;
        remoteDirContents = [[ NSMutableArray alloc ] init ];
        [ remoteBrowser reloadData ];

        /* done uploading, enable manual command button */
        [ commandButton setEnabled: YES ];
    }
}

- ( void )updateUploadProgressBarWithValue: ( double )value
	    amountTransfered: ( NSString * )amount
	    transferRate: ( NSString * )rate
	    ETA: ( NSString * )eta
{
    double		pc_done = value;

    [ uploadProgBar stopAnimation: nil ];
    [ uploadProgBar setIndeterminate: NO ];
    [ uploadProgBar setMinValue: 0.0 ];
    [ uploadProgBar setMaxValue: 100.0 ];
    [ uploadProgBar setDoubleValue: pc_done ];
    [ uploadProgInfo setStringValue: [ NSString stringWithFormat: @"%@ %@ %@", amount, rate, eta ]];
}

#pragma mark -- log --

- ( void )clearLog
{
    [ logField setString: @"" ];
}

- ( void )addToLog: ( NSString * )text
{
    [ logField setEditable: YES ];
    [ logField insertText: text ];
    [ logField setEditable: NO ];
}

#pragma mark -- connecting --

- ( void )updateHostList
{
    [ remoteHost insertItemWithObjectValue: [ remoteHost stringValue ] atIndex: 0 ];
    
    if ( [ remoteHost numberOfItems ] < 10 ) {
        [ remoteHost setNumberOfVisibleItems: [ remoteHost numberOfItems ]];
    } else {
        [ remoteHost setNumberOfVisibleItems: 10 ];
    }
}

- ( IBAction )showConnectingInterface: ( id )sender
{
    [ [ self window ] makeKeyAndOrderFront: nil ];
    
    [ connectingToField setStringValue:
        [ NSString stringWithFormat:
            NSLocalizedString( @"Connecting to %@...", @"Connecting to %@..." ), [[ self client ] server ]]];
    [ connectingProgress setUsesThreadedAnimation: YES ];
    [ connectingProgress startAnimation: nil ];
    [ remoteBox setContentView: nil ];
    [ remoteBox setContentView: connectingView ];
}

- (void)sftpClient:(SFTPClientTaskController*)c atUnknownHostKeyPrompt:(NSString*)prompt
{
    [ unauthHostInfo setStringValue: prompt ];
    [ unauthHostInfo setEditable: NO ];
#if 0 // XXX - no equivalent to this in new code yet
    [ hostKeyUnauth setStringValue: [ hostInfo objectForKey: @"key" ]];
    [ hostKeyUnauth setEditable: NO ];
#endif

    [ NSApp beginSheet: verifyConnectSheet modalForWindow: [ self window ]
            modalDelegate: self didEndSelector: NULL contextInfo: nil ];
}

- ( IBAction )continueConnecting: ( id )sender
{
    [ verifyConnectSheet orderOut: nil ];
    [ NSApp endSheet: verifyConnectSheet ];

    [ [ self client ] sendUnknownHostKeyResponse: YES ];
}

- ( IBAction )cancelConnection: ( id )sender
{
    // XXX - only approximates original behavior? - TKM
    if ( [ [ [ self client ] commandQueue ] count ] )
    {
	int rc = NSRunAlertPanel( NSLocalizedString( @"Warning: Upload queue not empty.",
                                                     @"Warning: Upload queue not empty." ),
                                  NSLocalizedString( @"There are %d items left in the "
                                                     @"upload queue. Cancelling will end this session. Do "
                                                     @"you want to disconnect without uploading all of them?",
                                                     @"There are %d items left in the "
                                                     @"upload queue. Cancelling will end this session. Do "
                                                     @"you want to disconnect without uploading all of them?" ),
                                  NSLocalizedString( @"Disconnect", @"Disconnect" ),
                                  NSLocalizedString( @"Cancel", @"Cancel" ), @"", [ [ [ self client ] commandQueue ] count ] );
        
        switch( rc )
        {
            case NSAlertDefaultReturn:
                break;
                
            default:
            case NSAlertAlternateReturn:
                return;
        }
    }

    [ [ self client ] disconnect ];

    if ( [ uploadProgPanel isVisible ] )
        [ self updateUploadProgress ];
    
    if ( [ downloadSheet isVisible ] )
        [ self finishedDownload: nil ];
    
    [ logField insertText: @"\nCaught request to cancel connection. Ending....\n" ];
    
    if ( [ verifyConnectSheet isVisible ] )
    {
        [ [ self client ] sendUnknownHostKeyResponse: NO ];

        [ verifyConnectSheet orderOut: nil ];
        [ NSApp endSheet: verifyConnectSheet ];
    }
        
    [ remoteBox setContentView: nil ];
    [ remoteBox setContentView: loginView ];
    [ connectButton setEnabled: YES ];
}

- ( BOOL )firstPasswordPrompt
{
    return( firstPasswordPrompt );
}

- ( void )setFirstPasswordPrompt: ( BOOL )first
{
    firstPasswordPrompt = first;
}

- (void)sftpClient:(SFTPClientTaskController*)c atPasswordPrompt:(NSString*)header
{
    [ passHeader setStringValue: header ];

    BOOL isFirst = [ self firstPasswordPrompt ];

    if ( isFirst )
    {
        [ self setFirstPasswordPrompt: NO ];
    }
    else
    {
        [ passErrorField setStringValue:
            NSLocalizedString( @"Permission denied. Try again.",
                               @"Permission denied. Try again." ) ];
        [ addToKeychainSwitch setState: NSOffState ];
    }        

    /* if we find the password in the keychain, don't show the prompt */
    if ( isFirst && [ self retrievePasswordFromKeychain ] )
        return;

    /*
     * if we get another password prompt and we tried using the password
     *  from the keychain, the password is wrong. Let the user know.
     */
    if ( ! isFirst && [ self gotPasswordFromKeychain ] )
    {
        int			rc;
        
        rc = NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                              NSLocalizedString( @"The password for this server in your "
                                                 @"keychain does not seem to be correct. Would "
                                                 @"you like to open Keychain Access to edit it?",
                                                 @"The password for this server in your "
                                                 @"keychain does not seem to be correct. Would "
                                                 @"you like to open Keychain Access to edit it?" ),
                              NSLocalizedString( @"Open Keychain Access", @"Open Keychain Access" ),
                              NSLocalizedString( @"Cancel", @"Cancel" ), @"" );
        
        switch ( rc )
        {
            case NSAlertDefaultReturn:
                if ( ! [[ NSWorkspace sharedWorkspace ] launchApplication: @"Keychain Access" ] )
                {
                    NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                                     @"Couldn't open Keychain Access", NSLocalizedString( @"OK", @"OK" ), @"", @"" );
                }
                break;

            default:
                break;
        }
    }

    [ authProgBar stopAnimation: nil ];
    [ authProgBar retain ];
    [ authProgBar removeFromSuperview ];
    [ passAuthButton setEnabled: YES ];
    [ passWord setEnabled: YES ];
    [ connectingProgress stopAnimation: nil ];
    [ remoteBox setContentView: nil ];
    [ remoteBox setContentView: passView ];
    [[ passWord window ] makeFirstResponder: passWord ];
}

- ( BOOL )retrievePasswordFromKeychain
{
    char		*password;
    OSStatus		error;
    
    if (( password = getpwdfromkeychain( [[ remoteHost stringValue  ] UTF8String ],
                                [[ userName stringValue ] UTF8String ], &error )) == NULL )
    {
        if ( error == errSecItemNotFound )
        {
            NSLog( @"Keychain item not found" );
        }
        else
        {
            NSLog( @"Attempting to retrieve password from keychain return error %d", error );
        }
        return NO;
    }

    [ self setGotPasswordFromKeychain: YES ];
    [ [ self client ] sendPassword: [ NSString stringWithCString: password ] ];  // XXX get rid of Cstring usage if possible - TKM

    free( password );

    return YES;
}

- ( void )addPasswordToKeychain
{
    addpwdtokeychain( [[ remoteHost stringValue ] UTF8String ],
                        [[ userName stringValue ] UTF8String ],
                        [[ passWord stringValue ] UTF8String ] );
}

- ( void )setGotPasswordFromKeychain: ( BOOL )rp
{
    gotPasswordFromKeychain = rp;
}

- ( BOOL )gotPasswordFromKeychain
{
    return( gotPasswordFromKeychain );
}

- ( IBAction )sendPassword: ( id )sender
{
    if ( [[ passWord stringValue ] length ] > _PASSWORD_LEN )  // XXX is this needed now that fixed buffer isn't used?  Check.  -tkm
    {
	NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                         NSLocalizedString( @"Password is too long.", @"Password is too long." ),
                         NSLocalizedString( @"OK", @"OK" ), @"", @"" );
	[ passWord setStringValue: @"" ];
	return;
    }
    
    if ( [ addToKeychainSwitch state ] == NSOnState )
    {
        [ self addPasswordToKeychain ];
    }
    
    [ passErrorField setStringValue: @"" ];
    [ passView addSubview: authProgBar ];
    [ authProgBar setUsesThreadedAnimation: YES ];
    [ authProgBar startAnimation: nil ];
    [ passHeader setStringValue: NSLocalizedString( @"Authenticating...",
                                                    @"Authenticating..." ) ];
    [ passAuthButton setEnabled: NO ];
    
    [ passWord setEnabled: NO ];
    
    [ [ self client ] sendPassword: [ passWord stringValue ] ];

    [ passWord setStringValue: @"" ];
}

- ( IBAction )addToFavorites:( id )sender
{
    if ( [ remoteHost stringValue ] ) {
        NSUserDefaults		*defaults;
        NSMutableArray		*favarray;
        NSArray			*tmp;
        NSDictionary		*dict;
        NSString		*host = [ remoteHost stringValue ];
        NSString		*user = [ userName stringValue ];
        NSString		*port = [ portField stringValue ];
        NSString		*dir = [ loginDirField stringValue ];
        NSString                *opts = [ advAdditionalOptionsField stringValue ];
        int                     ssh1 = [ advForceSSH1Switch state ];
        int                     compress = [ advEnableCompressionSwitch state ];
        
        if ( ! user ) user = @"";
        if ( ! port ) port = @"";
        if ( ! dir ) dir = @"";
        if ( ! opts ) opts = @"";
        
        dict = [ NSMutableDictionary dictionaryWithObjectsAndKeys:
                    @"", @"nick",
                    host, @"host",
                    user, @"user",
                    port, @"port",
                    dir, @"dir",
                    opts, @"options",
                    [ NSNumber numberWithInt: ssh1 ], @"ssh1",
                    [ NSNumber numberWithInt: compress ], @"compress", nil ];
                        
        defaults = [ NSUserDefaults standardUserDefaults ];
        tmp = [ defaults objectForKey: @"Favorites" ];
        favarray = [ NSMutableArray array ];
        if ( tmp ) {
            [ favarray addObjectsFromArray: tmp ];
        }
            
        [ favarray addObject: dict ];
        [[ NSUserDefaults standardUserDefaults ] setObject: favarray forKey: @"Favorites" ];
        [ defaults synchronize ];
        [ self reloadDefaults ];
    }
}

- ( IBAction )selectFromFavorites: ( id )sender
{
    int i = ( [ popUpFavs indexOfSelectedItem ] - 1 );

    if ( i > -1 )
    {
        
        NSArray* favarray = [[ NSUserDefaults standardUserDefaults ] objectForKey: @"Favorites" ];
        id       fobj     = [ favarray objectAtIndex: i ];
        
        if ( [ fobj isKindOfClass: [ NSString class ]] )
        {
            [ remoteHost setStringValue: fobj ];
        }
        else if ( [ fobj isKindOfClass: [ NSDictionary class ]] )
        {
            [ remoteHost setStringValue: [ fobj objectForKey: @"host" ]];
            [ userName setStringValue: [ fobj objectForKey: @"user" ]];
            [ portField setStringValue: [ fobj objectForKey: @"port" ]];
            [ loginDirField setStringValue: [ fobj objectForKey: @"dir" ]];
            if ( [ fobj objectForKey: @"options" ] )
            {
                [ advAdditionalOptionsField setStringValue: [ fobj objectForKey: @"options" ]];
            }
            [ advForceSSH1Switch setState:
                [[ fobj objectForKey: @"ssh1" ] intValue ]];
            [ advEnableCompressionSwitch setState:
                [[ fobj objectForKey: @"compress" ] intValue ]];
            
            [[ remoteHost window ] makeFirstResponder: remoteHost ];
        }
    }
}

#pragma mark -- end connecting --

// XXX - rename not in menu?  Add
- ( IBAction )renameLocalItem: ( id )sender
{
    int row = [ localBrowser selectedRow ];

    if ( row > -1 )
    {
        NSArray*  items = ( dotflag ? localDirContents : dotlessLDir );
        NSString* name  = [[ items objectAtIndex: row ] objectForKey: @"name" ];

        if ( name != nil )
        {
            [[ localBrowser selectedCell ] setEditable: YES ];
            [[ localBrowser selectedCell ] setScrollable: YES ];
            [ localBrowser editColumn: 0 row: row withEvent: nil select: YES ];
            [[ localBrowser selectedCell ] setStringValue: [ name lastPathComponent ]];
        }
    }
}

- ( IBAction )renameRemoteItem: ( id )sender
{
    int row = [ remoteBrowser selectedRow ];

    if ( row > -1 )
    {
        NSArray*  items = ( dotflag ? remoteDirContents : dotlessRDir );
        NSString* name  = [[ items objectAtIndex: row ] objectForKey: @"name" ];

        if ( name != nil )
        {
            [[ remoteBrowser selectedCell ] setEditable: YES ];
            [[ remoteBrowser selectedCell ] setScrollable: YES ];
            [ remoteBrowser editColumn: 0 row: row withEvent: nil select: YES ];
            [[ remoteBrowser selectedCell ] setStringValue: name ];
        }
    }
}

- ( IBAction )previewItem: ( id )sender
{
    id browser = [ [ self window ] firstResponder ];

    if ( ! [ browser isKindOfClass: [ SFTPTableView class ]] )
    {
        NSBeep();
    }
    else
    {
        if ( [ browser isEqual: localBrowser ] )
        {
            [ self previewLocalItem: sender ];
        }
        else if ( [ browser isEqual: remoteBrowser ] )
        {
            [ self previewRemoteItem: sender ];
        }
    }
}

- ( IBAction )previewLocalItem: ( id )sender
{
    if ( [ localBrowser selectedRow ] < 0 )
    {
        NSBeep();
    }
    else
    {
        NSArray*  items    = ( dotflag ? localDirContents : dotlessLDir );
        NSString* filepath = [[ items objectAtIndex: [ localBrowser selectedRow ] ] objectForKey: @"name" ];

        if ( [ [ NSSet validImageExtensions ] containsObject: [ [ filepath pathExtension ] lowercaseString ]] )
        {
            [ self setPreviewedImage: filepath ];
        }

        [ self displayPreview ];
    }
}

- ( IBAction )previewRemoteItem: ( id )sender
{
    if ( [ remoteBrowser selectedRow ] < 0 )
    {
        NSBeep();
    }
    else
    {
        NSArray*  items = ( dotflag ? remoteDirContents : dotlessRDir );
        NSString* filename  = [[ items objectAtIndex: [ remoteBrowser selectedRow ] ] objectForKey: @"name" ];
        NSString* filepath  = [ NSString stringWithFormat: @"%@/%@", remoteDirPath, filename ];
        
        NSEnumerator* previewEnumerator = [ [ self cachedPreviews ] objectEnumerator ];
        NSDictionary* preview;

        while ( ( preview = [ previewEnumerator nextObject ] ) != nil )
        {
            if ( [ [ preview objectForKey: @"RemotePath" ] isEqualToString: filepath ] )
            {
                [ self setPreviewedImage: [ preview objectForKey: @"LocalPath" ]];
                [ self displayPreview ];
                return;
            }
        }

        if ( ! [ [ NSSet validImageExtensions ] containsObject: [ [ filename pathExtension ] lowercaseString ]] )
        {
            [ self displayPreview ];
            return;
        }

        NSString* tempDir = [[ NSFileManager defaultManager ] makeTemporaryDirectoryWithMode: ( mode_t )0700 ];

        if ( tempDir == nil )
        {
            NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                             @"mkdir: %s", NSLocalizedString( @"OK", @"OK" ), @"", @"",
                             strerror( errno ));
            return;
        }

        tempDir = [ tempDir stringByAppendingPathComponent: filename ];

        [ self setPreviewedImage: tempDir ];

        [ self addToCachedPreviews:
            [ NSDictionary dictionaryWithObjectsAndKeys:
                filepath, @"RemotePath",
                tempDir, @"LocalPath",
                nil ]];

        [ [ self client ] queueCommand: [ [ [ SFTPFileTransferCommand alloc ] initWithLocalPath: tempDir
                                                                                     remotePath: [ remoteDirPath stringByAppendingPathComponent: filename ]
                                                                              transferDirection: SFTPFileTransferDirectionDownload ]
            autorelease ] ];
    }
}
    
- ( void )displayPreview
{
    NSImage			*image = nil;
    NSString			*name = nil, *path = nil;
    NSSize			imageSize;
    NSRect			originalPanelFrame;
    NSPoint			cellorigin;
    id				table = [ [ self window ] firstResponder ];
    int				row;
        
    if ( ! [ table isKindOfClass: [ SFTPTableView class ]] ) {
        return;
    }
    
    if (( row = [ table selectedRow ] ) < 0 ) {
        NSBeep();
        return;
    }
    
    if ( [ self previewedImage ] == nil ) {
        NSArray			*items = nil;
        NSString		*extension;

        if ( [ table isEqual: localBrowser ] ) {
            items = ( dotflag ? localDirContents : dotlessLDir );
            name = [[ items objectAtIndex: row ] objectForKey: @"name" ];
            image = [[ NSWorkspace sharedWorkspace ] iconForFile: name ];
        } else if ( [ table isEqual: remoteBrowser ] ) {
            items = ( dotflag ? remoteDirContents : dotlessRDir );
            name = [[ items objectAtIndex: row ] objectForKey: @"name" ];
            extension = [ name pathExtension ];
            
            if ( [ extension isEqualToString: @"" ] ) {
                if ( [[[ items objectAtIndex: row ] objectForKey: @"perm" ]
                            characterAtIndex: 0 ] == 'd' ) {
                    image = [[ NSWorkspace sharedWorkspace ] iconForFileType: @"'fldr'" ];
                } else {
                    image = [[ NSWorkspace sharedWorkspace ] iconForFileType: @"'doc '" ];
                }
            } else {
                image = [[ NSWorkspace sharedWorkspace ] iconForFileType: extension ];
            }
        }
        [ image setScalesWhenResized: YES ];
        [ image setSize: NSMakeSize( 128.0, 128.0 ) ];
        [ image retain ];
    } else {
        image = [[ NSImage alloc ] initWithContentsOfFile: [ self previewedImage ]];
        
        if ( image == nil ) {
            NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                    NSLocalizedString( @"Couldn't create preview of %@",
                                        @"Couldn't create preview of %@" ),
                    NSLocalizedString( @"OK", @"OK" ), @"", @"",
                    [[ self previewedImage ] lastPathComponent ] );
            return;
        }
        path = [ self previewedImage ];
        name = [ path lastPathComponent ];
    }
    originalPanelFrame = [ imagePreviewPanel frame ];
    cellorigin = [ table originOfSelectedCell ];
    
    imageSize = [ image size ];
    [ image setScalesWhenResized: YES ];

    [ imagePreviewPanel setTitle: [ NSString stringWithFormat: @"Preview: %@", name ]];
    [ imagePreviewTextField setStringValue:
            [ NSString stringWithFormat:
                    NSLocalizedString( @"Image size: %.0f x %.0f", @"Image size: %.0f x %.0f" ),
                    imageSize.width, imageSize.height ]];
                    
    [ imagePreview setImage: image ];
    [ imagePreview setImageLocationPath: path ];
                        
    if ( ! [ imagePreviewPanel isVisible ] ) {
        [ imagePreviewBox setContentView: nil ];
        [ imagePreviewPanel zoomFromRect: NSMakeRect( cellorigin.x, cellorigin.y, 1.0, 1.0 )
                                toRect: originalPanelFrame ];
        [ imagePreviewBox setContentView: imagePreviewView ];
    }
    [ image release ];
    [ self setPreviewedImage: nil ];
}

#pragma mark -- Open File in Editor --

- ( IBAction )editFile: ( id )sender
{
    id			table = [ [ self window ] firstResponder ];
    
    if ( ! [ table isKindOfClass: [ SFTPTableView class ]] ) {
        return;
    }
    
    if ( [ table isEqual: localBrowser ] ) {
        [ self openLocalFileInEditor: sender ];
    } else if ( [ table isEqual: remoteBrowser ] ) {
        [ self openRemoteFileInEditor: sender ];
    }
}

- ( NSString * )promptUserForApplication
{
    NSOpenPanel* openPanel = [ NSOpenPanel openPanel ];
    NSArray*     paths = NSSearchPathForDirectoriesInDomains(NSApplicationDirectory,
                                                             NSLocalDomainMask,
                                                             NO);
    NSArray*     types = [ NSArray arrayWithObject: @"app" ];
    NSString*    application = nil;

    NSAssert( [ paths count ] == 1, @"Unexpected number of paths returned." );

    if ( [ openPanel runModalForDirectory: [ paths objectAtIndex: 0 ] 
                                     file: nil
                                    types: types ] == NSOKButton )
    {
        application = [ openPanel filename ];
    }

    return application;
}

- ( IBAction )openLocalFileInEditor: ( id )sender
{
    NSString*     filepath    = [[ self localSelectedFile ] objectForKey: @"name" ];
    NSString*     application = nil;

    if ( [ sender tag ] == FILE_MENU_ITEM_OPEN_WITH )
    {
        application = [ sender title ];
    }
    else if ( [ sender tag ] == OPEN_WITH_MENU_ITEM_OTHER )
    {
        application  = [ self promptUserForApplication ];
        
        if ( application == nil )
        {
            // Cancelled
            return;
        }
    }

    if ( [ [ NSWorkspace sharedWorkspace ] openFile: filepath
                                    withApplication: application ] == NO )
    {
	NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                         @"Couldn't open %@", NSLocalizedString( @"OK", @"OK" ),
                         @"", @"", [ filepath lastPathComponent ] );
	return;
    }
}

- ( IBAction )openRemoteFileInEditor: ( id )sender
{
    NSString*     application = nil;

    if ( [ sender tag ] == FILE_MENU_ITEM_OPEN_WITH )
    {
        application = [ sender title ];
    }
    else if ( [ sender tag ] == OPEN_WITH_MENU_ITEM_OTHER )
    {
        application  = [ self promptUserForApplication ];
        
        if ( application == nil )
        {
            // Cancelled
            return;
        }
    }

    NSString* name           = [[ self remoteSelectedFile ] objectForKey: @"name" ];
    NSString* remoteFilePath = [ remoteDirPath stringByAppendingPathComponent: name ];
    NSString* localFilePath  = [[ self editedDocumentWithRemotePath: remoteFilePath ] objectForKey: @"localpath" ];

    if (localFilePath != nil)
    {
        if ( [ [ NSWorkspace sharedWorkspace ] openFile: localFilePath
                                        withApplication: application ] == NO )
        {
            NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                             @"Couldn't open %@", NSLocalizedString( @"OK", @"OK" ),
                             @"", @"", [ localFilePath lastPathComponent ] );
        }
    }
    else
    {
        NSString* tempDir = [[ NSFileManager defaultManager ] makeTemporaryDirectoryWithMode: ( mode_t )0700 ];
        
        if ( tempDir == nil )
        {
            NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                             @"mkdir: %s", NSLocalizedString( @"OK", @"OK" ), @"", @"",
                             strerror( errno ));
        }
        else
        {
            localFilePath = [ tempDir stringByAppendingPathComponent: name ];
            
            [ self addToEditedDocuments: localFilePath
                             remotePath: remoteFilePath
                                 editor: application ];
            
            [ [ self client ] queueCommand: [ [ [ SFTPFileTransferCommand alloc ] initWithLocalPath: localFilePath
                                                                                         remotePath: remoteFilePath
                                                                                  transferDirection: SFTPFileTransferDirectionDownload ]
                autorelease ] ];
        }
    }
}

#define FileNotifyEvents (UKKQueueNotifyAboutRename | UKKQueueNotifyAboutWrite  | UKKQueueNotifyAboutDelete | UKKQueueNotifyAboutSizeIncrease)

- ( void )editFile: ( NSString * )filepath remotePath: ( NSString * )remotepath withEditor: ( NSString * )editor;
{
    [kqueue addPathToQueue: filepath notifyingAbout: FileNotifyEvents];

    if ([ [ NSWorkspace sharedWorkspace ] openFile: filepath
                                       withApplication: editor ] == NO)
    {
        NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                         @"Failed to open %@.", NSLocalizedString( @"OK", @"OK" ),
                         @"", @"", [ filepath lastPathComponent ] );

        [kqueue removePathFromQueue: filepath];
    }
}

-(void) checkRemovedDocument:(NSString*)fpath
{
    BOOL  doUpload = NO;

    // Most likely, a new file was written with the same name
    if ([[NSFileManager defaultManager] isReadableFileAtPath: fpath])
    {
        // Start watching the new file
        [kqueue addPathToQueue: fpath notifyingAbout: FileNotifyEvents];
        
        doUpload = YES;
    }
    else  // File is gone; this means we are done editing
    {
        NSLog(@"Edited file was deleted/renamed and no new file exists (or is unreadable): %@", fpath);
        
        [ self performSelector: @selector( removeEditedDocumentWithLocalPath: )
                    withObject: fpath afterDelay: 1.0 ];
        
        if ( [[ NSUserDefaults standardUserDefaults ]
                integerForKey: @"PostEditBehaviour" ] == 0 ) {
            [ NSApp activateIgnoringOtherApps: YES ];
        }
    }

    if (doUpload)
    {
        [ self uploadFile: fpath toPath: [self remotePathForEditedDocumentWithLocalPath: fpath] ];
    }
}

-(void) kqueue: (UKKQueue*)kq receivedNotification: (NSString*)nm forFile: (NSString*)fpath
{
    BOOL  doUpload = NO;

    if ([nm isEqualToString: UKKQueueFileWrittenToNotification] ||
        [nm isEqualToString: UKKQueueFileSizeIncreasedNotification])
    {
        doUpload = YES;
    }
    else if ([nm isEqualToString: UKKQueueFileRenamedNotification] ||
                 [nm isEqualToString: UKKQueueFileDeletedNotification])
    {
        // File was deleted or re-named; quit watching it
        [kqueue removePathFromQueue: fpath];

        [ self performSelector: @selector( checkRemovedDocument: )
                    withObject: fpath afterDelay: 1.0 ];
    }
    else
    {
        NSLog(@"Unknown kqueue notification: %@, %@", nm, fpath);
    }
    
    if (doUpload)
    {
        [ self uploadFile: fpath toPath: [self remotePathForEditedDocumentWithLocalPath: fpath] ];
    }
}

- ( IBAction )getInfo: ( id )sender
{
    NSDictionary	*attribDict;
    NSDictionary	*rAttribDict;
    NSString		*fullPath;
    NSArray		*items = nil;
    id			activeTableView;
    CFStringRef		type;
    CFURLRef		url;
    OSStatus		status;
    
    if ( ! [ infoPanel isVisible ] ) {
        [ infoPanel makeKeyAndOrderFront: nil ];
    }
    
    activeTableView = [ [ self window ] firstResponder ];
        
    if ( [ activeTableView isEqual: localBrowser ] || ! [ [ self client ] isConnected ] ) {
        [ infoTabView selectTabViewItemWithIdentifier: @"local" ];
    } else {
        [ infoTabView selectTabViewItemWithIdentifier: @"remote" ];
    }
    
    items = ( dotflag ? localDirContents : dotlessLDir );
    if ( [ localBrowser selectedRow ] == -1 ) {
        fullPath = localDirPath;
        attribDict = [[ SFTPNode sharedInstance ]
                            statInformationForPath: fullPath ];
    } else {
        fullPath = [[ items objectAtIndex: [ localBrowser selectedRow ]] objectForKey: @"name" ];
        attribDict = [ items objectAtIndex: [ localBrowser selectedRow ]];
    }
    
    if ( ! [ fullPath isEqualToString: [ whereField stringValue ]] ) {
        [ infoPathField setStringValue: [ fullPath lastPathComponent ]];
        [ whereField setStringValue: fullPath ];
        
        [ largeIcon setImage: [[ NSWorkspace sharedWorkspace ]
                iconForFile: fullPath ]];
        [ self localCheckSetup: [ attribDict objectForKey: @"perm" ]];
	
	url = CFURLCreateFromFileSystemRepresentation( kCFAllocatorDefault,
		    ( const UInt8 * )[ fullPath fileSystemRepresentation ],
		    strlen( [ fullPath fileSystemRepresentation ] ), false );
	if ( url ) {
	    if (( status = LSCopyKindStringForURL( url, &type )) != noErr ) {
		NSLog( @"LSCopyKindStringForURL failed: error %d", ( int )status );
		[ typeField setStringValue: [ attribDict objectForKey: @"type" ]];
	    } else {
		[ typeField setStringValue: ( NSString * )type ];
		CFRelease( type );
	    }
	    CFRelease( url );
	}
                
        [ ownerField setStringValue: [ attribDict objectForKey: @"owner" ]];
        [ groupField setStringValue: [ attribDict objectForKey: @"group" ]];
        [ sizeField setStringValue: [ NSString stringWithFormat: @"%@ (%@ bytes)",
                                    [[ attribDict objectForKey: @"size" ] descriptiveSizeString ],
                                    [ attribDict objectForKey: @"size" ]]];
        [ modDateField setObjectValue: [ attribDict objectForKey: @"date" ]];
        [ permField setStringValue: [ attribDict objectForKey: @"perm" ]];
    }
    
    if ( ![ [ self client ] isConnected ] ) return;
    
    rAttribDict = (( dotflag ) ? [[ remoteDirContents objectAtIndex:
                                                    [ remoteBrowser selectedRow ]] copy ] :
                                [[ dotlessRDir objectAtIndex: [ remoteBrowser selectedRow ]] copy ] );

    [ rInfoPathField setStringValue: [ rAttribDict objectForKey: @"name" ]];
    [ rWhereField setStringValue:
            [ NSString stringWithFormat: @"%@/%@", remoteDirPath,
            [ rAttribDict objectForKey: @"name" ]]];
    
    [ rIcon setImage: [ NSImage iconForType:
            [[ rAttribDict objectForKey: @"name" ] pathExtension ]]];

    [ rOwnerField setStringValue: [ rAttribDict objectForKey: @"owner" ]];
    
    [ rGroupField setStringValue: [ rAttribDict objectForKey: @"group" ]];
    
    [ rSizeField setStringValue: [ NSString stringWithFormat: @"%@ (%@ bytes)",
                                    [[ rAttribDict objectForKey: @"size" ] descriptiveSizeString ],
                                    [ rAttribDict objectForKey: @"size" ]]];
    [ rModDateField setObjectValue: [ rAttribDict objectForKey: @"date" ]];

    [ rPermField setStringValue: [[ rAttribDict objectForKey: @"perm" ] octalRepresentation ]];
    [ self remoteCheckSetup: [ rAttribDict objectForKey: @"perm" ]];
    [ rTypeField setStringValue: [ rAttribDict objectForKey: @"type" ]];
    if ( [[ rAttribDict objectForKey: @"type" ] isEqualToString: @"directory" ] ) {
        [ rIcon setImage: nil ];
        [ rIcon setImage: [[ NSWorkspace sharedWorkspace ] iconForFileType: @"'fldr'" ]];
    }
    [ rAttribDict release ];
}

#pragma mark -- Permissions --

- ( void )localCheckSetup: ( NSString * )octalmode
{
    int			ow, own, gr, grp, ot, oth;
    
    /* switch all off first */
    [ loReadSwitch setState: NSOffState ];
    [ loWriteSwitch setState: NSOffState ];
    [ loExecSwitch setState: NSOffState ];
    [ lgReadSwitch setState: NSOffState ];
    [ lgWriteSwitch setState: NSOffState ];
    [ lgExecSwitch setState: NSOffState ];
    [ laReadSwitch setState: NSOffState ];
    [ laWriteSwitch setState: NSOffState ];
    [ laExecSwitch setState: NSOffState ];
    
    ow = own = ( [ octalmode characterAtIndex: 1 ] - '0' );
    gr = grp = ( [ octalmode characterAtIndex: 2 ] - '0' );
    ot = oth = ( [ octalmode characterAtIndex: 3 ] - '0' );
    
    if (( own = ( own - 4 )) >= 0 )
                    [ loReadSwitch setState: NSOnState ];
    if (( own = ( own - 2 )) >= 0 || ( own == -2 && ow == 2 ) || ( own == -3 && ow == 3 ))
                    [ loWriteSwitch setState: NSOnState ];
    if (( own = ( own - 1 )) >= 0 || own == -2 || own == -4 || own == -6 )
                    [ loExecSwitch setState: NSOnState ];
    
    if (( grp = ( grp - 4 )) >= 0 )
                    [ lgReadSwitch setState: NSOnState ];
    if (( grp = ( grp - 2 )) >= 0 || ( grp == -2 && gr == 2 ) || ( grp == -3 && gr == 3 ))
                    [ lgWriteSwitch setState: NSOnState ];
    if (( grp = ( grp - 1 )) >= 0 || grp == -2 || grp == -4 || grp == -6 )
                    [ lgExecSwitch setState: NSOnState ];
    
    if (( oth = ( oth - 4 )) >= 0 )
                    [ laReadSwitch setState: NSOnState ];
    if (( oth = ( oth - 2 )) >= 0 || ( oth == -2 && ot == 2 ) || ( oth == -3 && ot == 3 ))
                    [ laWriteSwitch setState: NSOnState ];
    if (( oth = ( oth - 1 )) >= 0 || oth == -2 || oth == -4 || oth == -6 )
                    [ laExecSwitch setState: NSOnState ];
}

- ( void )remoteCheckSetup: ( NSString * )permissions
{
    int			i;
    
    for ( i = 0; i < [ rSwitchArray count ]; i++ ) {
        [[ rSwitchArray objectAtIndex: i ] setState: NSOffState ];
    }
    
    for ( i = 1; i < 4; i++ ) {
        switch( [ permissions characterAtIndex: i ] ) {
        case 'r':
            [ roReadSwitch setState: NSOnState ];
            break;
        case 'w':
            [ roWriteSwitch setState: NSOnState ];
            break;
        case 'x':
        case 's':
            [ roExecSwitch setState: NSOnState ];
            break;
        }
    }
    
    for ( i = 4; i < 7; i++ ) {
        switch( [ permissions characterAtIndex: i ] ) {
        case 'r':
            [ rgReadSwitch setState: NSOnState ];
            break;
        case 'w':
            [ rgWriteSwitch setState: NSOnState ];
            break;
        case 'x':
        case 's':
            [ rgExecSwitch setState: NSOnState ];
            break;
        }
    }
    
    for ( i = 7; i < 10; i++ ) {
        switch( [ permissions characterAtIndex: i ] ) {
        case 'r':
            [ raReadSwitch setState: NSOnState ];
            break;
        case 'w':
            [ raWriteSwitch setState: NSOnState ];
            break;
        case 'x':
        case 't':
            [ raExecSwitch setState: NSOnState ];
            break;
        }
    }
}

#pragma mark -- Connect --

- ( IBAction )toggleAdvConnectionView: ( id )sender
{
    NSRect favsConnectBoxFrame = [ favsConnectBox frame ];
    NSSize advConBoxSize = ((NSRect)[ advConnectionBox frame ]).size;

    [ advConnectionBox setHidden: ![advConnectionBox isHidden] ];

    if ([ advConnectionBox isHidden ]) {
        favsConnectBoxFrame.origin.y += advConBoxSize.height;
    }
    else
    {
        favsConnectBoxFrame.origin.y -= advConBoxSize.height;
    }
    [ favsConnectBox setFrame: favsConnectBoxFrame ];
    [ loginView setNeedsDisplay: YES];
}

- ( IBAction )sftpConnect: ( id )sender
{
    [ connectButton setEnabled: NO ];

#if 0  // XXX use field validator for this:
    if ( [ portString length ] )
    {
        port = [ portString intValue ];
        if ( port == 0 )
        {
            NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                             NSLocalizedString( @"Invalid port number.", @"Invalid port number." ),
                             NSLocalizedString( @"OK", @"OK" ), @"", @"" );
            goto INVALID_CONNECTION_SETTINGS;
        }
    }
#endif

    if ( [ [ userName stringValue ] length ] == 0 || [ [ remoteHost stringValue ] length ] == 0 )
    {
        NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                         NSLocalizedString( @"You must provide a username and server address to continue.",
                                            @"You must provide a username and server address to continue." ),
                         NSLocalizedString( @"OK", @"OK" ), @"", @"" );
        goto INVALID_CONNECTION_SETTINGS;
    }

    NSMutableArray*     params = [NSMutableArray array];

    if ( [ advForceSSH1Switch state ] == NSOnState )
    {
        [ params addObject: @"-1" ];
    }

    if ( [ advEnableCompressionSwitch state ] == NSOnState )
    {
        [ params addObject: @"-C" ];
    }
    
    if ( [[ advAdditionalOptionsField stringValue ] isNotEqualTo: @"" ] )
    {
        [ params addObjectsFromArray:  [ [ advAdditionalOptionsField stringValue ] componentsSeparatedByString: @" " ] ];
    }

    // XXX - implement
#if 0
    if ( /* XXX - check supported sftp client version */)
    {
        /* Let's not pretend we support SSH.com's sftp client */
        NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                @"Unsupported version of sftp.",
                NSLocalizedString( @"OK", @"OK" ), @"", @"" );
        goto INVALID_CONNECTION_SETTINGS;
        break;
     }
#endif

    [ self setFirstPasswordPrompt: YES ];

    remoteDirContents = [[ NSMutableArray alloc ] init ];
    dotlessRDir	      = [[ NSMutableArray alloc ] init ];
    remoteDirPath     = [[ NSString alloc ] init ];

    NSNumber*  port = nil;

    if ( [ [ portField stringValue ] isNotEqualTo: @"" ] )
    {
        port = [ NSNumber numberWithUnsignedShort: [ portField intValue ] ];
    }
    
    [ self clearLog ];

    [ self setClient:
        [ [ [ SFTPClientTaskController alloc ] initWithServer: [ remoteHost stringValue ]
                                                         port: port
                                                         user: [ userName stringValue ]
                                                      options: params ]
            autorelease ] ];
    
    [ [ self client ] connect ];
    
    [ self updateHostList ];	/* adds new host to pop-up list */
    
    [ self showConnectingInterface: nil ];
    
    // Update Recent servers, ensuring uniqueness
    NSArray* uniqueServerList = [ [ NSSet setWithArray: [ remoteHost objectValues ] ] allObjects ];

    [ remoteHost removeAllItems ];
    [ remoteHost addItemsWithObjectValues: uniqueServerList ];

    [[ NSUserDefaults standardUserDefaults ] setObject: uniqueServerList
                                                forKey: @"RecentServers" ];

    return;

INVALID_CONNECTION_SETTINGS:
    [ connectButton setEnabled: YES ];
    return;
}

- ( BOOL )handleChangedText: ( NSString * )newstring
                   forTable: ( SFTPTableView * )table
                     column: ( int )column
{
    int  row     = [ table editedRow ];
    BOOL success = NO;

    if ( row < 0 )
    {
        NSBeep();
    }
    else
    {
        if ( [ newstring isNotEqualTo: @"" ] )
        {
            if ( table == localBrowser )
            {
                NSMutableArray* items   = ( dotflag ? localDirContents : dotlessLDir );
                NSDictionary*   dict    = [ items objectAtIndex: row ];
                NSString*       oldpath = [ dict objectForKey: @"name" ];
                NSString*       newpath =
                    [ NSString stringWithFormat: @"%@/%@",
                        [ oldpath stringByDeletingLastPathComponent ],
                        newstring ];
                
                if ( rename( [ oldpath UTF8String ], [ newpath UTF8String ] ) < 0 )
                {
                    NSRunAlertPanel( NSLocalizedString( @"Error", @"Error" ),
                                     NSLocalizedString( @"rename %@ to %@: %s", @"rename %@ to %@: %s" ),
                                     NSLocalizedString( @"OK", @"OK" ), @"", @"",
                                     [ dict objectForKey: @"name" ],
                                     newpath, strerror( errno ));
                }
                else
                {
                    dict = [[ SFTPNode sharedInstance ] statInformationForPath: newpath ];
                    [ items replaceObjectAtIndex: row withObject: dict ];
                    
                    success = YES;
                }
            }
            else if ( table == remoteBrowser )
            {
                NSArray*  items = ( dotflag ? remoteDirContents : dotlessRDir );
                NSString* oldpath = [[ items objectAtIndex: row ] objectForKey: @"name" ];

                [ [ self client ] queueCommand: [ [ [ SFTPRenameCommand alloc ] initWithOrigPath: oldpath
                                                                                         newPath: newstring ]
                    autorelease ] ];

                success = YES;
            }
        }
    }

    return success;
}

- ( BOOL )handleEvent: ( NSEvent * )theEvent fromTable: ( SFTPTableView * )table
{
    unichar 		key = [[ theEvent charactersIgnoringModifiers ] characterAtIndex: 0 ];
    
    if ( key == NSDeleteCharacter ) {
        [ self delete: nil ];
        return( YES );
    } else if ( key == NSRightArrowFunctionKey ) {
        [ self changeDirectory: nil ];
        return( YES );
    } else if ( key == NSLeftArrowFunctionKey ) {
        [ self dotdot: nil ];
        return( YES );
    } else if ( key == NSEnterCharacter ) {
        if ( [ table isEqual: localBrowser ] ) {
            [ self renameLocalItem: nil ];
        } else if ( [ table isEqual: remoteBrowser ] ) {
            [ self renameRemoteItem: nil ];
        }
        return( YES );
    } else if ( key == NSF8FunctionKey ) {
        int		row = [ table selectedRow ];
        NSMenu		*menu = nil;
        NSEvent		*e = nil;
        NSRect		rect; 
        NSPoint		p;
        
        if ( row < 0 ) {
            return( NO );
        }
        
        rect = [ table frameOfCellAtColumn: 0
                        row: [ table selectedRow ]];
        p = NSMakePoint( rect.origin.x, rect.origin.y );
        p = [[[ table window ] contentView ] convertPoint: p fromView: table ];
        
        if ( [ table isFlipped ] ) {
            p.x += 15.0;
            p.y -= 15.0;
        } else {
            p.x -= 15.0;
            p.y += 15.0;
        }
        
        if ( [ table isEqual: localBrowser ] ) {
            menu = localTableMenu;
        } else if ( [ table isEqual: remoteBrowser ] ) {
            menu = remoteTableMenu;
        }
        e = [ NSEvent mouseEventWithType: NSRightMouseUp location: p
                        modifierFlags: 0 timestamp: 1
                        windowNumber: [[ table window ] windowNumber ]
                        context: [ NSGraphicsContext currentContext ]
                        eventNumber: 1 clickCount: 1 pressure: 0.0 ];
                    
        [ NSMenu popUpContextMenu: menu withEvent: e forView: nil ];
        return( YES );
    } else if ( isascii( key )) {
        int		row;
        
        [ NSObject cancelPreviousPerformRequestsWithTarget: self
                    selector: @selector( clearTypeAheadString )
                    object: nil ];
        [ self performSelector: @selector( clearTypeAheadString )
                withObject: nil
                afterDelay: 0.5 ];
        if ( typeAheadString == nil ) {
            typeAheadString = [[ NSMutableString alloc ] init ];
            [ typeAheadString setString: @"" ];
        }
        [ typeAheadString appendString: [ NSString stringWithFormat: @"%c", key ]];
        row = [ self matchingIndexForString: typeAheadString inTable: table ];
        
        if ( row < 0 ) return( NO );
        [ table scrollRowToVisible: row ];
        [ table selectRow: row byExtendingSelection: NO ];
        return( YES );
    }
        
    return( NO );
}

- ( void )clearTypeAheadString
{
    if ( typeAheadString != nil ) {
        [ typeAheadString release ];
    }
    typeAheadString = nil;
}

- ( int )matchingIndexForString: ( NSString * )string inTable: ( SFTPTableView * )table
{
    NSMutableArray		*items;
    NSDictionary		*dict;
    int				index = -1, i;
    
    if ( string == nil ) return( -1 );
    
    if ( [ table isEqual: localBrowser ] ) {
        items = ( dotflag ? localDirContents : dotlessLDir );
        for ( i = 0; i < [ items count ]; i++ ) {
            dict = [ items objectAtIndex: i ];
            if ( [[[ dict objectForKey: @"name" ]
                        lastPathComponent ] beginsWithString: string ] ) {
                index = i;
                break;
            }
        }
    } else if ( [ table isEqual: remoteBrowser ] && [ [ self client ] isConnected ] ) {
        items = ( dotflag ? remoteDirContents : dotlessRDir );
        for ( i = 0; i < [ items count ]; i++ ) {
            dict = [ items objectAtIndex: i ];
            if ( [[ dict objectForKey: @"name" ] beginsWithString: string ] ) {
                index = i;
                break;
            }
        }
    }
    return( index );
}

#pragma mark -- Edited Documents --

- ( NSArray * )editedDocuments
{
    return [ NSArray arrayWithArray: editedDocuments ];
}

+ (BOOL)accessInstanceVariablesDirectly { return NO; }

- ( void )addToEditedDocuments: ( NSString * )docpath remotePath: ( NSString * )rpath editor: ( NSString * )editor
{
    NSString			*user = [ userName stringValue ];
    NSString			*host = [ remoteHost stringValue ];
    
    if ( editedDocuments == nil ) {
	editedDocuments = [[ NSMutableArray alloc ] init ];
    }
    
    if ( [ user isEqualToString: @"" ] ) {
	user = @"N/A";
    }
    if ( [ host isEqualToString: @"" ] ) {
	host = @"N/A";
    }

    [ self willChangeValueForKey: @"editedDocuments" ];

    [ editedDocuments addObject: [ NSDictionary dictionaryWithObjectsAndKeys:
		    docpath, @"localpath",
		    rpath, @"remotepath",
		    user, @"username",
		    host, @"hostname",
		    editor, @"editor",
		    nil ]];

    [ self didChangeValueForKey: @"editedDocuments" ];
}

- ( NSDictionary* )editedDocumentWithRemotePath: ( NSString * )remotePath
{
    NSEnumerator* e = [ editedDocuments objectEnumerator ];
    NSDictionary* editedDocument;

    while ( ( editedDocument = [ e nextObject ] ) != nil )
    {
	if ( [ remotePath isEqualToString: [ editedDocument objectForKey: @"remotepath" ] ] )
        {
            return editedDocument;
        }
    }

    /*NOTREACHED*/
    return nil;
}

- ( NSDictionary* )editedDocumentWithLocalPath: (NSString *)localPath
{
    NSEnumerator* e = [ editedDocuments objectEnumerator ];
    NSDictionary* editedDocument;

    while ( ( editedDocument = [ e nextObject ] ) != nil )
    {
	if ( [ localPath isEqualToString: [ editedDocument objectForKey: @"localpath" ] ] )
        {
            return editedDocument;
        }
    }

    // Error if we get here
    [NSException raise: NSInternalInconsistencyException format: @"Edited document not found: %@", localPath];

    /*NOTREACHED*/
    return nil;
}

- ( NSString * )remotePathForEditedDocumentWithLocalPath: (NSString *)localPath
{
    return [[ self editedDocumentWithLocalPath: localPath ] objectForKey: @"remotepath" ];
}

- ( void )insertObject:(id)object inEditedDocumentsAtIndex: ( int )index
{
    // XXX - not needed?
}

- ( void )removeObjectFromEditedDocumentsAtIndex: ( int )index
{
    NSString* localPath = [[ editedDocuments objectAtIndex: index ] objectForKey: @"localpath" ];

//    NSLog(@"Removing edited document: %@", localPath);

    [ kqueue removePathFromQueue: localPath ];

    if ( [ [ NSFileManager defaultManager ] removeFileAtPath: [ localPath stringByDeletingLastPathComponent ]
                                                     handler: nil ] == NO )
    {
        NSRunAlertPanel( NSLocalizedString( @"Warning", @"Warning" ),
                         @"Failed to delete temporary file %@", NSLocalizedString( @"OK", @"OK" ), @"", @"",
                         localPath);
    }

    [ kqueue removePathFromQueue: localPath ];
    [ editedDocuments removeObjectAtIndex: index ];
}

- ( void )removeEditedDocumentWithLocalPath: ( NSString * )localPath
{
    NSDictionary* editedDocument = [ self editedDocumentWithLocalPath: localPath ];
    
    [ self removeObjectFromEditedDocumentsAtIndex: [ editedDocuments indexOfObjectIdenticalTo: editedDocument ] ];
}

#pragma mark -- Preview --

- ( void )setPreviewedImage: ( NSString * )imagepath
{
    if ( previewedImage != nil ) {
        [ previewedImage release ];
        previewedImage = nil;
    }
    
    if ( imagepath == nil ) {
        return;
    }
    
    previewedImage = [[ NSString alloc ] initWithString: imagepath ];
}

- ( NSString * )previewedImage
{
    return( previewedImage );
}

- ( void )addToCachedPreviews: ( NSDictionary *)cachedData
{
    if ( cachedPreviews == nil ) {
        cachedPreviews = [[ NSMutableArray alloc ] init ];
    }
    [ cachedPreviews addObject: cachedData ];
}

- ( NSArray * )cachedPreviews
{
    return( cachedPreviews );
}

#pragma mark -- Rendezvous --

- ( IBAction )selectRendezvousServer: ( id )sender
{
    id  server = [[[ NSApp delegate ] rendezvousServers] objectAtIndex: [ sender indexOfSelectedItem ] - 1];

    [ remoteHost setStringValue: [ server name ]];
    [ portField setObjectValue: [ server port ]];
}

#pragma mark -- NSSplitView delegate methods --

- ( float )splitView: ( NSSplitView * )splitview constrainMaxCoordinate: ( float )proposedMax
            ofSubviewAt: ( int )offset
{
    return(( proposedMax - 175 ));
}

- ( float )splitView: ( NSSplitView * )splitview constrainMinCoordinate: ( float )proposedMax
            ofSubviewAt: ( int )offset
{
    return(( proposedMax + 175 ));
}

#pragma mark -- NSTabView delegate methods --

- ( BOOL )tabView: ( NSTabView * )tabView shouldSelectTabViewItem: ( NSTabViewItem * )tabViewItem
{
    /* don't allow user to select remote info tab if not connected */
    if ( ! [ [ self client ] isConnected ] && [[ tabViewItem identifier ] isEqualToString: @"remote" ] ) {
        return( NO );
    }
    return( YES );
}

#pragma mark -- Open With... Menu --

static int sortAppURLs(id url1, id url2, void *context)
{
    return [ [ [ url1 path ] lastPathComponent ] caseInsensitiveCompare: [ [ url2 path ] lastPathComponent] ];
}

- ( NSMenu * )openWithMenu
{
    return openWithMenu;
}

- ( void )updateOpenWithMenuForSelectionsInTableView: (NSTableView*) tv
{
    NSArray*       fileURLs = nil;
    NSFileManager* fm       = [ NSFileManager defaultManager ];

//    [ self willChangeValueForKey: @"openWithMenu" ];

    if ( tv == localBrowser )
    {
        NSArray*  dir   = dotflag ? localDirContents : dotlessLDir ;
        NSArray*  files = [ dir subarrayWithIndicesFromEnumerator: [ tv selectedRowEnumerator ] ];

        fileURLs = [ files valueForKeyPath: @"name.fileURL" ];
    }
    else if ( tv == remoteBrowser )
    {
        NSArray*       dir   = dotflag ? remoteDirContents : dotlessRDir ;
        NSArray*       files = [ dir subarrayWithIndicesFromEnumerator: [ tv selectedRowEnumerator ] ];
        NSEnumerator*  fileEnumerator = [ files objectEnumerator ];
        NSDictionary*  file;

        // KLUDGE:
        //
        // LaunchServices insists on requiring a URL to a file that actually
        // exists in the local file system, when attempting to determine the
        // applications for a given file type, instead of providing a way
        // to look this information up based on type only.
        //
        // Therefore, here we create a bunch of temporary files that will
        // be deleted after we get the information we need.

        fileURLs = [ NSMutableArray arrayWithCapacity: [ files count ] ];

        while ( (file = [fileEnumerator nextObject]) != nil )
        {
            NSString* pathExt = [ file valueForKeyPath: @"name.lastPathComponent.pathExtension" ];
            NSURL*    fileURL = [ [ fm makeTemporaryFileWithExtension: pathExt ] fileURL ];

            [ (NSMutableArray*)fileURLs addObject: fileURL ];
        }
    }

    NSArray*      appsURLs = [ [ NDLaunchServicesObject defaultLaunchServices ] applicationsForURLs: fileURLs
                                                                                             inRole: kLSRolesAll ];
    NSMutableSet* apps     = [ NSMutableSet set ];
    NSEnumerator* appURLEnumerator = [ [ appsURLs sortedArrayUsingFunction: sortAppURLs
                                                                   context: nil ] objectEnumerator ];
    NSURL*        appURL;

    [ openWithMenu removeAllItems ];

    while ( (appURL = [appURLEnumerator nextObject]) != nil )
    {
        NSString* appPath = [ appURL path ];
        NSString* appName = [ [ appPath lastPathComponent ] stringByDeletingPathExtension ];

        if ( [ apps member: appName ] == nil )  // New application
        {
            NSImage*    appImage = [ [ NSWorkspace sharedWorkspace ] iconForFile: appPath ];
            NSMenuItem* appItem  = [ openWithMenu addItemWithTitle: appName
                                                            action: @selector(editFile:)
                                                     keyEquivalent: @"" ];

            [ appImage setSize: NSMakeSize(16, 16) ];
            [ appItem setImage: appImage ];
            [ appItem setTag: FILE_MENU_ITEM_OPEN_WITH ];

            [ apps addObject: appName ];
        }
    }

    // Create "Other..." menu item
    [ [ openWithMenu addItemWithTitle: NSLocalizedString(@"Other...", @"Open With > Other... Menu Item")
                               action: @selector(editFile:)
                        keyEquivalent: @""]
        setTag: OPEN_WITH_MENU_ITEM_OTHER ];

    [ localMenuOpenWithMenuItem  setSubmenu: [ [ openWithMenu copy ] autorelease ] ];
    [ remoteMenuOpenWithMenuItem setSubmenu: [ [ openWithMenu copy ] autorelease ] ];

//    [ self didChangeValueForKey: @"openWithMenu" ];

    // KLUDGE:
    //
    // Clean up temporary files created by kludge above:

    if ( tv == remoteBrowser )
    {
        NSEnumerator*  fileURLEnumerator = [ fileURLs objectEnumerator ];
        NSURL*  file;

        while ( (file = [ fileURLEnumerator nextObject ]) != nil )
        {
            BOOL ok = [ fm removeFileAtPath: [ file path ]
                                    handler: nil ];

            NSAssert( ok, @"Failed to delete temporary files!" );
        }
    }
}

#pragma mark -- NSTableView delegate methods --

/* tableview delegate methods */
/* contextual menu for tables */
- ( NSMenu * )menuForTable: ( SFTPTableView * )table
                column: ( int )column row: ( int )row
{
    if ( [ table isEqual: localBrowser ] ) {
        return( localTableMenu );
    } else if ( [ table isEqual: remoteBrowser ] ) {
        return( remoteTableMenu );
    }
    return( nil );
}

- ( void )tableViewDidReloadData: ( NSTableView * )tableView
{
//    NSLog(@"Table view reloaded.");

    [ self updateOpenWithMenuForSelectionsInTableView: tableView ];
}

- ( void )tableViewSelectionDidChange: ( NSNotification * )aNotification
{
    id 				tv = [ aNotification object ];
    int				numselected;
    
    if ( ! [ tv isKindOfClass: [ SFTPTableView class ]] ) {
        return;
    }
    
    numselected = [ tv numberOfSelectedRows ];
    
    if ( [ tv editedRow ] >= 0 ) {
        [ tv abortEditing ];
    }
    
    if ( [ infoPanel isVisible ] && numselected == 1 ) {
        [ self getInfo: nil ];
    }
    
    if ( [ imagePreviewPanel isVisible ] && numselected == 1 ) {
        [ self previewItem: nil ];
    }

    [ self updateOpenWithMenuForSelectionsInTableView: tv ];
}

- ( void )tableView: ( NSTableView * )table
        didClickTableColumn: ( NSTableColumn * )tableColumn
{
    NSArray		*columns = [ table tableColumns ];
    int			i, sortdirection;
    int			( *sortFunction )( id, id, void * );
    unsigned		context;
    id			identifier;
    NSImage		*image = nil;
    
    image = [ table indicatorImageInTableColumn: tableColumn ];
    if ( [ image isEqual:
            [ NSImage imageNamed: @"NSAscendingSortIndicator" ]] ) {
        image = [ NSImage imageNamed: @"NSDescendingSortIndicator" ];
        sortdirection = 1;
    } else {
        image = [ NSImage imageNamed: @"NSAscendingSortIndicator" ];
        sortdirection = 0;
    }
    [ image retain ];
    
    for ( i = 0; i < [ columns count ]; i++ ) {
        [ table setIndicatorImage: nil
                inTableColumn: [ columns objectAtIndex: i ]];
    }
            
    identifier = [ tableColumn identifier ];
    
    sortFunction = sortFunctionForIdentifier( identifier );
    
    if ( [[ NSUserDefaults standardUserDefaults ]
            boolForKey: @"ASCIIOrderSorting" ] ) {
        context = NSLiteralSearch;
    } else {
        context = NSCaseInsensitiveSearch;
    }
    
    if ( [ table isEqual: localBrowser ] ) {
        if ( [[[ table highlightedTableColumn ] identifier ]
                        isEqualToString: identifier ] ) {
            [ localDirContents reverse ];
        } else {
            [ localDirContents sortUsingFunction: sortFunction
                            context: ( void * )&context ];
        }
        [ dotlessLDir removeAllObjects ];
        [ dotlessLDir addObjectsFromArray: [ localDirContents visibleItems ]];
        [[ NSUserDefaults standardUserDefaults ] setObject: identifier
                                forKey: @"LocalBrowserSortingIdentifier" ];
        [[ NSUserDefaults standardUserDefaults ] setObject:
            [ NSNumber numberWithInt: sortdirection ]
                            forKey: @"LocalBrowserSortDirection" ];
    } else if ( [ table isEqual: remoteBrowser ] ) {
        if ( [[[ table highlightedTableColumn ] identifier ]
                        isEqualToString: identifier ] ) {
            [ remoteDirContents reverse ];
        } else {
            [ remoteDirContents sortUsingFunction: sortFunction
                            context: ( void * )&context ];
        }
        [ dotlessRDir removeAllObjects ];
        [ dotlessRDir addObjectsFromArray: [ remoteDirContents visibleItems ]];
        [[ NSUserDefaults standardUserDefaults ] setObject: identifier
                                forKey: @"RemoteBrowserSortingIdentifier" ];
        [[ NSUserDefaults standardUserDefaults ] setObject:
            [ NSNumber numberWithInt: sortdirection ]
                            forKey: @"RemoteBrowserSortDirection" ];
    }
    
    [ table setHighlightedTableColumn: tableColumn ];
    [ table setIndicatorImage: image
            inTableColumn: tableColumn ];
    [ image release ];
    [ table reloadData ];
}

- ( BOOL )tableView: ( NSTableView * )aTableView
        shouldEditTableColumn: ( NSTableColumn * )aTableColumn
        row: ( int )rowIndex
{
    /* if user's chosen to rename a file, allow editing */
    if ( [ aTableView editedRow ] >= 0 ) return( YES );
    
    return( NO );	/* so doubleclicks register as such */
}
            

- ( int )numberOfRowsInTableView: ( NSTableView * )aTableView
{
    NSMutableArray	*items = nil;
    
    if ( [ aTableView isEqual: localBrowser ] ) {
        if ( !dotflag ) {
            items = dotlessLDir;
        } else {
            items = localDirContents;
        }
    } else if ( [ aTableView isEqual: remoteBrowser ] ) {
        if ( !dotflag ) {
            items = dotlessRDir;
        } else {
            items = remoteDirContents;
        }
    }
    
    return( [ items count ] );
}

#ifndef GNUSTEP
- (void)tableView:(NSTableView *)aTableView willDisplayCell:(id)aCell forTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex
{
    NSString * identifier = [aTableColumn identifier];

    // XXX - ugly kludge because Cocoa doesn't handle bindings for NSButtonCell in NSTableView
    if ([identifier isEqualToString: @"cancel"])
    {
        if ([aCell isEnabled])
        {
            id task = [[mTasksController arrangedObjects] objectAtIndex:rowIndex];

            // Make sure that any old bindings are undone, first.
            // Failure to do this can lead to crashes (bindings to freed objects)
            [aCell unbind: @"target"];
            [aCell unbind: @"argument"];

            [aCell bind: @"target"   toObject: [self client] withKeyPath: @"self" options: [NSDictionary dictionaryWithObject: @"cancelCommand:" forKey: NSSelectorNameBindingOption]];
            [aCell bind: @"argument" toObject: task withKeyPath: @"self" options: nil];
        }
    }
    else if ([identifier isEqualToString: @"remove"])
    {
        id document = [[mEditedDocumentsController arrangedObjects] objectAtIndex:rowIndex];
        
        // Make sure that any old bindings are undone, first.
        // Failure to do this can lead to crashes (bindings to freed objects)
        [aCell unbind: @"target"];
        [aCell unbind: @"argument"];
        
        [aCell bind: @"target"   toObject: self withKeyPath: @"self" options: [NSDictionary dictionaryWithObject: @"removeEditedDocumentWithLocalPath:" forKey: NSSelectorNameBindingOption]];
        [aCell bind: @"argument" toObject: document withKeyPath: @"localpath" options: nil];
    }
}
#endif

- ( id )tableView: ( NSTableView * )view
        objectValueForTableColumn: ( NSTableColumn * )aTableColumn
        row: ( int )row
{
    NSString			*path = nil, *ext = nil;
    NSAttributedString          *attrString = nil;
    NSImage			*image = nil;
    int				therow = row;
    NSString			*name = nil;
    id				cell = [ aTableColumn dataCell ];
    double                      width = [ aTableColumn width ];
    
    if ( row < 0 ) {
        return( nil );
    }

    if ( view == localBrowser ) {
        NSDictionary		*dict;
        NSString		*resolvedpath;
        
        if ( !dotflag ) {
            dict = [ dotlessLDir objectAtIndex: therow ];
        } else {
            dict = [ localDirContents objectAtIndex: therow ];
        }

        resolvedpath = [ dict objectForKey: @"resolvedAlias" ];
        
        if ( [[ aTableColumn identifier ] isEqualToString: @"namecolumn" ] ) {
            path = [ dict objectForKey: @"name" ];
            ext = [ path pathExtension ];

            if ( access(( char * )[ path UTF8String ], F_OK ) == 0 &&
                        resolvedpath == nil ) {
                /* display name at path shows resolved symlink names */
                name = [[ NSFileManager defaultManager ] displayNameAtPath: path ];
            } else {
                name = [ path lastPathComponent ];
            }
            
            image = [ dict objectForKey: @"icon" ];
            
            [ cell setImage: image ];
            [ cell setEditable: YES ];
            attrString = [[[ NSAttributedString alloc ] initWithString: name ] autorelease ];
            
            attrString = [ attrString ellipsisAbbreviatedStringForWidth: width ];
            
            return( attrString );
        } else if ( [[ aTableColumn identifier ] isEqualToString: @"datecolumn" ] ) {
            NSString		*date = [ dict objectForKey: @"date" ];
            
            if ( date != nil ) {
                return( date );
            }
        } else if ( [[ aTableColumn identifier ] isEqualToString: @"sizecolumn" ] ) {
            NSString		*size = [ dict objectForKey: @"size" ];
            
            if ( size != nil ) {
                return( [ size descriptiveSizeString ] );
            }
        } else if ( [[ aTableColumn identifier ] isEqualToString: @"ownercolumn" ] ) {
            NSString		*owner = [ dict objectForKey: @"owner" ];
            
            if ( owner != nil ) {
                return( owner );
            }
        } else if ( [[ aTableColumn identifier ] isEqualToString: @"groupcolumn" ] ) {
            NSString		*group = [ dict objectForKey: @"group" ];
            
            if ( group != nil ) {
                return( group );
            }
        } else if ( [[ aTableColumn identifier ] isEqualToString: @"permcolumn" ] ) {
            NSString		*mode = [ dict objectForKey: @"perm" ];
            
            if ( mode != nil ) {
                return( [ mode stringRepresentationOfOctalMode ] );
            }
        }
    } else if ( [ view isEqual: remoteBrowser ] ) {
        NSDictionary		*dict;
        
        if ( !dotflag ) {
            dict = [ dotlessRDir objectAtIndex: therow ];
        } else {
            dict = [ remoteDirContents objectAtIndex: therow ];
        }

        if ( [[ aTableColumn identifier ] isEqualToString: @"namecolumn" ] ) {
            name = [[ dict objectForKey: @"name" ] lastPathComponent ];
            ext = [ name pathExtension ];
        
            switch ( [[ dict objectForKey: @"perm" ] characterAtIndex: 0 ] ) {
            case '-':
            default:
                if ( [ ext isEqualToString: @"" ] ) {
                    image = fileImage;
                } else {
                    image = [ NSImage iconForType: ext ];
                    [ image setSize: NSMakeSize( 16.0, 16.0 ) ];
                }
                break;
                
            case 'd':
                image = dirImage;
                break;
                
            case 'l':
                //[ cell italicizeStringValue ];
                image = linkImage;
                break;
            }
            
            [ cell setImage: image ];
            [ cell setEditable: YES ];
            
            attrString = [[[ NSAttributedString alloc ] initWithString: name ] autorelease ];
            attrString = [ attrString ellipsisAbbreviatedStringForWidth: width ];
            
            return( attrString );
        } else if ( [[ aTableColumn identifier ] isEqualToString: @"sizecolumn" ] ) {
            NSString		*size = [ dict objectForKey: @"size" ];
            
            if ( size != nil ) {
                return( [ size descriptiveSizeString ] );
            }
        } else if ( [[ aTableColumn identifier ] isEqualToString: @"ownercolumn" ] ) {
            NSString		*owner = [ dict objectForKey: @"owner" ];
            
            if ( owner == nil ) {
                owner = @"";
            }
            [ cell setEditable: NO ];
            return( owner );
        } else if ( [[ aTableColumn identifier ] isEqualToString: @"datecolumn" ] ) {
            NSString		*date = [ dict objectForKey: @"date" ];
            
            if ( date == nil ) {
                date = @"";
            }
            [ cell setEditable: NO ];
            return( date );
        } else if ( [[ aTableColumn identifier ] isEqualToString: @"groupcolumn" ] ) {
            NSString		*group = [ dict objectForKey: @"group" ];
            
            if ( group != nil ) {
                return( group );
            }
        } else if ( [[ aTableColumn identifier ] isEqualToString: @"permcolumn" ] ) {
            NSString		*mode = [ dict objectForKey: @"perm" ];
            
            if ( mode != nil ) {
                return( mode );
            }
        }
    }

    [ cell setEditable: NO ];
    return( @"" );
}

- ( void )tableView: ( NSTableView * )aTableView
        setObjectValue: ( id )anObject
        forTableColumn: ( NSTableColumn * )aTableColumn
        row: ( int )rowIndex
{
}

/* drag and drop related methods */

- ( void )tableView: ( NSTableView * )tableView dragOperationEnded:(NSDragOperation)operation
{
    NSPasteboard		*pb;

    // NSLog(@"Drag operation ended: %d", operation);

    pb = [NSPasteboard pasteboardWithName: NSDragPboard];

    if ( tableView == localBrowser ) {
        switch (operation) {
            case NSDragOperationDelete:
            {
                NSArray* files = [ pb propertyListForType: NSFilenamesPboardType ];

                [ self deleteLocalFiles: [ files valueForKey: @"lastPathComponent" ] ];
            } break;
        }
    } else if ( tableView == remoteBrowser ) {
        NSArray* files = [ pb propertyListForType: FuguRemoteFilesPboardType ];

        switch (operation) {
            case NSDragOperationDelete:
            {
                [ self deleteRemoteFiles: files ];
            } break;
                
            case NSDragOperationCopy:
            case NSDragOperationGeneric:  // What a mess Drag & Drop is.  See http://www.cocoadev.com/index.pl?NSDragOperationGeneric
            {
                if ( promisedFilesDestinationURL != nil )
                {
                    // NSLog(@"Downloading promised files");

                    [ self downloadFiles: files
                             toDirectory: [ promisedFilesDestinationURL path] ];
                    
                    [ promisedFilesDestinationURL release ];
                    promisedFilesDestinationURL = nil;
                }
            } break;
        }
    }
}

- ( BOOL )tableView: ( NSTableView * )tableView acceptDrop: ( id <NSDraggingInfo> )info
    row: ( int )row dropOperation: ( NSTableViewDropOperation )operation
{
    NSPasteboard		*pb;
    id				dragData;

    pb = [ info draggingPasteboard ];
    
    /* clear spring-loaded root path */
    if ( _springLoadedRootPath != nil ) {
        [ _springLoadedRootPath release ];
        _springLoadedRootPath = nil;
    }
    
    if ( tableView == localBrowser ) {
        NSString		*lpath = localDirPath;
        
        if ( operation == NSTableViewDropOn && row >= 0 ) {
            lpath = [[ ( dotflag ? localDirContents : dotlessLDir )
                        objectAtIndex: row ] objectForKey: @"name" ];
        }

        dragData = [ pb propertyListForType: FuguRemoteFilesPboardType ];

        if ( [ dragData count ] == 1 &&
                [[[ dragData objectAtIndex: 0 ] objectForKey: @"type" ]
                isEqualToString: @"directory" ] ) {
            int			rc;

            rc = NSRunAlertPanel(
                NSLocalizedString( @"Warning: OpenSSH's sftp client cannot yet download directories.",
                                @"Warning: OpenSSH's sftp client cannot yet download directories." ),
                NSLocalizedString( @"Would you like to download %@ to %@ with SCP instead?",
                                @"Would you like to download %@ to %@ with SCP instead?" ),
                NSLocalizedString( @"Download", @"Download" ),
                NSLocalizedString( @"Cancel", @"Cancel" ), @"",
                [[ dragData objectAtIndex: 0 ] objectForKey: @"name" ], lpath );
            switch ( rc ) {
            case NSAlertDefaultReturn:
                [ self scpRemoteItem: [[ dragData objectAtIndex: 0 ] objectForKey: @"name" ]
                            fromHost: [ remoteHost stringValue ]
                         toLocalPath: lpath userName: [ userName stringValue ]];
            case NSAlertAlternateReturn:
            default:
                return( YES );
            }
        }

        [ self downloadFiles: dragData toDirectory: lpath ];
        return( YES );
    } else if ( tableView == remoteBrowser ) {
        NSString		*rpath = remoteDirPath;
        
        if ( ![ [ self client ] isConnected ] ) return( NO );
        if ( operation == NSTableViewDropOn && row >= 0 ) {
            rpath = [[ ( dotflag ? remoteDirContents : dotlessRDir ) objectAtIndex: row ]
                                                                        objectForKey: @"name" ];
        }

        dragData = [ pb propertyListForType: NSFilenamesPboardType ];

        [ self uploadFiles: dragData toDirectory: rpath ];        
        return( YES );
    }
    
    return( NO );
}

- ( NSDragOperation )tableView: ( NSTableView * )tableView validateDrop:( id <NSDraggingInfo> )info
    proposedRow: ( int )row proposedDropOperation: ( NSTableViewDropOperation )operation
{
    id			dsrc;
    
    if ( operation == NSTableViewDropOn && row == -1 ) {
        return( NSDragOperationCopy );
    } else if ( operation == NSTableViewDropAbove ) {
        [ tableView setDropRow: -1 dropOperation: NSTableViewDropOn ];
        return( NSDragOperationCopy );
    } else if ( operation == NSTableViewDropOn && row >= 0 ) {
        if ( tableView == remoteBrowser ) {
            if ( dotflag ) dsrc = [ remoteDirContents copy ];
            else dsrc = [ dotlessRDir copy ];
            
            if ( [[[ dsrc objectAtIndex: row ]
                    objectForKey: @"perm" ] characterAtIndex: 0 ] == 'd' ) {
                [ dsrc release ];
                return( NSDragOperationCopy );
            }
            [ dsrc release ];
        } else {
            BOOL	isDir;
            
            if ( dotflag ) dsrc = [ localDirContents copy ];
            else dsrc = [ dotlessLDir copy ];
            
            if ( [[ NSFileManager defaultManager ] fileExistsAtPath:
                    [[ dsrc objectAtIndex: row ] objectForKey: @"name" ]
                    isDirectory: &isDir ] && isDir ) {
                [ dsrc release ];
                return( NSDragOperationCopy );
            }
            [ dsrc release ];
        }
    }

    return( NSDragOperationNone );
}

- ( BOOL )tableView: ( NSTableView * )tableView writeRows: ( NSArray * )rows
    toPasteboard: ( NSPasteboard * )pboard
{
    NSAssert( [ [ self client ] isConnected ] || (![ [ self client ] isConnected ] && tableView == localBrowser),
              @"Drag from remote browser while not connected!");

    if ( tableView == localBrowser ) {
        NSArray*   dir   = dotflag ? localDirContents : dotlessLDir;
        NSArray*   files = [ dir subarrayWithIndicesFromEnumerator: [ rows objectEnumerator] ];

        [ pboard declareTypes: [ NSArray arrayWithObject: NSFilenamesPboardType ]
                        owner: self ];
        [ pboard setPropertyList: [ files valueForKey: @"name" ]
                         forType: NSFilenamesPboardType ];
    } else if ( tableView == remoteBrowser ) {
        NSArray*   dir   = dotflag ? remoteDirContents : dotlessRDir;
        NSArray*   files = [ dir subarrayWithIndicesFromEnumerator: [ rows objectEnumerator] ];
        NSArray*   exts  = [ files valueForKeyPath: @"name.pathExtension" ];
        NSArray*   pboardTypes;

        // Declare pasteboard datatypes we're going to write
        pboardTypes =
            [ NSArray arrayWithObjects:
                FuguRemoteFilesPboardType,
                NSFilesPromisePboardType,
                nil ];
        [ pboard declareTypes: pboardTypes owner: self ];

        // Write pasteboard data
        [ pboard setPropertyList: files forType: FuguRemoteFilesPboardType ];
        [ pboard setPropertyList: exts  forType: NSFilesPromisePboardType ];
    }

    return( YES );
}

- ( NSArray* )tableView: ( NSTableView * )tableView namesOfPromisedFilesDroppedAtDestination: ( NSURL * )url
{
    NSArray*       result = nil;
    NSPasteboard*  pboard = [ NSPasteboard pasteboardWithName: NSDragPboard ];

    // NSLog( @"[%@] Dropped to: %@", (tableView == localBrowser ? @"Local" : @"Remote"), [ url path ] );

    if ( tableView == remoteBrowser )
    {
        NSArray* rfiles = [ pboard propertyListForType: FuguRemoteFilesPboardType ];

        NSAssert( rfiles != nil, @"No remote file data on pasteboard." );

        promisedFilesDestinationURL = [ url retain ];

        result = [ rfiles valueForKeyPath: @"name" ];
        
        // NSLog(@"Returning promised file names: %@", result);
    }

    return result;
}

#pragma mark -- SCP --

/* download files/directories with scp */
- ( IBAction )scp: ( id )sender
{
    int			row;
    id			item;
    
    if ( [ sender isKindOfClass: [ NSMenuItem class ]] )
    {
        /* check to make sure we've got something valid */
        NSString	*user = [ userName stringValue ], *host = [ remoteHost stringValue ];
        if ( ! [ user length ] ) user = NSUserName();
        if ( ! [ host length ] ) host = @"127.0.0.1";

        if ( [[ sender menu ] isEqual: localTableMenu ] ) {
            
            row = [ localBrowser selectedRow ];
            if ( row < 0 ) return;
            
            if ( dotflag ) {
                item = [[ localDirContents objectAtIndex: row ]
                                objectForKey: @"name" ];
            } else {
                item = [[ dotlessLDir objectAtIndex: row ]
                                objectForKey: @"name" ];
            }
            
            
            [ self scpLocalItem: item
                         toHost: host
                       userName: user ];
        }
        else if ( [[ sender menu ] isEqual: remoteTableMenu ] )
        {
            row = [ remoteBrowser selectedRow ];
            if ( row < 0 ) return;
            
            if ( dotflag ) {
                item = [[ remoteDirContents objectAtIndex: row ]
                                objectForKey: @"name" ];
            } else {
                item = [[ dotlessRDir objectAtIndex: row ] objectForKey: @"name" ];
            }
            
            [ self scpRemoteItem: item
                        fromHost: host
                     toLocalPath: localDirPath
                        userName: user ];
        }
    }
}

- ( void )scpRemoteItem: ( NSString * )rdir fromHost: ( NSString * )rhost
            toLocalPath: ( NSString * )ldir userName: ( NSString * )user
{
    NSString* fullRemoteDirPath = [ NSString stringWithFormat: @"%@/%@", remoteDirPath, rdir ];
    
    [ [ [ NSApp delegate ]  scp ] getSecureCopyWindowForFile: fullRemoteDirPath
                                                     scpType: 1
                                                  copyToPath: ldir
                                                    fromHost: rhost
                                                    userName: user
                                                    delegate: self ];
}

- ( void )scpLocalItem: ( NSString * )item toHost: ( NSString * )rhost
              userName: ( NSString * )user
{
    NSString* dir = remoteDirPath;
    
    if ( dir == nil ) dir = @" ";
    
    [ [ [ NSApp delegate ] scp ] getSecureCopyWindowForFile: item
                                                    scpType: 0
                                                 copyToPath: dir
                                                   fromHost: rhost
                                                   userName: user
                                                   delegate: self ];
}

/* SCPController delegate method, called when an scp completes */
- ( void )scpFinished
{
    if ( [ [ self window ] isVisible ] && localDirPath != nil )
    {
        [ self localBrowserReloadForPath: localDirPath ];
    }
}

#pragma mark -- NSMenu delegate methods & notifications --

- ( void )rendezvousMenuUpdate: ( NSNotification * )note
{
    int index = [[[ note userInfo] objectForKey: @"NSMenuItemIndex" ] intValue ];

    if ( index == 0 )
    {
        [[ rendezvousPopUp itemAtIndex: index] setImage: [ NSImage imageNamed: @"zeroconf" ] ];
    }
}

@end
