//
//  SFTPFileTransferCommand.m
//  Fugu
//
//  Created by Tim McIntosh on 2005.10.08.
//  Copyright 2005 __MyCompanyName__. All rights reserved.
//

#import "SFTPFileTransferCommand.h"
#import "SFTPClientTaskController.h"
#import "NSString+CocoaDevUsersAdditions.h"
#import "NSString+SFTPAdditions.h"


@implementation SFTPFileTransferCommand

- (SFTPFileTransferCommand*)initWithLocalPath:(NSString*)localPath
                                   remotePath:(NSString*)remotePath
                            transferDirection:(SFTPFileTransferDirection)direction
{
    self = [super init];
    if (self != nil)
    {
        mDirection  = direction;
        mLocalPath  = [localPath copy];
        mRemotePath = [remotePath copy];
    }
    return self;
}

- (void) dealloc
{
    [mLocalPath release];
    [mRemotePath release];

    [mEta release];
    [mBytesTransferred release];
    [mTransferRate release];

    [super dealloc];
}

- (NSString*)path
{
    return [self isUpload] ? mLocalPath : mRemotePath;
}

- (NSString*)file
{
    return [[self path] lastPathComponent];
}

- (double)progress
{
    return mProgress;
}

- (NSString*)eta
{
    return mEta ? mEta : @"--:--";
}

- (NSString*)bytesTransferred
{
    return mBytesTransferred ? mBytesTransferred : @"0";
}

- (NSString*)transferRate
{
    return mTransferRate ? mTransferRate : @"0.0KB/s";
}

// private
- (void)setEta:(NSString*)eta
{
    if (eta != mEta)
    {
        [mEta release];
        mEta = [eta copy];
    }
}

- (void)setBytesTransferred:(NSString*)bytesTransferred
{
    if (bytesTransferred != mBytesTransferred)
    {
        [mBytesTransferred release];
        mBytesTransferred = [bytesTransferred copy];
    }
}

- (void)setTransferRate:(NSString*)transferRate
{
    if (transferRate != mTransferRate)
    {
        [mTransferRate release];
        mTransferRate = [transferRate copy];
    }
}

- (void)setProgress:(double)progress
{
    mProgress = progress;
}

- (void)setIsComplete:(BOOL)isComplete
{
    [self setProgress: isComplete ? 0 : 100];
    [super setIsComplete: isComplete];
}

- (BOOL)isUpload
{
    return mDirection == SFTPFileTransferDirectionUpload;
}

- (void)updateProgress:(NSMutableString*)response
{
    /* Output looks like:
       status line \r\n   (note:  controller converts \r\n to \n)
       progress line \r
       progress line \r
       ...
       progress line \n
       prompt
     */
    NSCharacterSet*  crSet     = [NSCharacterSet characterSetWithCharactersInString: @"\r"];
    NSCharacterSet*  nlSet     = [NSCharacterSet characterSetWithCharactersInString: @"\n"];
    NSCharacterSet*  crNlSet   = [NSCharacterSet characterSetWithCharactersInString: @"\r\n"];

    // Delete all lines from first nl to cr prior to last cr/nl, leaving at most 1 progress line
    NSRange          firstNl   = [response rangeOfCharacterFromSet: nlSet];
    NSRange          lastCrNl  = [response rangeOfCharacterFromSet: crNlSet options: NSBackwardsSearch];

    if (firstNl.location  != lastCrNl.location &&
        firstNl.location  != NSNotFound &&
        lastCrNl.location != NSNotFound)
    {
        //NSLog(@"Checking for progress string in \"%@\"", response);
        unsigned int start = firstNl.location + firstNl.length;

        NSRange      previousCr = [response rangeOfCharacterFromSet: crSet
                                                            options: NSBackwardsSearch
                                                              range: NSMakeRange(start,
                                                                                 lastCrNl.location - start)];

        NSString*   progressString;

        if (previousCr.location == NSNotFound)
        {
            // Only 1 progress line present.
            progressString = [response substringWithRange: NSMakeRange(start, lastCrNl.location - start)];
        }
        else
        {
            // Found obsolete status lines - delete them after grabbing progress string
            unsigned int end = previousCr.location + previousCr.length;

            progressString = [response substringWithRange: NSMakeRange(end, lastCrNl.location - end)];

            [response deleteCharactersInRange: NSMakeRange(start,
                                                           end - start)];
        }

        //NSLog(@"Progress: \"%@\"", progressString);

        // format:  filename   0% 1KB 0.0KB/s MM:SS ETA
        // ETA only printed while in progress
        // KB on size field only printed when != 0
        // fields:
        enum
        {
            kEntireMatch,
            kProgress,
            kBytesTransferred,
            kBytesTransferredUnits,
            kTransferRate,
            kTransferRateUnits,
            kEta,
            kEtaMarker,
            kNumFields
        };
        
        static NSString* const re =
            @"^.*[^[:blank:]][[:blank:]]+([[:digit:]]+)%[[:blank:]]+([[:digit:]]+(KB|MB)?)[[:blank:]]+([[:digit:]]+[[:punct:]][[:digit:]]+(KB|MB)/s)[[:blank:]]+(--:--|[[:digit:]]{2}:[[:digit:]]{2})[[:blank:]]*(ETA)?$";
        
        NSArray* matches = [progressString findRegularExpression:(NSString *)re ignoreCase:YES];  // XXX - ignore case may not be needed
        
        if (matches != nil)
        {
            NSAssert1([matches count] == kNumFields, @"Number of fields matched is incorrect: %@", matches);
            
            //NSLog(@"Matched: %@", matches);
            
            [self setProgress:        [[matches objectAtIndex: kProgress] doubleValue]];
            [self setBytesTransferred: [matches objectAtIndex: kBytesTransferred]];
            [self setTransferRate:     [matches objectAtIndex: kTransferRate]];
            [self setEta:              [matches objectAtIndex: kEta]];
        }
    }
}

- (NSString*)command
{
    if (mDirection == SFTPFileTransferDirectionDownload)
    {
        return [NSString stringWithFormat: @"get \"%@\" \"%@\"",
            [[self controller] absolutePathFromString: mRemotePath],
            [[self controller] absolutePathFromString: mLocalPath]];
    }
    else
    {
        return [NSString stringWithFormat: @"put \"%@\" \"%@\"",
            [[self controller] absolutePathFromString: mLocalPath],
            [[self controller] absolutePathFromString: mRemotePath]];
    }
}

- (NSString*)displayName
{
    if (mDirection == SFTPFileTransferDirectionDownload)
    {    
        return [NSString stringWithFormat: NSLocalizedString(@"download %@", @"get command"), mRemotePath];
    }
    else
    {
        return [NSString stringWithFormat: NSLocalizedString(@"upload %@", @"put command"), mLocalPath];
    }
}

- (void)parseResponse:(NSArray*)response
{
    NSAssert([response count] >= 1, @"Unexpected count in 'get'/'put'");
    
    NSString* status = [response lastObject];
    
    if ([status containsSftpError])
    {
        [[self controller] setError: status];
    }
    else
    {
//        NSLog(@"Transfer complete");

        if (mDirection == SFTPFileTransferDirectionUpload)
        {
            [[self controller] setNeedsRefresh: YES];
        }
    }

    [super parseResponse: response];
}

@end
