/*
 * Copyright (c) 2003 Regents of The University of Michigan.
 * All Rights Reserved.  See COPYRIGHT.
 */

#import "SFTPNode.h"
#import "NSImage(IconForType).h"
#import "NSString(SSHAdditions).h"
#import "NSString(FSRefAdditions).h"

#include <tzfile.h>

// XXX
time_t
sixmo_in_secs()
{
    return(( DAYSPERNYEAR / 2 ) * SECSPERDAY );
}

@implementation SFTPNode

+ ( SFTPNode * )sharedInstance
{
    SFTPNode *sharedInstance = nil;
    
    if ( !sharedInstance ) {
        sharedInstance = [[ SFTPNode alloc ] init ];
    }
    
    return( [ sharedInstance autorelease ] );
}

- ( NSDictionary * )statInformationForPath: ( NSString * )path
{
    NSMutableDictionary* file = nil;
    NSDictionary*        attr;

    attr = [[ NSFileManager defaultManager ] fileAttributesAtPath: path traverseLink: NO ];

    if ( attr != nil )
    {
        file = [ NSMutableDictionary dictionary ];
        
        // XXX get rid of this conversion & use returned value directly - see also category on NSString used by remote files
        static NSDictionary*  remap;
        
        if (!remap)
        {
            remap = [[NSDictionary alloc] initWithObjectsAndKeys:
                @"symbolic link", NSFileTypeSymbolicLink,
                @"file", NSFileTypeRegular,
                @"directory", NSFileTypeDirectory,
                @"character special", NSFileTypeCharacterSpecial,
                @"block special", NSFileTypeBlockSpecial,
                @"socket", NSFileTypeSocket,
                // named pipe missing
                nil];
        }
        
        NSString* remapType = [ remap objectForKey: [attr fileType] ];
        
        if (!remapType)
        {
            remapType = @"unknown";
        }
        
        [ file setObject: NSLocalizedString(remapType, @"file type") forKey: @"type" ];
        
        /* only check if a file's an alias if it has a size of 0 */
        if ( [ attr fileSize ] == 0L && [ attr fileType ] == NSFileTypeRegular )
        {
            if ( [ path isAliasFile ] )
            {
                NSString* resolvedPath = [ path stringByResolvingAliasInPath ];
                
                if (resolvedPath != nil )
                {
                    [ file setObject: resolvedPath forKey: @"resolvedAlias" ];
                    [ file setObject: NSLocalizedString(@"alias", @"alias") forKey: @"type" ];
                }
            }
        }
        else if ( [ attr fileType ] == NSFileTypeSymbolicLink )
        {
            [ file setObject: path forKey: @"resolvedAlias" ];
        }

        NSDate*   theDate = [ attr fileModificationDate ];
        NSString* dateStr;
        
        if ( fabs( [ theDate timeIntervalSinceNow ] ) < sixmo_in_secs())
        {
            dateStr = [ theDate descriptionWithCalendarFormat: @"%b %e %H:%M"
                                                     timeZone: [ NSTimeZone defaultTimeZone ] locale: nil ];
        }
        else
        {
            dateStr = [ theDate descriptionWithCalendarFormat: @"%b %e %Y"
                                                     timeZone: [ NSTimeZone defaultTimeZone ]  locale: nil ];
        }
        
        NSString* owner = [ attr fileOwnerAccountName ];

        if ( owner == nil )
        {
            owner = [ [ attr fileOwnerAccountID ] stringValue ];
        }

        NSString* group = [ attr fileGroupOwnerAccountName ];

        if ( group == nil )
        {
            group = [ [ attr fileGroupOwnerAccountID ] stringValue ];
        }
        
        [ file setObject: [ NSImage iconForFile: path ] forKey: @"icon" ];
        [ file setObject: owner forKey: @"owner" ];
        [ file setObject: group forKey: @"group" ];
        [ file setObject: [ NSString stringWithFormat: @"%.4lo", [ attr filePosixPermissions ] ] forKey: @"perm" ];
        [ file setObject: [ NSString stringWithFormat: @"%lld", [ attr fileSize ] ] forKey: @"size" ];
        [ file setObject: dateStr forKey: @"date" ];
        [ file setObject: path forKey: @"name" ];
    }

    return file;
}

// Note: showHidden currently ignored by SFTPController.  XXX
- ( NSMutableArray * )itemsAtPath: ( NSString * )directory showHiddenFiles: ( BOOL )showHidden
{
    NSMutableArray* items = [ NSMutableArray array ];

    NSArray* directoryContents = [[ NSFileManager defaultManager ] directoryContentsAtPath: directory ];

    [ NSImage cacheIconsForPath: directory contents: directoryContents ];

    NSEnumerator* e = [ directoryContents objectEnumerator ];
    NSString*     fileName;

    while ( ( fileName = [ e nextObject ] ) != nil )
    {
        NSAutoreleasePool   *pool = [[ NSAutoreleasePool alloc ] init ];

        NSString* path = [ directory stringByAppendingPathComponent: fileName ];

        if ( showHidden || [ path pathIsVisible ] )
        {
            NSDictionary* file = [ self statInformationForPath: path ];
            
            if ( file == nil )
            {
                NSLog(@"%s: couldn't get information for %@ (%@)", __FUNCTION__, path, fileName);
            }
            else
            {
                [ items addObject: file ];
            }
        }

        [ pool release ];
    }

    return items;
}

@end
